#line 1 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
/*
* Copyright (c) 2015-2017, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  * Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*  * Neither the name of Intel Corporation nor the names of its contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

/** \file
* \brief Parser code (generated with Ragel from Parser.rl).
*/

#include "config.h"

/* Parser.cpp is a built source, may not be in same dir as parser files */
#include "parser/check_refs.h"
#include "parser/control_verbs.h"
#include "parser/ComponentAlternation.h"
#include "parser/ComponentAssertion.h"
#include "parser/ComponentAtomicGroup.h"
#include "parser/ComponentBackReference.h"
#include "parser/ComponentBoundary.h"
#include "parser/ComponentByte.h"
#include "parser/ComponentClass.h"
#include "parser/ComponentCondReference.h"
#include "parser/ComponentEmpty.h"
#include "parser/ComponentEUS.h"
#include "parser/Component.h"
#include "parser/ComponentRepeat.h"
#include "parser/ComponentSequence.h"
#include "parser/ComponentWordBoundary.h"
#include "parser/parse_error.h"
#include "parser/Parser.h"
#include "ue2common.h"
#include "util/compare.h"
#include "util/flat_containers.h"
#include "util/unicode_def.h"
#include "util/verify_types.h"

#include <cassert>
#include <cctype>
#include <cstring>
#include <cstdlib>
#include <map>
#include <sstream>
#include <string>
#include <vector>

using namespace std;

namespace ue2 {
	
#define PUSH_SEQUENCE do {\
		sequences.push_back(ExprState(currentSeq, (size_t)(ts - ptr), \
		mode)); \
	} while(0)
#define POP_SEQUENCE do {\
		currentSeq = sequences.back().seq; \
		mode = sequences.back().mode; \
		sequences.pop_back(); \
	} while(0)
	
	namespace {
		
		/** \brief Structure representing current state as we're parsing (current
		* sequence, current options). Stored in the 'sequences' vector. */
		struct ExprState {
			ExprState(ComponentSequence *seq_in, size_t offset,
			const ParseMode &mode_in) :
			seq(seq_in), seqOffset(offset), mode(mode_in) {}
			
			ComponentSequence *seq; //!< current sequence
			size_t seqOffset; //!< offset seq was entered, for error reporting
			ParseMode mode; //!< current mode flags
		};
		
	} // namespace
	
	static
	unsigned parseAsDecimal(unsigned oct) {
		// The input was parsed as octal, but should have been parsed as decimal.
		// Deconstruct the octal number and reconstruct into decimal
		unsigned ret = 0;
		unsigned multiplier = 1;
		while (oct) {
			ret += (oct & 0x7) * multiplier;
			oct >>= 3;
			multiplier *= 10;
		}
		return ret;
	}
	
	/** \brief Maximum value for a positive integer. We use INT_MAX, as that's what
	* PCRE uses. */
	static constexpr u32 MAX_NUMBER = INT_MAX;
	
	static
	void pushDec(u32 *acc, char raw_digit) {
		assert(raw_digit >= '0' && raw_digit <= '9');
		u32 digit_val = raw_digit - '0';
		
		// Ensure that we don't overflow.
		u64a val = ((u64a)*acc * 10) + digit_val;
		if (val > MAX_NUMBER) {
			throw LocatedParseError("Number is too big");
		}
		
		*acc = verify_u32(val);
	}
	
	static
	void pushOct(u32 *acc, char raw_digit) {
		assert(raw_digit >= '0' && raw_digit <= '7');
		u32 digit_val = raw_digit - '0';
		
		// Ensure that we don't overflow.
		u64a val = ((u64a)*acc * 8) + digit_val;
		if (val > MAX_NUMBER) {
			throw LocatedParseError("Number is too big");
		}
		
		*acc = verify_u32(val);
	}
	
	static
	void throwInvalidRepeat(void) {
		throw LocatedParseError("Invalid repeat");
	}
	
	static
	void throwInvalidUtf8(void) {
		throw ParseError("Expression is not valid UTF-8.");
	}
	
	/**
	* Adds the given child component to the parent sequence, returning a pointer
	* to the new (child) "current sequence".
	*/
	static
	ComponentSequence *enterSequence(ComponentSequence *parent,
	unique_ptr<ComponentSequence> child) {
		assert(parent);
		assert(child);
		
		ComponentSequence *seq = child.get();
		parent->addComponent(std::move(child));
		return seq;
	}
	
	static
	void addLiteral(ComponentSequence *currentSeq, char c, const ParseMode &mode) {
		if (mode.utf8 && mode.caseless) {
			/* leverage ComponentClass to generate the vertices */
			auto cc = getComponentClass(mode);
			assert(cc);
			cc->add(c);
			cc->finalize();
			currentSeq->addComponent(std::move(cc));
		} else {
			currentSeq->addComponent(getLiteralComponentClass(c, mode.caseless));
		}
	}
	
	static
	void addEscaped(ComponentSequence *currentSeq, unichar accum,
	const ParseMode &mode, const char *err_msg) {
		if (mode.utf8) {
			/* leverage ComponentClass to generate the vertices */
			auto cc = getComponentClass(mode);
			assert(cc);
			cc->add(accum);
			cc->finalize();
			currentSeq->addComponent(std::move(cc));
		} else {
			if (accum > 255) {
				throw LocatedParseError(err_msg);
			}
			addLiteral(currentSeq, (char)accum, mode);
		}
	}
	
	static
	void addEscapedOctal(ComponentSequence *currentSeq, unichar accum,
	const ParseMode &mode) {
		addEscaped(currentSeq, accum, mode, "Octal value is greater than \\377");
	}
	
	static
	void addEscapedHex(ComponentSequence *currentSeq, unichar accum,
	const ParseMode &mode) {
		addEscaped(currentSeq, accum, mode,
		"Hexadecimal value is greater than \\xFF");
	}
	
#define SLASH_C_ERROR "\\c must be followed by an ASCII character"
	
	static
	u8 decodeCtrl(char raw) {
		if (raw & 0x80) {
			throw LocatedParseError(SLASH_C_ERROR);
		}
		return mytoupper(raw) ^ 0x40;
	}
	
	static
	unichar readUtf8CodePoint2c(const char *s) {
		auto *ts = (const u8 *)s;
		assert(ts[0] >= 0xc0 && ts[0] < 0xe0);
		assert(ts[1] >= 0x80 && ts[1] < 0xc0);
		unichar val = ts[0] & 0x1f;
		val <<= 6;
		val |= ts[1] & 0x3f;
		DEBUG_PRINTF("utf8 %02hhx %02hhx ->\\x{%x}\n", ts[0],
		ts[1], val);
		return val;
	}
	
	static
	unichar readUtf8CodePoint3c(const char *s) {
		auto *ts = (const u8 *)s;
		assert(ts[0] >= 0xe0 && ts[0] < 0xf0);
		assert(ts[1] >= 0x80 && ts[1] < 0xc0);
		assert(ts[2] >= 0x80 && ts[2] < 0xc0);
		unichar val = ts[0] & 0x0f;
		val <<= 6;
		val |= ts[1] & 0x3f;
		val <<= 6;
		val |= ts[2] & 0x3f;
		DEBUG_PRINTF("utf8 %02hhx %02hhx %02hhx ->\\x{%x}\n", ts[0],
		ts[1], ts[2], val);
		return val;
	}
	
	static
	unichar readUtf8CodePoint4c(const char *s) {
		auto *ts = (const u8 *)s;
		assert(ts[0] >= 0xf0 && ts[0] < 0xf8);
		assert(ts[1] >= 0x80 && ts[1] < 0xc0);
		assert(ts[2] >= 0x80 && ts[2] < 0xc0);
		assert(ts[3] >= 0x80 && ts[3] < 0xc0);
		unichar val = ts[0] & 0x07;
		val <<= 6;
		val |= ts[1] & 0x3f;
		val <<= 6;
		val |= ts[2] & 0x3f;
		val <<= 6;
		val |= ts[3] & 0x3f;
		DEBUG_PRINTF("utf8 %02hhx %02hhx %02hhx %02hhx ->\\x{%x}\n", ts[0],
		ts[1], ts[2], ts[3], val);
		return val;
	}
	
	
#line 1910 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
	
	
	
#line 280 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
	static const short _regex_actions[] = {
		0, 1, 0, 1, 1, 1, 2, 1,
		3, 1, 4, 1, 7, 1, 8, 1,
		9, 1, 10, 1, 11, 1, 12, 1,
		13, 1, 15, 1, 16, 1, 17, 1,
		18, 1, 19, 1, 20, 1, 21, 1,
		22, 1, 23, 1, 24, 1, 25, 1,
		26, 1, 27, 1, 28, 1, 29, 1,
		30, 1, 31, 1, 32, 1, 33, 1,
		34, 1, 35, 1, 36, 1, 37, 1,
		38, 1, 39, 1, 40, 1, 41, 1,
		42, 1, 43, 1, 44, 1, 45, 1,
		46, 1, 47, 1, 48, 1, 49, 1,
		50, 1, 51, 1, 52, 1, 53, 1,
		54, 1, 55, 1, 56, 1, 57, 1,
		58, 1, 59, 1, 60, 1, 61, 1,
		62, 1, 63, 1, 64, 1, 65, 1,
		66, 1, 67, 1, 68, 1, 69, 1,
		70, 1, 71, 1, 72, 1, 73, 1,
		74, 1, 75, 1, 76, 1, 77, 1,
		78, 1, 79, 1, 80, 1, 81, 1,
		82, 1, 83, 1, 84, 1, 85, 1,
		86, 1, 87, 1, 88, 1, 89, 1,
		90, 1, 91, 1, 92, 1, 93, 1,
		94, 1, 95, 1, 96, 1, 97, 1,
		98, 1, 99, 1, 100, 1, 101, 1,
		102, 1, 103, 1, 104, 1, 105, 1,
		106, 1, 107, 1, 108, 1, 109, 1,
		110, 1, 111, 1, 112, 1, 113, 1,
		114, 1, 115, 1, 116, 1, 117, 1,
		118, 1, 119, 1, 120, 1, 121, 1,
		122, 1, 123, 1, 124, 1, 125, 1,
		126, 1, 127, 1, 128, 1, 129, 1,
		130, 1, 131, 1, 132, 1, 133, 1,
		134, 1, 135, 1, 136, 1, 137, 1,
		138, 1, 139, 1, 140, 1, 141, 1,
		142, 1, 143, 1, 144, 1, 145, 1,
		146, 1, 147, 1, 148, 1, 149, 1,
		150, 1, 151, 1, 152, 1, 153, 1,
		154, 1, 155, 1, 156, 1, 157, 1,
		158, 1, 159, 1, 160, 1, 161, 1,
		162, 1, 163, 1, 164, 1, 165, 1,
		166, 1, 167, 1, 168, 1, 169, 1,
		170, 1, 171, 1, 172, 1, 173, 1,
		174, 1, 175, 1, 176, 1, 177, 1,
		178, 1, 179, 1, 180, 1, 181, 1,
		182, 1, 183, 1, 184, 1, 185, 1,
		186, 1, 187, 1, 188, 1, 189, 1,
		190, 1, 191, 1, 192, 1, 193, 1,
		194, 1, 195, 1, 196, 1, 197, 1,
		198, 1, 199, 1, 200, 1, 201, 1,
		202, 1, 203, 1, 204, 1, 205, 1,
		206, 1, 207, 1, 208, 1, 209, 1,
		210, 1, 211, 1, 212, 1, 213, 1,
		214, 1, 215, 1, 216, 1, 217, 1,
		218, 1, 219, 1, 220, 1, 221, 1,
		222, 1, 223, 1, 224, 1, 225, 1,
		226, 1, 227, 1, 228, 1, 229, 1,
		230, 1, 231, 1, 232, 1, 233, 1,
		234, 1, 235, 1, 236, 1, 237, 1,
		240, 1, 242, 1, 243, 1, 244, 1,
		245, 1, 246, 1, 247, 1, 248, 1,
		249, 1, 250, 1, 251, 1, 252, 1,
		253, 1, 254, 1, 255, 1, 256, 1,
		257, 1, 258, 1, 259, 1, 260, 1,
		261, 1, 262, 1, 263, 1, 264, 1,
		265, 1, 266, 1, 267, 1, 268, 1,
		269, 1, 270, 1, 271, 1, 272, 1,
		273, 1, 274, 1, 275, 1, 276, 1,
		277, 1, 278, 1, 279, 1, 280, 1,
		281, 1, 282, 1, 283, 1, 284, 1,
		285, 1, 286, 1, 287, 1, 288, 1,
		289, 1, 290, 1, 291, 1, 292, 1,
		293, 1, 294, 1, 295, 1, 296, 1,
		297, 1, 298, 1, 299, 1, 300, 1,
		301, 1, 302, 1, 303, 1, 307, 1,
		308, 1, 309, 1, 310, 1, 311, 1,
		312, 1, 313, 1, 314, 1, 315, 1,
		316, 1, 317, 1, 318, 1, 319, 1,
		320, 1, 321, 1, 322, 1, 323, 1,
		324, 1, 325, 1, 326, 1, 327, 1,
		328, 1, 329, 1, 330, 1, 331, 1,
		332, 1, 333, 1, 334, 1, 335, 1,
		336, 1, 337, 1, 338, 1, 342, 1,
		343, 1, 344, 1, 345, 1, 346, 1,
		347, 1, 348, 1, 349, 1, 350, 1,
		352, 1, 353, 1, 354, 1, 355, 1,
		356, 1, 357, 1, 358, 1, 359, 1,
		360, 1, 361, 1, 362, 1, 363, 1,
		364, 1, 365, 1, 366, 1, 367, 1,
		368, 1, 369, 1, 370, 1, 371, 1,
		372, 1, 373, 1, 374, 1, 375, 1,
		376, 1, 377, 1, 378, 1, 379, 1,
		380, 1, 381, 1, 382, 1, 383, 1,
		384, 1, 385, 1, 386, 1, 387, 1,
		388, 1, 389, 1, 390, 1, 391, 1,
		392, 1, 393, 1, 394, 1, 395, 1,
		396, 1, 397, 1, 398, 1, 399, 1,
		400, 1, 401, 1, 402, 1, 403, 1,
		404, 1, 405, 1, 406, 1, 407, 1,
		408, 1, 409, 1, 410, 1, 411, 1,
		412, 1, 413, 1, 414, 1, 415, 1,
		416, 1, 417, 1, 418, 1, 419, 1,
		420, 1, 421, 1, 422, 1, 423, 1,
		424, 1, 425, 1, 426, 1, 427, 1,
		428, 1, 429, 1, 430, 1, 431, 1,
		432, 1, 433, 1, 434, 1, 435, 1,
		436, 2, 3, 0, 2, 4, 5, 2,
		5, 1, 2, 9, 10, 2, 9, 238,
		2, 9, 239, 2, 9, 339, 2, 10,
		1, 2, 10, 340, 2, 10, 341, 2,
		11, 241, 2, 11, 351, 2, 12, 241,
		2, 12, 351, 2, 13, 241, 2, 13,
		351, 2, 14, 375, 2, 14, 376, 2,
		25, 0, 2, 25, 3, 2, 25, 6,
		2, 25, 14, 3, 25, 5, 306, 3,
		25, 10, 305, 3, 25, 14, 15, 4,
		25, 9, 304, 10, 0
	};
	
	static const signed char _regex_to_state_actions[] = {
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 41,
		0, 41, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 41, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 41, 0, 0, 41, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 41, 41, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 41, 0, 41, 0,
		0, 0, 0, 41, 0, 0, 0, 0,
		41, 41, 0
	};
	
	static const int regex_start = 746;
	static const int regex_error = 0;
	
	static const int regex_en_readVerb = 787;
	static const int regex_en_readUCP = 790;
	static const int regex_en_readBracedUCP = 559;
	static const int regex_en_readUCPSingle = 818;
	static const int regex_en_charClassGuts = 819;
	static const int regex_en_readClass = 836;
	static const int regex_en_readQuotedLiteral = 838;
	static const int regex_en_readQuotedClass = 843;
	static const int regex_en_readComment = 848;
	static const int regex_en_readNewlineTerminatedComment = 849;
	static const int regex_en_main = 746;
	
	
#line 1912 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
	
	
	/** \brief Main parser call, returns root Component or nullptr. */
	unique_ptr<Component> parse(const char *ptr, ParseMode &globalMode) {
		assert(ptr);
		
		const char *p = ptr;
		const char *pe = ptr + strlen(ptr);
		
		// First, read the control verbs, set any global mode flags and move the
		// ptr forward.
		p = read_control_verbs(p, pe, 0, globalMode);
		
		const char *eof = pe;
		int cs;
		UNUSED int act;
		int top;
		vector<int> stack;
		const char *ts, *te;
		unichar accumulator = 0;
		unichar octAccumulator = 0; /* required as we are also accumulating for
		* back ref when looking for octals */
		unsigned repeatN = 0;
		unsigned repeatM = 0;
		string label;
		
		ParseMode mode = globalMode;
		ParseMode newMode;
		
		bool negated = false;
		bool inComment = false;
		
		// Stack of sequences and flags used to store state when we enter
		// sub-sequences.
		vector<ExprState> sequences;
		
		// Index of the next capturing group. Note that zero is reserved for the
		// root sequence.
		unsigned groupIndex = 1;
		
		// Set storing group names that are currently in use.
		flat_set<string> groupNames;
		
		// Root sequence.
		unique_ptr<ComponentSequence> rootSeq = std::make_unique<ComponentSequence>();
		rootSeq->setCaptureIndex(0);
		
		// Current sequence being appended to
		ComponentSequence *currentSeq = rootSeq.get();
		
		// The current character class being appended to. This is used as the
		// accumulator for both character class and UCP properties.
		unique_ptr<ComponentClass> currentCls;
		
		// True if the machine is currently inside a character class, i.e. square
		// brackets [..].
		bool inCharClass = false;
		
		// True if the machine is inside a character class but it has not processed
		// any "real" elements yet, i.e. it's still processing meta-characters like
		// '^'.
		bool inCharClassEarly = false;
		
		// Location at which the current character class began.
		const char *currentClsBegin = p;
		
		// We throw exceptions on various parsing failures beyond this point: we
		// use a try/catch block here to clean up our allocated memory before we
		// re-throw the exception to the caller.
		try {
			// Embed the Ragel machine here
			
#line 600 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			{
				cs = (int)regex_start;
				top = 0;
				ts = 0;
				te = 0;
				act = 0;
			}
			
#line 1983 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
			
			
#line 612 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			{
				int _ck;
				const short * _acts;
				unsigned int _nacts;
				
				_resume: {}
				if ( p == pe && p != eof )
					goto _out;
				switch ( cs ) {
					case 746:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 627 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr882;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 0u: {
								goto _ctr883;
							}
							case 32u: {
								goto _ctr885;
							}
							case 35u: {
								_ck = 0;
								if ( ( mode.ignore_space )
#line 643 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr887;
								} else {
									goto _ctr886;
								}
								
							}
							case 36u: {
								goto _ctr888;
							}
							case 40u: {
								goto _ctr889;
							}
							case 41u: {
								goto _ctr890;
							}
							case 42u: {
								goto _ctr891;
							}
							case 43u: {
								goto _ctr892;
							}
							case 46u: {
								goto _ctr893;
							}
							case 63u: {
								goto _ctr894;
							}
							case 91u: {
								goto _ctr895;
							}
							case 92u: {
								goto _ctr896;
							}
							case 94u: {
								goto _ctr897;
							}
							case 123u: {
								goto _ctr898;
							}
							case 124u: {
								goto _ctr899;
							}
						}
						if ( ( (*( p))) < 192u ) {
							if ( ( (*( p))) > 13u ) {
								if ( 128u <= ( (*( p))) ) {
									_ck = 0;
									if ( ( mode.utf8 )
#line 695 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr900;
									} else {
										goto _ctr886;
									}
									
								}
							} else if ( ( (*( p))) >= 9u ) {
								goto _ctr885;
							}
						} else if ( ( (*( p))) > 223u ) {
							if ( ( (*( p))) < 240u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 713 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr902;
									} else {
										goto _ctr886;
									}
									
								}
							} else if ( ( (*( p))) > 247u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 727 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr900;
									} else {
										goto _ctr886;
									}
									
								}
							} else {
								_ck = 0;
								if ( ( mode.utf8 )
#line 740 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr903;
								} else {
									goto _ctr886;
								}
								
							}
						} else {
							_ck = 0;
							if ( ( mode.utf8 )
#line 753 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr901;
							} else {
								goto _ctr886;
							}
							
						}
						goto _ctr884;
					}
					case 747:
					if ( p == eof ) {
						goto _ctr904;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 42u: {
								goto _ctr905;
							}
							case 63u: {
								goto _ctr906;
							}
						}
						goto _ctr904;
					}
					case 1:
					if ( p == eof ) {
						goto _ctr1;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr1;
						}
						goto _ctr2;
					}
					case 2:
					if ( p == eof ) {
						goto _ctr1;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 33u: {
								goto _ctr4;
							}
							case 35u: {
								goto _ctr5;
							}
							case 38u: {
								goto _ctr6;
							}
							case 39u: {
								goto _ctr7;
							}
							case 40u: {
								goto _ctr8;
							}
							case 41u: {
								goto _ctr9;
							}
							case 43u: {
								goto _ctr10;
							}
							case 45u: {
								goto _ctr11;
							}
							case 58u: {
								goto _ctr13;
							}
							case 60u: {
								goto _ctr14;
							}
							case 61u: {
								goto _ctr15;
							}
							case 62u: {
								goto _ctr16;
							}
							case 63u: {
								goto _ctr17;
							}
							case 67u: {
								goto _ctr18;
							}
							case 80u: {
								goto _ctr19;
							}
							case 105u: {
								goto _ctr20;
							}
							case 109u: {
								goto _ctr20;
							}
							case 115u: {
								goto _ctr20;
							}
							case 120u: {
								goto _ctr20;
							}
							case 123u: {
								goto _ctr21;
							}
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr12;
						}
						goto _ctr3;
					}
					case 748:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr24;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr24;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr24;
							}
						} else {
							goto _ctr24;
						}
						goto _ctr907;
					}
					case 3:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr23;
							}
							case 95u: {
								goto _ctr24;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr24;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr24;
							}
						} else {
							goto _ctr24;
						}
						goto _ctr22;
					}
					case 749:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr26;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr26;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr26;
							}
						} else {
							goto _ctr26;
						}
						goto _ctr907;
					}
					case 4:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr25;
							}
							case 95u: {
								goto _ctr26;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr26;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr26;
							}
						} else {
							goto _ctr26;
						}
						goto _ctr22;
					}
					case 750:
					if ( p == eof ) {
						goto _ctr908;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr909;
							}
							case 48u: {
								goto _ctr31;
							}
							case 60u: {
								goto _ctr912;
							}
							case 63u: {
								goto _ctr913;
							}
							case 82u: {
								goto _ctr914;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 56u ) {
							if ( 49u <= ( (*( p))) ) {
								goto _ctr910;
							}
						} else if ( ( (*( p))) > 57u ) {
							if ( ( (*( p))) > 90u ) {
								if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
									goto _ctr31;
								}
							} else if ( ( (*( p))) >= 65u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr911;
						}
						goto _ctr908;
					}
					case 5:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr28;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr28;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr28;
							}
						} else {
							goto _ctr28;
						}
						goto _ctr27;
					}
					case 6:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr29;
							}
							case 95u: {
								goto _ctr28;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr28;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr28;
							}
						} else {
							goto _ctr28;
						}
						goto _ctr27;
					}
					case 7:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr30;
						}
						goto _ctr27;
					}
					case 8:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr30;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr31;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr31;
						}
						goto _ctr27;
					}
					case 9:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr32;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr33;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr31;
						}
						goto _ctr27;
					}
					case 10:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr30;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr33;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr31;
						}
						goto _ctr27;
					}
					case 11:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr34;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr34;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr34;
							}
						} else {
							goto _ctr34;
						}
						goto _ctr27;
					}
					case 12:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 62u: {
								goto _ctr29;
							}
							case 95u: {
								goto _ctr34;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr34;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr34;
							}
						} else {
							goto _ctr34;
						}
						goto _ctr27;
					}
					case 13:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 33u: {
								goto _ctr35;
							}
							case 60u: {
								goto _ctr36;
							}
							case 61u: {
								goto _ctr37;
							}
						}
						goto _ctr27;
					}
					case 14:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 33u: {
								goto _ctr38;
							}
							case 61u: {
								goto _ctr39;
							}
						}
						goto _ctr27;
					}
					case 15:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 38u: {
								goto _ctr40;
							}
							case 41u: {
								goto _ctr41;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr42;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr31;
						}
						goto _ctr27;
					}
					case 16:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr43;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr43;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr43;
							}
						} else {
							goto _ctr43;
						}
						goto _ctr27;
					}
					case 17:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr41;
							}
							case 95u: {
								goto _ctr43;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr43;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr43;
							}
						} else {
							goto _ctr43;
						}
						goto _ctr27;
					}
					case 18:
					if ( p == eof ) {
						goto _ctr27;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr41;
							}
							case 95u: {
								goto _ctr31;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr42;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr31;
							}
						} else {
							goto _ctr31;
						}
						goto _ctr27;
					}
					case 751:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr45;
						}
						goto _ctr907;
					}
					case 19:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr44;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr45;
						}
						goto _ctr22;
					}
					case 752:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 105u: {
								goto _ctr48;
							}
							case 109u: {
								goto _ctr48;
							}
							case 115u: {
								goto _ctr48;
							}
							case 120u: {
								goto _ctr48;
							}
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr45;
						}
						goto _ctr907;
					}
					case 20:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr46;
							}
							case 58u: {
								goto _ctr47;
							}
							case 105u: {
								goto _ctr48;
							}
							case 109u: {
								goto _ctr48;
							}
							case 115u: {
								goto _ctr48;
							}
							case 120u: {
								goto _ctr48;
							}
						}
						goto _ctr22;
					}
					case 753:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr44;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr45;
						}
						goto _ctr907;
					}
					case 754:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 33u: {
								goto _ctr915;
							}
							case 61u: {
								goto _ctr916;
							}
							case 95u: {
								goto _ctr49;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr49;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr49;
							}
						} else {
							goto _ctr49;
						}
						goto _ctr907;
					}
					case 21:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 62u: {
								goto _ctr25;
							}
							case 95u: {
								goto _ctr49;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr49;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr49;
							}
						} else {
							goto _ctr49;
						}
						goto _ctr22;
					}
					case 755:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr917;
						}
						goto _ctr907;
					}
					case 756:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr50;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr51;
						}
						goto _ctr907;
					}
					case 22:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr50;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr51;
						}
						goto _ctr22;
					}
					case 757:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 60u: {
								goto _ctr918;
							}
							case 61u: {
								goto _ctr919;
							}
							case 62u: {
								goto _ctr920;
							}
						}
						goto _ctr907;
					}
					case 23:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr49;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr49;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr49;
							}
						} else {
							goto _ctr49;
						}
						goto _ctr22;
					}
					case 24:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr52;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr52;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr52;
							}
						} else {
							goto _ctr52;
						}
						goto _ctr22;
					}
					case 25:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr53;
							}
							case 95u: {
								goto _ctr52;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr52;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr52;
							}
						} else {
							goto _ctr52;
						}
						goto _ctr22;
					}
					case 26:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr24;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr24;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr24;
							}
						} else {
							goto _ctr24;
						}
						goto _ctr22;
					}
					case 758:
					if ( p == eof ) {
						goto _ctr907;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr46;
							}
							case 45u: {
								goto _ctr54;
							}
							case 58u: {
								goto _ctr47;
							}
							case 105u: {
								goto _ctr55;
							}
							case 109u: {
								goto _ctr55;
							}
							case 115u: {
								goto _ctr55;
							}
							case 120u: {
								goto _ctr55;
							}
						}
						goto _ctr907;
					}
					case 27:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 105u: {
								goto _ctr48;
							}
							case 109u: {
								goto _ctr48;
							}
							case 115u: {
								goto _ctr48;
							}
							case 120u: {
								goto _ctr48;
							}
						}
						goto _ctr22;
					}
					case 28:
					if ( p == eof ) {
						goto _ctr22;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr46;
							}
							case 45u: {
								goto _ctr54;
							}
							case 58u: {
								goto _ctr47;
							}
							case 105u: {
								goto _ctr55;
							}
							case 109u: {
								goto _ctr55;
							}
							case 115u: {
								goto _ctr55;
							}
							case 120u: {
								goto _ctr55;
							}
						}
						goto _ctr22;
					}
					case 759:
					if ( p == eof ) {
						goto _ctr921;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 43u: {
								goto _ctr922;
							}
							case 63u: {
								goto _ctr923;
							}
						}
						goto _ctr921;
					}
					case 760:
					if ( p == eof ) {
						goto _ctr924;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 43u: {
								goto _ctr925;
							}
							case 63u: {
								goto _ctr926;
							}
						}
						goto _ctr924;
					}
					case 761:
					if ( p == eof ) {
						goto _ctr927;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 43u: {
								goto _ctr928;
							}
							case 63u: {
								goto _ctr929;
							}
						}
						goto _ctr927;
					}
					case 762:
					if ( p == eof ) {
						goto _ctr930;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr57;
							}
							case 58u: {
								goto _ctr61;
							}
							case 61u: {
								goto _ctr65;
							}
						}
						goto _ctr930;
					}
					case 29:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr58;
							}
							case 92u: {
								goto _ctr59;
							}
							case 93u: {
								goto _ctr56;
							}
						}
						goto _ctr57;
					}
					case 30:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr58;
							}
							case 92u: {
								goto _ctr59;
							}
							case 93u: {
								goto _ctr60;
							}
						}
						goto _ctr57;
					}
					case 31:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr58;
							}
							case 92u: {
								goto _ctr59;
							}
						}
						goto _ctr57;
					}
					case 32:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr62;
							}
							case 92u: {
								goto _ctr63;
							}
							case 93u: {
								goto _ctr56;
							}
						}
						goto _ctr61;
					}
					case 33:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr62;
							}
							case 92u: {
								goto _ctr63;
							}
							case 93u: {
								goto _ctr64;
							}
						}
						goto _ctr61;
					}
					case 34:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr62;
							}
							case 92u: {
								goto _ctr63;
							}
						}
						goto _ctr61;
					}
					case 35:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr66;
							}
							case 92u: {
								goto _ctr67;
							}
							case 93u: {
								goto _ctr56;
							}
						}
						goto _ctr65;
					}
					case 36:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr66;
							}
							case 92u: {
								goto _ctr67;
							}
							case 93u: {
								goto _ctr60;
							}
						}
						goto _ctr65;
					}
					case 37:
					if ( p == eof ) {
						goto _ctr56;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr66;
							}
							case 92u: {
								goto _ctr67;
							}
						}
						goto _ctr65;
					}
					case 763:
					if ( p == eof ) {
						goto _ctr931;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 48u: {
								goto _ctr933;
							}
							case 65u: {
								goto _ctr936;
							}
							case 66u: {
								goto _ctr937;
							}
							case 67u: {
								goto _ctr938;
							}
							case 68u: {
								goto _ctr939;
							}
							case 69u: {
								goto _ctr940;
							}
							case 71u: {
								goto _ctr941;
							}
							case 72u: {
								goto _ctr942;
							}
							case 75u: {
								goto _ctr943;
							}
							case 76u: {
								goto _ctr944;
							}
							case 78u: {
								goto _ctr944;
							}
							case 80u: {
								goto _ctr945;
							}
							case 81u: {
								goto _ctr946;
							}
							case 82u: {
								goto _ctr947;
							}
							case 83u: {
								goto _ctr948;
							}
							case 85u: {
								goto _ctr944;
							}
							case 86u: {
								goto _ctr949;
							}
							case 87u: {
								goto _ctr950;
							}
							case 88u: {
								goto _ctr951;
							}
							case 90u: {
								goto _ctr952;
							}
							case 97u: {
								goto _ctr953;
							}
							case 98u: {
								goto _ctr954;
							}
							case 99u: {
								goto _ctr955;
							}
							case 100u: {
								goto _ctr956;
							}
							case 101u: {
								goto _ctr957;
							}
							case 102u: {
								goto _ctr958;
							}
							case 103u: {
								goto _ctr959;
							}
							case 104u: {
								goto _ctr960;
							}
							case 107u: {
								goto _ctr961;
							}
							case 108u: {
								goto _ctr944;
							}
							case 110u: {
								goto _ctr962;
							}
							case 111u: {
								goto _ctr963;
							}
							case 112u: {
								goto _ctr964;
							}
							case 114u: {
								goto _ctr965;
							}
							case 115u: {
								goto _ctr966;
							}
							case 116u: {
								goto _ctr967;
							}
							case 117u: {
								goto _ctr944;
							}
							case 118u: {
								goto _ctr968;
							}
							case 119u: {
								goto _ctr969;
							}
							case 120u: {
								goto _ctr970;
							}
							case 122u: {
								goto _ctr971;
							}
						}
						if ( ( (*( p))) > 55u ) {
							if ( ( (*( p))) <= 57u ) {
								goto _ctr935;
							}
						} else if ( ( (*( p))) >= 49u ) {
							goto _ctr934;
						}
						goto _ctr932;
					}
					case 764:
					if ( p == eof ) {
						goto _ctr972;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr973;
						}
						goto _ctr972;
					}
					case 765:
					if ( p == eof ) {
						goto _ctr972;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr974;
						}
						goto _ctr972;
					}
					case 766:
					if ( p == eof ) {
						goto _ctr975;	goto _again;
					}
					else {
						if ( ( (*( p))) > 55u ) {
							if ( ( (*( p))) <= 57u ) {
								goto _ctr977;
							}
						} else if ( ( (*( p))) >= 48u ) {
							goto _ctr976;
						}
						goto _ctr975;
					}
					case 767:
					if ( p == eof ) {
						goto _ctr978;	goto _again;
					}
					else {
						if ( ( (*( p))) > 55u ) {
							if ( ( (*( p))) <= 57u ) {
								goto _ctr977;
							}
						} else if ( ( (*( p))) >= 48u ) {
							goto _ctr979;
						}
						goto _ctr978;
					}
					case 768:
					if ( p == eof ) {
						goto _ctr980;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr977;
						}
						goto _ctr980;
					}
					case 769:
					if ( p == eof ) {
						goto _ctr981;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr983;
						}
						goto _ctr982;
					}
					case 770:
					if ( p == eof ) {
						goto _ctr984;	goto _again;
					}
					else {
						goto _ctr985;
					}
					case 771:
					if ( p == eof ) {
						goto _ctr986;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr69;
							}
							case 45u: {
								goto _ctr987;
							}
							case 60u: {
								goto _ctr72;
							}
							case 123u: {
								goto _ctr989;
							}
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr988;
						}
						goto _ctr986;
					}
					case 38:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( ( (*( p))) == 39u ) {
							goto _ctr70;
						}
						goto _ctr69;
					}
					case 39:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr71;
						}
						goto _ctr68;
					}
					case 772:
					if ( p == eof ) {
						goto _ctr990;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr991;
						}
						goto _ctr990;
					}
					case 773:
					if ( p == eof ) {
						goto _ctr990;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr992;
						}
						goto _ctr990;
					}
					case 774:
					if ( p == eof ) {
						goto _ctr993;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr994;
						}
						goto _ctr993;
					}
					case 775:
					if ( p == eof ) {
						goto _ctr993;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr995;
						}
						goto _ctr993;
					}
					case 40:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( ( (*( p))) == 62u ) {
							goto _ctr70;
						}
						goto _ctr72;
					}
					case 41:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 45u: {
								goto _ctr73;
							}
							case 95u: {
								goto _ctr75;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr74;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr75;
							}
						} else {
							goto _ctr75;
						}
						goto _ctr68;
					}
					case 42:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr76;
						}
						goto _ctr68;
					}
					case 43:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr78;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr77;
						}
						goto _ctr68;
					}
					case 44:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr78;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr79;
						}
						goto _ctr68;
					}
					case 45:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr78;
						}
						goto _ctr68;
					}
					case 46:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr75;
							}
							case 125u: {
								goto _ctr81;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr80;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr75;
							}
						} else {
							goto _ctr75;
						}
						goto _ctr68;
					}
					case 47:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr75;
							}
							case 125u: {
								goto _ctr81;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr82;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr75;
							}
						} else {
							goto _ctr75;
						}
						goto _ctr68;
					}
					case 48:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr75;
							}
							case 125u: {
								goto _ctr81;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr75;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr75;
							}
						} else {
							goto _ctr75;
						}
						goto _ctr68;
					}
					case 49:
					if ( p == eof ) {
						goto _ctr68;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr75;
							}
							case 125u: {
								goto _ctr83;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr75;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr75;
							}
						} else {
							goto _ctr75;
						}
						goto _ctr68;
					}
					case 776:
					if ( p == eof ) {
						goto _ctr996;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr997;
							}
							case 60u: {
								goto _ctr998;
							}
							case 123u: {
								goto _ctr999;
							}
						}
						goto _ctr996;
					}
					case 50:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr85;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr85;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr85;
							}
						} else {
							goto _ctr85;
						}
						goto _ctr84;
					}
					case 51:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 39u: {
								goto _ctr86;
							}
							case 95u: {
								goto _ctr85;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr85;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr85;
							}
						} else {
							goto _ctr85;
						}
						goto _ctr84;
					}
					case 52:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr87;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr87;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr87;
							}
						} else {
							goto _ctr87;
						}
						goto _ctr84;
					}
					case 53:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 62u: {
								goto _ctr88;
							}
							case 95u: {
								goto _ctr87;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr87;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr87;
							}
						} else {
							goto _ctr87;
						}
						goto _ctr84;
					}
					case 54:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr89;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr89;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr89;
							}
						} else {
							goto _ctr89;
						}
						goto _ctr84;
					}
					case 55:
					if ( p == eof ) {
						goto _ctr84;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr89;
							}
							case 125u: {
								goto _ctr90;
							}
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr89;
							}
						} else if ( ( (*( p))) > 90u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 122u ) {
								goto _ctr89;
							}
						} else {
							goto _ctr89;
						}
						goto _ctr84;
					}
					case 777:
					if ( p == eof ) {
						goto _ctr1000;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1001;
						}
						goto _ctr1000;
					}
					case 56:
					if ( p == eof ) {
						goto _ctr91;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr92;
						}
						goto _ctr91;
					}
					case 57:
					if ( p == eof ) {
						goto _ctr91;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr93;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr92;
						}
						goto _ctr91;
					}
					case 778:
					if ( p == eof ) {
						goto _ctr1002;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1004;
						}
						goto _ctr1003;
					}
					case 779:
					if ( p == eof ) {
						goto _ctr1005;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1009;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr1006;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr1008;
							}
						} else {
							goto _ctr1007;
						}
						goto _ctr1005;
					}
					case 780:
					if ( p == eof ) {
						goto _ctr1005;	goto _again;
					}
					else {
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr1010;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr1012;
							}
						} else {
							goto _ctr1011;
						}
						goto _ctr1005;
					}
					case 781:
					if ( p == eof ) {
						goto _ctr1013;	goto _again;
					}
					else {
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr95;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr95;
							}
						} else {
							goto _ctr95;
						}
						goto _ctr1013;
					}
					case 58:
					if ( p == eof ) {
						goto _ctr94;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr96;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr95;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr95;
							}
						} else {
							goto _ctr95;
						}
						goto _ctr94;
					}
					case 782:
					if ( p == eof ) {
						goto _ctr1014;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr99;
						}
						goto _ctr1014;
					}
					case 59:
					if ( p == eof ) {
						goto _ctr97;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 44u: {
								goto _ctr98;
							}
							case 125u: {
								goto _ctr100;
							}
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr99;
						}
						goto _ctr97;
					}
					case 60:
					if ( p == eof ) {
						goto _ctr97;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr102;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr101;
						}
						goto _ctr97;
					}
					case 61:
					if ( p == eof ) {
						goto _ctr97;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr103;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
							goto _ctr101;
						}
						goto _ctr97;
					}
					case 783:
					if ( p == eof ) {
						goto _ctr1015;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 43u: {
								goto _ctr1016;
							}
							case 63u: {
								goto _ctr1017;
							}
						}
						goto _ctr1015;
					}
					case 784:
					if ( p == eof ) {
						goto _ctr1018;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2738 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1020;
							} else {
								goto _ctr1019;
							}
							
						}
						goto _ctr1018;
					}
					case 785:
					if ( p == eof ) {
						goto _ctr1018;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2758 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1021;
							} else {
								goto _ctr1019;
							}
							
						}
						goto _ctr1018;
					}
					case 62:
					if ( p == eof ) {
						goto _ctr104;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2778 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr106;
							} else {
								goto _ctr105;
							}
							
						}
						goto _ctr104;
					}
					case 786:
					if ( p == eof ) {
						goto _ctr1018;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2798 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1022;
							} else {
								goto _ctr1019;
							}
							
						}
						goto _ctr1018;
					}
					case 63:
					if ( p == eof ) {
						goto _ctr104;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2818 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr107;
							} else {
								goto _ctr105;
							}
							
						}
						goto _ctr104;
					}
					case 64:
					if ( p == eof ) {
						goto _ctr104;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 2838 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr108;
							} else {
								goto _ctr105;
							}
							
						}
						goto _ctr104;
					}
					case 559:
					if ( p == eof ) {
						goto _ctr651;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr653;
						}
						goto _ctr652;
					}
					case 0:
					if ( p == eof ) {
						goto _ctr0;	goto _again;
					}
					else {
						goto _again;
					}
					case 560:
					if ( p == eof ) {
						goto _ctr654;	goto _again;
					}
					else {
						if ( ( (*( p))) == 94u ) {
							goto _ctr656;
						}
						goto _ctr655;
					}
					case 561:
					if ( p == eof ) {
						goto _ctr657;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr658;
						}
						goto _ctr652;
					}
					case 817:
					if ( p == eof ) {
						goto _ctr1162;	goto _again;
					}
					else {
						goto _ctr652;
					}
					case 562:
					if ( p == eof ) {
						goto _ctr659;	goto _again;
					}
					else {
						if ( ( (*( p))) == 94u ) {
							goto _ctr652;
						}
						goto _ctr655;
					}
					case 787:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 2909 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1023;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr1025;
							}
							case 85u: {
								goto _ctr1026;
							}
						}
						goto _ctr1024;
					}
					case 788:
					if ( p == eof ) {
						goto _ctr1027;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr111;
						}
						goto _ctr110;
					}
					case 65:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr111;
						}
						goto _ctr110;
					}
					case 789:
					if ( p == eof ) {
						goto _ctr1027;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr111;
							}
							case 67u: {
								goto _ctr1028;
							}
							case 84u: {
								goto _ctr1029;
							}
						}
						goto _ctr110;
					}
					case 66:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr111;
							}
							case 80u: {
								goto _ctr112;
							}
						}
						goto _ctr110;
					}
					case 67:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr113;
						}
						goto _ctr110;
					}
					case 68:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr111;
							}
							case 70u: {
								goto _ctr114;
							}
						}
						goto _ctr110;
					}
					case 69:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 41u: {
								goto _ctr115;
							}
							case 56u: {
								goto _ctr116;
							}
						}
						goto _ctr110;
					}
					case 70:
					if ( p == eof ) {
						goto _ctr109;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr117;
						}
						goto _ctr110;
					}
					case 790:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 3033 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1030;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 65u: {
								goto _ctr1032;
							}
							case 66u: {
								goto _ctr1033;
							}
							case 67u: {
								goto _ctr1034;
							}
							case 68u: {
								goto _ctr1035;
							}
							case 69u: {
								goto _ctr1036;
							}
							case 71u: {
								goto _ctr1037;
							}
							case 72u: {
								goto _ctr1038;
							}
							case 73u: {
								goto _ctr1039;
							}
							case 74u: {
								goto _ctr1040;
							}
							case 75u: {
								goto _ctr1041;
							}
							case 76u: {
								goto _ctr1042;
							}
							case 77u: {
								goto _ctr1043;
							}
							case 78u: {
								goto _ctr1044;
							}
							case 79u: {
								goto _ctr1045;
							}
							case 80u: {
								goto _ctr1046;
							}
							case 82u: {
								goto _ctr1047;
							}
							case 83u: {
								goto _ctr1048;
							}
							case 84u: {
								goto _ctr1049;
							}
							case 85u: {
								goto _ctr1050;
							}
							case 86u: {
								goto _ctr1051;
							}
							case 88u: {
								goto _ctr1052;
							}
							case 89u: {
								goto _ctr1053;
							}
							case 90u: {
								goto _ctr1054;
							}
						}
						goto _ctr1031;
					}
					case 791:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 110u: {
								goto _ctr1056;
							}
							case 114u: {
								goto _ctr1057;
							}
							case 118u: {
								goto _ctr1058;
							}
						}
						goto _ctr1055;
					}
					case 71:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr119;
						}
						goto _ctr118;
					}
					case 72:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr120;
							}
							case 109u: {
								goto _ctr121;
							}
						}
						goto _ctr118;
					}
					case 73:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 98u ) {
							goto _ctr122;
						}
						goto _ctr118;
					}
					case 74:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr123;
						}
						goto _ctr118;
					}
					case 75:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr124;
						}
						goto _ctr118;
					}
					case 76:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr125;
						}
						goto _ctr118;
					}
					case 77:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr126;
						}
						goto _ctr118;
					}
					case 78:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr127;
						}
						goto _ctr118;
					}
					case 79:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr128;
						}
						goto _ctr118;
					}
					case 80:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr129;
						}
						goto _ctr118;
					}
					case 81:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr130;
						}
						goto _ctr118;
					}
					case 82:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr131;
						}
						goto _ctr118;
					}
					case 83:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr132;
						}
						goto _ctr118;
					}
					case 84:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr133;
						}
						goto _ctr118;
					}
					case 85:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr134;
						}
						goto _ctr118;
					}
					case 792:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1059;
							}
							case 101u: {
								goto _ctr1060;
							}
							case 111u: {
								goto _ctr1061;
							}
							case 114u: {
								goto _ctr1062;
							}
							case 117u: {
								goto _ctr1063;
							}
						}
						goto _ctr1055;
					}
					case 86:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 108u: {
								goto _ctr135;
							}
							case 109u: {
								goto _ctr136;
							}
							case 116u: {
								goto _ctr137;
							}
						}
						goto _ctr118;
					}
					case 87:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr138;
						}
						goto _ctr118;
					}
					case 88:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr139;
						}
						goto _ctr118;
					}
					case 89:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr140;
						}
						goto _ctr118;
					}
					case 90:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr141;
						}
						goto _ctr118;
					}
					case 91:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr142;
						}
						goto _ctr118;
					}
					case 92:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr143;
						}
						goto _ctr118;
					}
					case 93:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr144;
						}
						goto _ctr118;
					}
					case 94:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr145;
						}
						goto _ctr118;
					}
					case 95:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr146;
						}
						goto _ctr118;
					}
					case 96:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr147;
						}
						goto _ctr118;
					}
					case 97:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr148;
						}
						goto _ctr118;
					}
					case 98:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr149;
						}
						goto _ctr118;
					}
					case 99:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr150;
						}
						goto _ctr118;
					}
					case 100:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr151;
						}
						goto _ctr118;
					}
					case 101:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr152;
						}
						goto _ctr118;
					}
					case 102:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr153;
						}
						goto _ctr118;
					}
					case 103:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr154;
						}
						goto _ctr118;
					}
					case 104:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr155;
						}
						goto _ctr118;
					}
					case 105:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 102u ) {
							goto _ctr156;
						}
						goto _ctr118;
					}
					case 106:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr157;
						}
						goto _ctr118;
					}
					case 107:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr158;
						}
						goto _ctr118;
					}
					case 108:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 104u: {
								goto _ctr159;
							}
							case 105u: {
								goto _ctr160;
							}
						}
						goto _ctr118;
					}
					case 109:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr161;
						}
						goto _ctr118;
					}
					case 110:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr162;
						}
						goto _ctr118;
					}
					case 111:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr163;
						}
						goto _ctr118;
					}
					case 112:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr164;
						}
						goto _ctr118;
					}
					case 113:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr165;
						}
						goto _ctr118;
					}
					case 114:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 103u: {
								goto _ctr166;
							}
							case 104u: {
								goto _ctr167;
							}
						}
						goto _ctr118;
					}
					case 115:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr168;
						}
						goto _ctr118;
					}
					case 116:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr169;
						}
						goto _ctr118;
					}
					case 117:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr170;
						}
						goto _ctr118;
					}
					case 118:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr171;
						}
						goto _ctr118;
					}
					case 119:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr172;
						}
						goto _ctr118;
					}
					case 120:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr173;
						}
						goto _ctr118;
					}
					case 121:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr174;
						}
						goto _ctr118;
					}
					case 793:
					if ( p == eof ) {
						goto _ctr1064;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1065;
							}
							case 99u: {
								goto _ctr1066;
							}
							case 102u: {
								goto _ctr1067;
							}
							case 104u: {
								goto _ctr1068;
							}
							case 110u: {
								goto _ctr1069;
							}
							case 111u: {
								goto _ctr1070;
							}
							case 115u: {
								goto _ctr1071;
							}
							case 117u: {
								goto _ctr1072;
							}
							case 121u: {
								goto _ctr1073;
							}
						}
						goto _ctr1064;
					}
					case 122:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 110u: {
								goto _ctr176;
							}
							case 114u: {
								goto _ctr177;
							}
						}
						goto _ctr175;
					}
					case 123:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr178;
						}
						goto _ctr175;
					}
					case 124:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr179;
						}
						goto _ctr175;
					}
					case 125:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr180;
						}
						goto _ctr175;
					}
					case 126:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr181;
						}
						goto _ctr175;
					}
					case 127:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr182;
						}
						goto _ctr175;
					}
					case 128:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr183;
						}
						goto _ctr175;
					}
					case 129:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 65u ) {
							goto _ctr184;
						}
						goto _ctr175;
					}
					case 130:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 98u ) {
							goto _ctr185;
						}
						goto _ctr175;
					}
					case 131:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr186;
						}
						goto _ctr175;
					}
					case 132:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr187;
						}
						goto _ctr175;
					}
					case 133:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr188;
						}
						goto _ctr175;
					}
					case 134:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr189;
						}
						goto _ctr175;
					}
					case 135:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr190;
						}
						goto _ctr175;
					}
					case 136:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr191;
						}
						goto _ctr175;
					}
					case 137:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr192;
						}
						goto _ctr175;
					}
					case 138:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr193;
						}
						goto _ctr175;
					}
					case 139:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr194;
						}
						goto _ctr175;
					}
					case 140:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr195;
						}
						goto _ctr175;
					}
					case 141:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr196;
						}
						goto _ctr175;
					}
					case 142:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr197;
							}
							case 101u: {
								goto _ctr198;
							}
						}
						goto _ctr175;
					}
					case 143:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr199;
						}
						goto _ctr175;
					}
					case 144:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr200;
						}
						goto _ctr175;
					}
					case 145:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr201;
						}
						goto _ctr175;
					}
					case 146:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr202;
						}
						goto _ctr175;
					}
					case 147:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr203;
						}
						goto _ctr175;
					}
					case 148:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr204;
						}
						goto _ctr175;
					}
					case 794:
					if ( p == eof ) {
						goto _ctr1074;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 109u: {
								goto _ctr1075;
							}
							case 112u: {
								goto _ctr1076;
							}
						}
						goto _ctr1074;
					}
					case 149:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr206;
						}
						goto _ctr205;
					}
					case 150:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr207;
						}
						goto _ctr205;
					}
					case 151:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr208;
						}
						goto _ctr205;
					}
					case 152:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr209;
						}
						goto _ctr205;
					}
					case 153:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr210;
						}
						goto _ctr205;
					}
					case 154:
					if ( p == eof ) {
						goto _ctr205;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr211;
						}
						goto _ctr205;
					}
					case 155:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr212;
						}
						goto _ctr175;
					}
					case 156:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr213;
						}
						goto _ctr175;
					}
					case 157:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr214;
						}
						goto _ctr175;
					}
					case 158:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 102u ) {
							goto _ctr215;
						}
						goto _ctr175;
					}
					case 159:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr216;
						}
						goto _ctr175;
					}
					case 160:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr217;
						}
						goto _ctr175;
					}
					case 161:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr218;
						}
						goto _ctr175;
					}
					case 162:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 112u: {
								goto _ctr219;
							}
							case 114u: {
								goto _ctr220;
							}
						}
						goto _ctr175;
					}
					case 163:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr221;
						}
						goto _ctr175;
					}
					case 164:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr222;
						}
						goto _ctr175;
					}
					case 165:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr223;
						}
						goto _ctr175;
					}
					case 166:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr224;
						}
						goto _ctr175;
					}
					case 167:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr225;
						}
						goto _ctr175;
					}
					case 168:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr226;
						}
						goto _ctr175;
					}
					case 169:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr227;
						}
						goto _ctr175;
					}
					case 170:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr228;
						}
						goto _ctr175;
					}
					case 171:
					if ( p == eof ) {
						goto _ctr175;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr229;
						}
						goto _ctr175;
					}
					case 795:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr1077;
						}
						goto _ctr1055;
					}
					case 172:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 115u: {
								goto _ctr230;
							}
							case 118u: {
								goto _ctr231;
							}
						}
						goto _ctr118;
					}
					case 173:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr232;
						}
						goto _ctr118;
					}
					case 174:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr233;
						}
						goto _ctr118;
					}
					case 175:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr234;
						}
						goto _ctr118;
					}
					case 176:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr235;
						}
						goto _ctr118;
					}
					case 177:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr236;
						}
						goto _ctr118;
					}
					case 178:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr237;
						}
						goto _ctr118;
					}
					case 179:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr238;
						}
						goto _ctr118;
					}
					case 180:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr239;
						}
						goto _ctr118;
					}
					case 181:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr240;
						}
						goto _ctr118;
					}
					case 182:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr241;
						}
						goto _ctr118;
					}
					case 183:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr242;
						}
						goto _ctr118;
					}
					case 796:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 103u: {
								goto _ctr1078;
							}
							case 116u: {
								goto _ctr1079;
							}
						}
						goto _ctr1055;
					}
					case 184:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr243;
						}
						goto _ctr118;
					}
					case 185:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr244;
						}
						goto _ctr118;
					}
					case 186:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr245;
						}
						goto _ctr118;
					}
					case 187:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr246;
						}
						goto _ctr118;
					}
					case 188:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr247;
						}
						goto _ctr118;
					}
					case 189:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr248;
						}
						goto _ctr118;
					}
					case 190:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr249;
						}
						goto _ctr118;
					}
					case 191:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 72u ) {
							goto _ctr250;
						}
						goto _ctr118;
					}
					case 192:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr251;
						}
						goto _ctr118;
					}
					case 193:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr252;
						}
						goto _ctr118;
					}
					case 194:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr253;
						}
						goto _ctr118;
					}
					case 195:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr254;
						}
						goto _ctr118;
					}
					case 196:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr255;
						}
						goto _ctr118;
					}
					case 197:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr256;
						}
						goto _ctr118;
					}
					case 198:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr257;
						}
						goto _ctr118;
					}
					case 199:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr258;
						}
						goto _ctr118;
					}
					case 200:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr259;
						}
						goto _ctr118;
					}
					case 201:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr260;
						}
						goto _ctr118;
					}
					case 202:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr261;
						}
						goto _ctr118;
					}
					case 203:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr262;
						}
						goto _ctr118;
					}
					case 204:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr263;
						}
						goto _ctr118;
					}
					case 205:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr264;
						}
						goto _ctr118;
					}
					case 206:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr265;
						}
						goto _ctr118;
					}
					case 207:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr266;
						}
						goto _ctr118;
					}
					case 797:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 101u: {
								goto _ctr1080;
							}
							case 108u: {
								goto _ctr1081;
							}
							case 111u: {
								goto _ctr1082;
							}
							case 114u: {
								goto _ctr1083;
							}
							case 117u: {
								goto _ctr1084;
							}
						}
						goto _ctr1055;
					}
					case 208:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr267;
						}
						goto _ctr118;
					}
					case 209:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr268;
						}
						goto _ctr118;
					}
					case 210:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr269;
						}
						goto _ctr118;
					}
					case 211:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr270;
						}
						goto _ctr118;
					}
					case 212:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr271;
						}
						goto _ctr118;
					}
					case 213:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr272;
						}
						goto _ctr118;
					}
					case 214:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr273;
						}
						goto _ctr118;
					}
					case 215:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr274;
						}
						goto _ctr118;
					}
					case 216:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr275;
						}
						goto _ctr118;
					}
					case 217:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr276;
						}
						goto _ctr118;
					}
					case 218:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr277;
						}
						goto _ctr118;
					}
					case 219:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr278;
						}
						goto _ctr118;
					}
					case 220:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr279;
						}
						goto _ctr118;
					}
					case 221:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr280;
						}
						goto _ctr118;
					}
					case 222:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr281;
						}
						goto _ctr118;
					}
					case 223:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr282;
						}
						goto _ctr118;
					}
					case 224:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr283;
						}
						goto _ctr118;
					}
					case 225:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr284;
						}
						goto _ctr118;
					}
					case 226:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr285;
						}
						goto _ctr118;
					}
					case 227:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr286;
						}
						goto _ctr118;
					}
					case 228:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr287;
						}
						goto _ctr118;
					}
					case 229:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 106u: {
								goto _ctr288;
							}
							case 114u: {
								goto _ctr289;
							}
						}
						goto _ctr118;
					}
					case 230:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr290;
						}
						goto _ctr118;
					}
					case 231:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr291;
						}
						goto _ctr118;
					}
					case 232:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr292;
						}
						goto _ctr118;
					}
					case 233:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr293;
						}
						goto _ctr118;
					}
					case 234:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr294;
						}
						goto _ctr118;
					}
					case 235:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr295;
						}
						goto _ctr118;
					}
					case 236:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr296;
						}
						goto _ctr118;
					}
					case 237:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr297;
						}
						goto _ctr118;
					}
					case 238:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr298;
						}
						goto _ctr118;
					}
					case 239:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr299;
						}
						goto _ctr118;
					}
					case 798:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1085;
							}
							case 101u: {
								goto _ctr1086;
							}
							case 105u: {
								goto _ctr1087;
							}
						}
						goto _ctr1055;
					}
					case 240:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr300;
						}
						goto _ctr118;
					}
					case 799:
					if ( p == eof ) {
						goto _ctr1088;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 103u: {
								goto _ctr1089;
							}
							case 117u: {
								goto _ctr1090;
							}
						}
						goto _ctr1088;
					}
					case 241:
					if ( p == eof ) {
						goto _ctr301;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr302;
						}
						goto _ctr301;
					}
					case 242:
					if ( p == eof ) {
						goto _ctr301;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr303;
						}
						goto _ctr301;
					}
					case 243:
					if ( p == eof ) {
						goto _ctr301;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr304;
						}
						goto _ctr301;
					}
					case 244:
					if ( p == eof ) {
						goto _ctr301;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr305;
						}
						goto _ctr301;
					}
					case 245:
					if ( p == eof ) {
						goto _ctr301;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr306;
						}
						goto _ctr301;
					}
					case 246:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 98u ) {
							goto _ctr307;
						}
						goto _ctr118;
					}
					case 247:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr308;
						}
						goto _ctr118;
					}
					case 248:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr309;
						}
						goto _ctr118;
					}
					case 249:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 119u ) {
							goto _ctr310;
						}
						goto _ctr118;
					}
					case 250:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr311;
						}
						goto _ctr118;
					}
					case 251:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr312;
						}
						goto _ctr118;
					}
					case 252:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr313;
						}
						goto _ctr118;
					}
					case 253:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr314;
						}
						goto _ctr118;
					}
					case 254:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr315;
						}
						goto _ctr118;
					}
					case 255:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr316;
						}
						goto _ctr118;
					}
					case 800:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 109u: {
								goto _ctr1091;
							}
							case 110u: {
								goto _ctr1092;
							}
						}
						goto _ctr1055;
					}
					case 256:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr317;
						}
						goto _ctr118;
					}
					case 257:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr318;
						}
						goto _ctr118;
					}
					case 258:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr319;
						}
						goto _ctr118;
					}
					case 259:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr320;
						}
						goto _ctr118;
					}
					case 260:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr321;
						}
						goto _ctr118;
					}
					case 261:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr322;
						}
						goto _ctr118;
					}
					case 262:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr323;
						}
						goto _ctr118;
					}
					case 263:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 65u ) {
							goto _ctr324;
						}
						goto _ctr118;
					}
					case 264:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr325;
						}
						goto _ctr118;
					}
					case 265:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr326;
						}
						goto _ctr118;
					}
					case 266:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr327;
						}
						goto _ctr118;
					}
					case 267:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr328;
						}
						goto _ctr118;
					}
					case 268:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr329;
						}
						goto _ctr118;
					}
					case 269:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr330;
						}
						goto _ctr118;
					}
					case 270:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 104u: {
								goto _ctr331;
							}
							case 115u: {
								goto _ctr332;
							}
						}
						goto _ctr118;
					}
					case 271:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr333;
						}
						goto _ctr118;
					}
					case 272:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr334;
						}
						goto _ctr118;
					}
					case 273:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr335;
						}
						goto _ctr118;
					}
					case 274:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr336;
						}
						goto _ctr118;
					}
					case 275:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr337;
						}
						goto _ctr118;
					}
					case 276:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr338;
						}
						goto _ctr118;
					}
					case 277:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr339;
						}
						goto _ctr118;
					}
					case 278:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr340;
						}
						goto _ctr118;
					}
					case 279:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr341;
						}
						goto _ctr118;
					}
					case 280:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr342;
						}
						goto _ctr118;
					}
					case 281:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr343;
						}
						goto _ctr118;
					}
					case 282:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr344;
						}
						goto _ctr118;
					}
					case 283:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr345;
						}
						goto _ctr118;
					}
					case 284:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr346;
						}
						goto _ctr118;
					}
					case 285:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr347;
						}
						goto _ctr118;
					}
					case 286:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr348;
						}
						goto _ctr118;
					}
					case 287:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr349;
						}
						goto _ctr118;
					}
					case 288:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 80u ) {
							goto _ctr350;
						}
						goto _ctr118;
					}
					case 289:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr351;
						}
						goto _ctr118;
					}
					case 290:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 104u: {
								goto _ctr352;
							}
							case 114u: {
								goto _ctr353;
							}
						}
						goto _ctr118;
					}
					case 291:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr354;
						}
						goto _ctr118;
					}
					case 292:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr355;
						}
						goto _ctr118;
					}
					case 293:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 118u ) {
							goto _ctr356;
						}
						goto _ctr118;
					}
					case 294:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr357;
						}
						goto _ctr118;
					}
					case 295:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr358;
						}
						goto _ctr118;
					}
					case 296:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr359;
						}
						goto _ctr118;
					}
					case 297:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr360;
						}
						goto _ctr118;
					}
					case 298:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr361;
						}
						goto _ctr118;
					}
					case 299:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr362;
						}
						goto _ctr118;
					}
					case 801:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr1093;
						}
						goto _ctr1055;
					}
					case 300:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 118u ) {
							goto _ctr363;
						}
						goto _ctr118;
					}
					case 301:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr364;
						}
						goto _ctr118;
					}
					case 302:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr365;
						}
						goto _ctr118;
					}
					case 303:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr366;
						}
						goto _ctr118;
					}
					case 304:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr367;
						}
						goto _ctr118;
					}
					case 305:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr368;
						}
						goto _ctr118;
					}
					case 802:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1094;
							}
							case 104u: {
								goto _ctr1095;
							}
						}
						goto _ctr1055;
					}
					case 306:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 105u: {
								goto _ctr369;
							}
							case 110u: {
								goto _ctr370;
							}
							case 116u: {
								goto _ctr371;
							}
							case 121u: {
								goto _ctr372;
							}
						}
						goto _ctr118;
					}
					case 307:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr373;
						}
						goto _ctr118;
					}
					case 308:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr374;
						}
						goto _ctr118;
					}
					case 309:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr375;
						}
						goto _ctr118;
					}
					case 310:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr376;
						}
						goto _ctr118;
					}
					case 311:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr377;
						}
						goto _ctr118;
					}
					case 312:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr378;
						}
						goto _ctr118;
					}
					case 313:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr379;
						}
						goto _ctr118;
					}
					case 314:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr380;
						}
						goto _ctr118;
					}
					case 315:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr381;
						}
						goto _ctr118;
					}
					case 316:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr382;
						}
						goto _ctr118;
					}
					case 317:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr383;
						}
						goto _ctr118;
					}
					case 318:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr384;
						}
						goto _ctr118;
					}
					case 319:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr385;
						}
						goto _ctr118;
					}
					case 320:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr386;
						}
						goto _ctr118;
					}
					case 321:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr387;
						}
						goto _ctr118;
					}
					case 322:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 76u ) {
							goto _ctr388;
						}
						goto _ctr118;
					}
					case 323:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr389;
						}
						goto _ctr118;
					}
					case 324:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr390;
							}
							case 109u: {
								goto _ctr391;
							}
						}
						goto _ctr118;
					}
					case 325:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr392;
						}
						goto _ctr118;
					}
					case 326:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr393;
						}
						goto _ctr118;
					}
					case 327:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr394;
						}
						goto _ctr118;
					}
					case 328:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr395;
						}
						goto _ctr118;
					}
					case 329:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr396;
						}
						goto _ctr118;
					}
					case 330:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr397;
						}
						goto _ctr118;
					}
					case 331:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr398;
						}
						goto _ctr118;
					}
					case 332:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr399;
						}
						goto _ctr118;
					}
					case 333:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr400;
						}
						goto _ctr118;
					}
					case 803:
					if ( p == eof ) {
						goto _ctr1096;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 38u: {
								goto _ctr1097;
							}
							case 97u: {
								goto _ctr1098;
							}
							case 101u: {
								goto _ctr1099;
							}
							case 105u: {
								goto _ctr1100;
							}
							case 108u: {
								goto _ctr1101;
							}
							case 109u: {
								goto _ctr1102;
							}
							case 111u: {
								goto _ctr1103;
							}
							case 116u: {
								goto _ctr1104;
							}
							case 117u: {
								goto _ctr1105;
							}
							case 121u: {
								goto _ctr1106;
							}
						}
						goto _ctr1096;
					}
					case 334:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 111u: {
								goto _ctr402;
							}
							case 116u: {
								goto _ctr403;
							}
						}
						goto _ctr401;
					}
					case 335:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr404;
						}
						goto _ctr401;
					}
					case 336:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr405;
						}
						goto _ctr401;
					}
					case 337:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr406;
						}
						goto _ctr401;
					}
					case 338:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr407;
						}
						goto _ctr401;
					}
					case 339:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr408;
						}
						goto _ctr401;
					}
					case 340:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr409;
						}
						goto _ctr401;
					}
					case 341:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 109u: {
								goto _ctr410;
							}
							case 110u: {
								goto _ctr411;
							}
							case 115u: {
								goto _ctr412;
							}
						}
						goto _ctr401;
					}
					case 342:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 98u ) {
							goto _ctr413;
						}
						goto _ctr401;
					}
					case 343:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr414;
						}
						goto _ctr401;
					}
					case 344:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr415;
						}
						goto _ctr401;
					}
					case 345:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr416;
						}
						goto _ctr401;
					}
					case 346:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr417;
						}
						goto _ctr401;
					}
					case 347:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr418;
						}
						goto _ctr401;
					}
					case 348:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 66u ) {
							goto _ctr419;
						}
						goto _ctr401;
					}
					case 349:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr420;
						}
						goto _ctr401;
					}
					case 350:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 99u: {
								goto _ctr421;
							}
							case 100u: {
								goto _ctr422;
							}
						}
						goto _ctr401;
					}
					case 351:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr423;
						}
						goto _ctr401;
					}
					case 352:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr424;
						}
						goto _ctr401;
					}
					case 353:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr425;
						}
						goto _ctr401;
					}
					case 354:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr426;
						}
						goto _ctr401;
					}
					case 355:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr427;
						}
						goto _ctr401;
					}
					case 356:
					if ( p == eof ) {
						goto _ctr401;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr428;
						}
						goto _ctr401;
					}
					case 804:
					if ( p == eof ) {
						goto _ctr1107;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1108;
							}
							case 99u: {
								goto _ctr1109;
							}
							case 101u: {
								goto _ctr1110;
							}
							case 110u: {
								goto _ctr1111;
							}
							case 111u: {
								goto _ctr1112;
							}
							case 121u: {
								goto _ctr1113;
							}
						}
						goto _ctr1107;
					}
					case 357:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 108u: {
								goto _ctr430;
							}
							case 110u: {
								goto _ctr431;
							}
						}
						goto _ctr429;
					}
					case 358:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr432;
						}
						goto _ctr429;
					}
					case 359:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr433;
						}
						goto _ctr429;
					}
					case 360:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr434;
						}
						goto _ctr429;
					}
					case 361:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr435;
						}
						goto _ctr429;
					}
					case 362:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr436;
						}
						goto _ctr429;
					}
					case 363:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr437;
						}
						goto _ctr429;
					}
					case 364:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr438;
						}
						goto _ctr429;
					}
					case 365:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr439;
						}
						goto _ctr429;
					}
					case 366:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr440;
						}
						goto _ctr429;
					}
					case 367:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr441;
						}
						goto _ctr429;
					}
					case 805:
					if ( p == eof ) {
						goto _ctr1114;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr1115;
						}
						goto _ctr1114;
					}
					case 368:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr443;
						}
						goto _ctr442;
					}
					case 369:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr444;
						}
						goto _ctr442;
					}
					case 370:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr445;
						}
						goto _ctr442;
					}
					case 371:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr446;
						}
						goto _ctr442;
					}
					case 372:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 77u ) {
							goto _ctr447;
						}
						goto _ctr442;
					}
					case 373:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr448;
						}
						goto _ctr442;
					}
					case 374:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr449;
						}
						goto _ctr442;
					}
					case 375:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr450;
						}
						goto _ctr442;
					}
					case 376:
					if ( p == eof ) {
						goto _ctr442;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr451;
						}
						goto _ctr442;
					}
					case 377:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr452;
						}
						goto _ctr429;
					}
					case 378:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr453;
						}
						goto _ctr429;
					}
					case 379:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr454;
						}
						goto _ctr429;
					}
					case 380:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr455;
						}
						goto _ctr429;
					}
					case 381:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr456;
						}
						goto _ctr429;
					}
					case 382:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr457;
						}
						goto _ctr429;
					}
					case 383:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr458;
						}
						goto _ctr429;
					}
					case 384:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr459;
						}
						goto _ctr429;
					}
					case 385:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr460;
						}
						goto _ctr429;
					}
					case 386:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr461;
						}
						goto _ctr429;
					}
					case 387:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr462;
						}
						goto _ctr429;
					}
					case 388:
					if ( p == eof ) {
						goto _ctr429;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr463;
						}
						goto _ctr429;
					}
					case 806:
					if ( p == eof ) {
						goto _ctr1116;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 100u: {
								goto _ctr1117;
							}
							case 101u: {
								goto _ctr1118;
							}
							case 107u: {
								goto _ctr1119;
							}
							case 108u: {
								goto _ctr1120;
							}
							case 111u: {
								goto _ctr1121;
							}
						}
						goto _ctr1116;
					}
					case 389:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 119u ) {
							goto _ctr465;
						}
						goto _ctr464;
					}
					case 390:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr466;
						}
						goto _ctr464;
					}
					case 391:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 84u ) {
							goto _ctr467;
						}
						goto _ctr464;
					}
					case 392:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr468;
						}
						goto _ctr464;
					}
					case 393:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr469;
						}
						goto _ctr464;
					}
					case 394:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr470;
						}
						goto _ctr464;
					}
					case 395:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 76u ) {
							goto _ctr471;
						}
						goto _ctr464;
					}
					case 396:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr472;
						}
						goto _ctr464;
					}
					case 397:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr473;
						}
						goto _ctr464;
					}
					case 398:
					if ( p == eof ) {
						goto _ctr464;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr474;
						}
						goto _ctr464;
					}
					case 807:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 103u: {
								goto _ctr1122;
							}
							case 108u: {
								goto _ctr1123;
							}
							case 114u: {
								goto _ctr1124;
							}
							case 115u: {
								goto _ctr1125;
							}
						}
						goto _ctr1055;
					}
					case 399:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr475;
						}
						goto _ctr118;
					}
					case 400:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr476;
						}
						goto _ctr118;
					}
					case 401:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr477;
						}
						goto _ctr118;
					}
					case 402:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 95u: {
								goto _ctr478;
							}
							case 100u: {
								goto _ctr479;
							}
						}
						goto _ctr118;
					}
					case 403:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 67u ) {
							goto _ctr480;
						}
						goto _ctr118;
					}
					case 404:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr481;
						}
						goto _ctr118;
					}
					case 405:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr482;
						}
						goto _ctr118;
					}
					case 406:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr483;
						}
						goto _ctr118;
					}
					case 407:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr484;
						}
						goto _ctr118;
					}
					case 408:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr485;
						}
						goto _ctr118;
					}
					case 409:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 73u: {
								goto _ctr486;
							}
							case 80u: {
								goto _ctr487;
							}
							case 83u: {
								goto _ctr488;
							}
							case 84u: {
								goto _ctr489;
							}
						}
						goto _ctr118;
					}
					case 410:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr490;
						}
						goto _ctr118;
					}
					case 411:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr491;
						}
						goto _ctr118;
					}
					case 412:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr492;
						}
						goto _ctr118;
					}
					case 413:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr493;
						}
						goto _ctr118;
					}
					case 414:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr494;
						}
						goto _ctr118;
					}
					case 415:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr495;
						}
						goto _ctr118;
					}
					case 416:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr496;
						}
						goto _ctr118;
					}
					case 417:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr497;
						}
						goto _ctr118;
					}
					case 418:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr498;
						}
						goto _ctr118;
					}
					case 419:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr499;
						}
						goto _ctr118;
					}
					case 420:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr500;
						}
						goto _ctr118;
					}
					case 421:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr501;
						}
						goto _ctr118;
					}
					case 422:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr502;
						}
						goto _ctr118;
					}
					case 423:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr503;
						}
						goto _ctr118;
					}
					case 424:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr504;
						}
						goto _ctr118;
					}
					case 425:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr505;
						}
						goto _ctr118;
					}
					case 426:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 65u ) {
							goto _ctr506;
						}
						goto _ctr118;
					}
					case 427:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr507;
						}
						goto _ctr118;
					}
					case 428:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr508;
						}
						goto _ctr118;
					}
					case 429:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 98u ) {
							goto _ctr509;
						}
						goto _ctr118;
					}
					case 430:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr510;
						}
						goto _ctr118;
					}
					case 431:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr511;
						}
						goto _ctr118;
					}
					case 432:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr512;
						}
						goto _ctr118;
					}
					case 433:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr513;
						}
						goto _ctr118;
					}
					case 434:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr514;
						}
						goto _ctr118;
					}
					case 435:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 107u ) {
							goto _ctr515;
						}
						goto _ctr118;
					}
					case 436:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr516;
						}
						goto _ctr118;
					}
					case 437:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr517;
						}
						goto _ctr118;
					}
					case 438:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr518;
						}
						goto _ctr118;
					}
					case 439:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr519;
						}
						goto _ctr118;
					}
					case 440:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr520;
						}
						goto _ctr118;
					}
					case 441:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr521;
						}
						goto _ctr118;
					}
					case 442:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr522;
						}
						goto _ctr118;
					}
					case 443:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr523;
						}
						goto _ctr118;
					}
					case 444:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 121u ) {
							goto _ctr524;
						}
						goto _ctr118;
					}
					case 445:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr525;
						}
						goto _ctr118;
					}
					case 808:
					if ( p == eof ) {
						goto _ctr1126;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 99u: {
								goto _ctr1127;
							}
							case 100u: {
								goto _ctr1128;
							}
							case 101u: {
								goto _ctr1129;
							}
							case 102u: {
								goto _ctr1130;
							}
							case 104u: {
								goto _ctr1131;
							}
							case 105u: {
								goto _ctr1132;
							}
							case 111u: {
								goto _ctr1133;
							}
							case 115u: {
								goto _ctr1134;
							}
						}
						goto _ctr1126;
					}
					case 446:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr527;
							}
							case 111u: {
								goto _ctr528;
							}
						}
						goto _ctr526;
					}
					case 447:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr529;
						}
						goto _ctr526;
					}
					case 448:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr530;
						}
						goto _ctr526;
					}
					case 449:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr531;
						}
						goto _ctr526;
					}
					case 450:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 80u ) {
							goto _ctr532;
						}
						goto _ctr526;
					}
					case 451:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr533;
						}
						goto _ctr526;
					}
					case 452:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr534;
						}
						goto _ctr526;
					}
					case 453:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr535;
						}
						goto _ctr526;
					}
					case 454:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr536;
						}
						goto _ctr526;
					}
					case 455:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr537;
						}
						goto _ctr526;
					}
					case 456:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr538;
						}
						goto _ctr526;
					}
					case 457:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr539;
						}
						goto _ctr526;
					}
					case 458:
					if ( p == eof ) {
						goto _ctr526;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr540;
						}
						goto _ctr526;
					}
					case 809:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 101u: {
								goto _ctr1135;
							}
							case 117u: {
								goto _ctr1136;
							}
						}
						goto _ctr1055;
					}
					case 459:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 106u ) {
							goto _ctr541;
						}
						goto _ctr118;
					}
					case 460:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr542;
						}
						goto _ctr118;
					}
					case 461:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr543;
						}
						goto _ctr118;
					}
					case 462:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr544;
						}
						goto _ctr118;
					}
					case 463:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr545;
						}
						goto _ctr118;
					}
					case 464:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr546;
						}
						goto _ctr118;
					}
					case 465:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr547;
						}
						goto _ctr118;
					}
					case 810:
					if ( p == eof ) {
						goto _ctr1137;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1138;
							}
							case 99u: {
								goto _ctr1139;
							}
							case 104u: {
								goto _ctr1140;
							}
							case 105u: {
								goto _ctr1141;
							}
							case 107u: {
								goto _ctr1142;
							}
							case 109u: {
								goto _ctr1143;
							}
							case 111u: {
								goto _ctr1144;
							}
							case 117u: {
								goto _ctr1145;
							}
							case 121u: {
								goto _ctr1146;
							}
						}
						goto _ctr1137;
					}
					case 466:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 109u: {
								goto _ctr549;
							}
							case 117u: {
								goto _ctr550;
							}
						}
						goto _ctr548;
					}
					case 467:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr551;
						}
						goto _ctr548;
					}
					case 468:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr552;
						}
						goto _ctr548;
					}
					case 469:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr553;
						}
						goto _ctr548;
					}
					case 470:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr554;
						}
						goto _ctr548;
					}
					case 471:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr555;
						}
						goto _ctr548;
					}
					case 472:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr556;
						}
						goto _ctr548;
					}
					case 473:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr557;
						}
						goto _ctr548;
					}
					case 474:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr558;
						}
						goto _ctr548;
					}
					case 475:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr559;
						}
						goto _ctr548;
					}
					case 476:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr560;
						}
						goto _ctr548;
					}
					case 477:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr561;
						}
						goto _ctr548;
					}
					case 478:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr562;
						}
						goto _ctr548;
					}
					case 479:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr563;
						}
						goto _ctr548;
					}
					case 480:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr564;
						}
						goto _ctr548;
					}
					case 481:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 118u ) {
							goto _ctr565;
						}
						goto _ctr548;
					}
					case 482:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr566;
						}
						goto _ctr548;
					}
					case 483:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr567;
						}
						goto _ctr548;
					}
					case 484:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr568;
						}
						goto _ctr548;
					}
					case 485:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr569;
						}
						goto _ctr548;
					}
					case 486:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr570;
						}
						goto _ctr548;
					}
					case 487:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr571;
						}
						goto _ctr548;
					}
					case 488:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr572;
						}
						goto _ctr548;
					}
					case 489:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr573;
						}
						goto _ctr548;
					}
					case 490:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr574;
						}
						goto _ctr548;
					}
					case 491:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr575;
						}
						goto _ctr548;
					}
					case 492:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr576;
						}
						goto _ctr548;
					}
					case 493:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr577;
						}
						goto _ctr548;
					}
					case 494:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr578;
						}
						goto _ctr548;
					}
					case 495:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr579;
						}
						goto _ctr548;
					}
					case 496:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr580;
						}
						goto _ctr548;
					}
					case 497:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 108u: {
								goto _ctr581;
							}
							case 114u: {
								goto _ctr582;
							}
						}
						goto _ctr548;
					}
					case 498:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr583;
						}
						goto _ctr548;
					}
					case 499:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr584;
						}
						goto _ctr548;
					}
					case 500:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr585;
						}
						goto _ctr548;
					}
					case 501:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr586;
						}
						goto _ctr548;
					}
					case 502:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 78u ) {
							goto _ctr587;
						}
						goto _ctr548;
					}
					case 503:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr588;
						}
						goto _ctr548;
					}
					case 504:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr589;
						}
						goto _ctr548;
					}
					case 505:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr590;
						}
						goto _ctr548;
					}
					case 506:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr591;
						}
						goto _ctr548;
					}
					case 507:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr592;
						}
						goto _ctr548;
					}
					case 508:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr593;
						}
						goto _ctr548;
					}
					case 509:
					if ( p == eof ) {
						goto _ctr548;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr594;
						}
						goto _ctr548;
					}
					case 811:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1147;
							}
							case 101u: {
								goto _ctr1148;
							}
							case 104u: {
								goto _ctr1149;
							}
							case 105u: {
								goto _ctr1150;
							}
						}
						goto _ctr1055;
					}
					case 510:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 103u: {
								goto _ctr595;
							}
							case 105u: {
								goto _ctr596;
							}
							case 109u: {
								goto _ctr597;
							}
						}
						goto _ctr118;
					}
					case 511:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr598;
							}
							case 98u: {
								goto _ctr599;
							}
						}
						goto _ctr118;
					}
					case 512:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr600;
						}
						goto _ctr118;
					}
					case 513:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 111u ) {
							goto _ctr601;
						}
						goto _ctr118;
					}
					case 514:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr602;
						}
						goto _ctr118;
					}
					case 515:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr603;
						}
						goto _ctr118;
					}
					case 516:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr604;
						}
						goto _ctr118;
					}
					case 517:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 119u ) {
							goto _ctr605;
						}
						goto _ctr118;
					}
					case 518:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr606;
						}
						goto _ctr118;
					}
					case 519:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 95u ) {
							goto _ctr607;
						}
						goto _ctr118;
					}
					case 520:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 76u: {
								goto _ctr608;
							}
							case 84u: {
								goto _ctr609;
							}
							case 86u: {
								goto _ctr610;
							}
						}
						goto _ctr118;
					}
					case 521:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr611;
						}
						goto _ctr118;
					}
					case 522:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr612;
						}
						goto _ctr118;
					}
					case 523:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr613;
						}
						goto _ctr118;
					}
					case 524:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 109u ) {
							goto _ctr614;
						}
						goto _ctr118;
					}
					case 525:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr615;
						}
						goto _ctr118;
					}
					case 526:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr616;
						}
						goto _ctr118;
					}
					case 527:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr617;
						}
						goto _ctr118;
					}
					case 528:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr618;
						}
						goto _ctr118;
					}
					case 529:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr619;
						}
						goto _ctr118;
					}
					case 530:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 108u ) {
							goto _ctr620;
						}
						goto _ctr118;
					}
					case 531:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr621;
						}
						goto _ctr118;
					}
					case 532:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr622;
						}
						goto _ctr118;
					}
					case 533:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 117u ) {
							goto _ctr623;
						}
						goto _ctr118;
					}
					case 534:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr624;
						}
						goto _ctr118;
					}
					case 535:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr625;
							}
							case 105u: {
								goto _ctr626;
							}
						}
						goto _ctr118;
					}
					case 536:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr627;
						}
						goto _ctr118;
					}
					case 537:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr628;
						}
						goto _ctr118;
					}
					case 538:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 98u: {
								goto _ctr629;
							}
							case 102u: {
								goto _ctr630;
							}
						}
						goto _ctr118;
					}
					case 539:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 101u ) {
							goto _ctr631;
						}
						goto _ctr118;
					}
					case 540:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr632;
						}
						goto _ctr118;
					}
					case 541:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr633;
						}
						goto _ctr118;
					}
					case 542:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr634;
						}
						goto _ctr118;
					}
					case 543:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr635;
						}
						goto _ctr118;
					}
					case 544:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr636;
						}
						goto _ctr118;
					}
					case 545:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr637;
						}
						goto _ctr118;
					}
					case 546:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr638;
						}
						goto _ctr118;
					}
					case 547:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 104u ) {
							goto _ctr639;
						}
						goto _ctr118;
					}
					case 812:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						if ( ( (*( p))) == 103u ) {
							goto _ctr1151;
						}
						goto _ctr1055;
					}
					case 548:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr640;
						}
						goto _ctr118;
					}
					case 549:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 114u ) {
							goto _ctr641;
						}
						goto _ctr118;
					}
					case 550:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr642;
						}
						goto _ctr118;
					}
					case 551:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 116u ) {
							goto _ctr643;
						}
						goto _ctr118;
					}
					case 552:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr644;
						}
						goto _ctr118;
					}
					case 553:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 99u ) {
							goto _ctr645;
						}
						goto _ctr118;
					}
					case 813:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						if ( ( (*( p))) == 97u ) {
							goto _ctr1152;
						}
						goto _ctr1055;
					}
					case 554:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr646;
						}
						goto _ctr118;
					}
					case 814:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 97u: {
								goto _ctr1153;
							}
							case 112u: {
								goto _ctr1154;
							}
							case 115u: {
								goto _ctr1155;
							}
							case 119u: {
								goto _ctr1156;
							}
						}
						goto _ctr1055;
					}
					case 555:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 110u ) {
							goto _ctr647;
						}
						goto _ctr118;
					}
					case 556:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 115u ) {
							goto _ctr648;
						}
						goto _ctr118;
					}
					case 557:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 112u ) {
							goto _ctr649;
						}
						goto _ctr118;
					}
					case 558:
					if ( p == eof ) {
						goto _ctr118;	goto _again;
					}
					else {
						if ( ( (*( p))) == 100u ) {
							goto _ctr650;
						}
						goto _ctr118;
					}
					case 815:
					if ( p == eof ) {
						goto _ctr1055;	goto _again;
					}
					else {
						if ( ( (*( p))) == 105u ) {
							goto _ctr1157;
						}
						goto _ctr1055;
					}
					case 816:
					if ( p == eof ) {
						goto _ctr1158;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 108u: {
								goto _ctr1159;
							}
							case 112u: {
								goto _ctr1160;
							}
							case 115u: {
								goto _ctr1161;
							}
						}
						goto _ctr1158;
					}
					case 818:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 8666 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1163;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 67u: {
								goto _ctr1165;
							}
							case 76u: {
								goto _ctr1166;
							}
							case 77u: {
								goto _ctr1167;
							}
							case 78u: {
								goto _ctr1168;
							}
							case 80u: {
								goto _ctr1169;
							}
							case 83u: {
								goto _ctr1170;
							}
							case 90u: {
								goto _ctr1171;
							}
						}
						goto _ctr1164;
					}
					case 819:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 8702 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1172;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 45u: {
								goto _ctr1174;
							}
							case 91u: {
								goto _ctr1175;
							}
							case 92u: {
								goto _ctr1176;
							}
							case 93u: {
								goto _ctr1177;
							}
						}
						if ( ( (*( p))) < 224u ) {
							if ( ( (*( p))) > 191u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 8727 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1180;
									} else {
										goto _ctr1178;
									}
									
								}
							} else if ( ( (*( p))) >= 128u ) {
								_ck = 0;
								if ( ( mode.utf8 )
#line 8740 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1179;
								} else {
									goto _ctr1178;
								}
								
							}
						} else if ( ( (*( p))) > 239u ) {
							if ( ( (*( p))) > 247u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 8755 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1179;
									} else {
										goto _ctr1178;
									}
									
								}
							} else {
								_ck = 0;
								if ( ( mode.utf8 )
#line 8768 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1182;
								} else {
									goto _ctr1178;
								}
								
							}
						} else {
							_ck = 0;
							if ( ( mode.utf8 )
#line 8781 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1181;
							} else {
								goto _ctr1178;
							}
							
						}
						goto _ctr1173;
					}
					case 820:
					if ( p == eof ) {
						goto _ctr1183;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr661;
							}
							case 58u: {
								goto _ctr1184;
							}
							case 61u: {
								goto _ctr858;
							}
						}
						goto _ctr1183;
					}
					case 563:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr662;
							}
							case 92u: {
								goto _ctr663;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr661;
					}
					case 564:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr662;
							}
							case 92u: {
								goto _ctr663;
							}
							case 93u: {
								goto _ctr664;
							}
						}
						goto _ctr661;
					}
					case 565:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 46u: {
								goto _ctr662;
							}
							case 92u: {
								goto _ctr663;
							}
						}
						goto _ctr661;
					}
					case 566:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 94u: {
								goto _ctr668;
							}
							case 97u: {
								goto _ctr669;
							}
							case 98u: {
								goto _ctr670;
							}
							case 99u: {
								goto _ctr671;
							}
							case 100u: {
								goto _ctr672;
							}
							case 103u: {
								goto _ctr673;
							}
							case 108u: {
								goto _ctr674;
							}
							case 112u: {
								goto _ctr675;
							}
							case 115u: {
								goto _ctr676;
							}
							case 117u: {
								goto _ctr677;
							}
							case 119u: {
								goto _ctr678;
							}
							case 120u: {
								goto _ctr679;
							}
						}
						goto _ctr665;
					}
					case 567:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 568:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr680;
							}
						}
						goto _ctr665;
					}
					case 569:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
						}
						goto _ctr665;
					}
					case 570:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr681;
							}
							case 98u: {
								goto _ctr682;
							}
							case 99u: {
								goto _ctr683;
							}
							case 100u: {
								goto _ctr684;
							}
							case 103u: {
								goto _ctr685;
							}
							case 108u: {
								goto _ctr686;
							}
							case 112u: {
								goto _ctr687;
							}
							case 115u: {
								goto _ctr688;
							}
							case 117u: {
								goto _ctr689;
							}
							case 119u: {
								goto _ctr690;
							}
							case 120u: {
								goto _ctr691;
							}
						}
						goto _ctr665;
					}
					case 571:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr692;
							}
							case 115u: {
								goto _ctr693;
							}
						}
						goto _ctr665;
					}
					case 572:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr694;
							}
							case 112u: {
								goto _ctr695;
							}
						}
						goto _ctr665;
					}
					case 573:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 117u: {
								goto _ctr696;
							}
						}
						goto _ctr665;
					}
					case 574:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 109u: {
								goto _ctr697;
							}
						}
						goto _ctr665;
					}
					case 575:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr698;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 576:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr699;
							}
						}
						goto _ctr665;
					}
					case 577:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 104u: {
								goto _ctr700;
							}
						}
						goto _ctr665;
					}
					case 578:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr701;
							}
						}
						goto _ctr665;
					}
					case 579:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr702;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 580:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr703;
							}
						}
						goto _ctr665;
					}
					case 581:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr704;
							}
						}
						goto _ctr665;
					}
					case 582:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr705;
							}
						}
						goto _ctr665;
					}
					case 583:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr706;
							}
						}
						goto _ctr665;
					}
					case 584:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr707;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 585:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr708;
							}
						}
						goto _ctr665;
					}
					case 586:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr709;
							}
						}
						goto _ctr665;
					}
					case 587:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr710;
							}
						}
						goto _ctr665;
					}
					case 588:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr711;
							}
						}
						goto _ctr665;
					}
					case 589:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 107u: {
								goto _ctr712;
							}
						}
						goto _ctr665;
					}
					case 590:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr713;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 591:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr714;
							}
						}
						goto _ctr665;
					}
					case 592:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr715;
							}
						}
						goto _ctr665;
					}
					case 593:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr716;
							}
						}
						goto _ctr665;
					}
					case 594:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr717;
							}
						}
						goto _ctr665;
					}
					case 595:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr718;
							}
						}
						goto _ctr665;
					}
					case 596:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr719;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 597:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr720;
							}
						}
						goto _ctr665;
					}
					case 598:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr721;
							}
						}
						goto _ctr665;
					}
					case 599:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 103u: {
								goto _ctr722;
							}
						}
						goto _ctr665;
					}
					case 600:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr723;
							}
						}
						goto _ctr665;
					}
					case 601:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr724;
							}
						}
						goto _ctr665;
					}
					case 602:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr725;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 603:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr726;
							}
						}
						goto _ctr665;
					}
					case 604:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr727;
							}
						}
						goto _ctr665;
					}
					case 605:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr728;
							}
						}
						goto _ctr665;
					}
					case 606:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr729;
							}
						}
						goto _ctr665;
					}
					case 607:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 104u: {
								goto _ctr730;
							}
						}
						goto _ctr665;
					}
					case 608:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr731;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 609:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr732;
							}
						}
						goto _ctr665;
					}
					case 610:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 111u: {
								goto _ctr733;
							}
						}
						goto _ctr665;
					}
					case 611:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 119u: {
								goto _ctr734;
							}
						}
						goto _ctr665;
					}
					case 612:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr735;
							}
						}
						goto _ctr665;
					}
					case 613:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr736;
							}
						}
						goto _ctr665;
					}
					case 614:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr737;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 615:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr738;
							}
						}
						goto _ctr665;
					}
					case 616:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr739;
							}
							case 117u: {
								goto _ctr740;
							}
						}
						goto _ctr665;
					}
					case 617:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr741;
							}
						}
						goto _ctr665;
					}
					case 618:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr742;
							}
						}
						goto _ctr665;
					}
					case 619:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr743;
							}
						}
						goto _ctr665;
					}
					case 620:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr744;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 621:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr745;
							}
						}
						goto _ctr665;
					}
					case 622:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr746;
							}
						}
						goto _ctr665;
					}
					case 623:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr747;
							}
						}
						goto _ctr665;
					}
					case 624:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr748;
							}
						}
						goto _ctr665;
					}
					case 625:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr749;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 626:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr750;
							}
						}
						goto _ctr665;
					}
					case 627:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr751;
							}
						}
						goto _ctr665;
					}
					case 628:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr752;
							}
						}
						goto _ctr665;
					}
					case 629:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr753;
							}
						}
						goto _ctr665;
					}
					case 630:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr754;
							}
						}
						goto _ctr665;
					}
					case 631:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr755;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 632:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr756;
							}
						}
						goto _ctr665;
					}
					case 633:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr757;
							}
						}
						goto _ctr665;
					}
					case 634:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr758;
							}
						}
						goto _ctr665;
					}
					case 635:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr759;
							}
						}
						goto _ctr665;
					}
					case 636:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr760;
							}
						}
						goto _ctr665;
					}
					case 637:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr761;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 638:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr762;
							}
						}
						goto _ctr665;
					}
					case 639:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 111u: {
								goto _ctr763;
							}
						}
						goto _ctr665;
					}
					case 640:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr764;
							}
						}
						goto _ctr665;
					}
					case 641:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 100u: {
								goto _ctr765;
							}
						}
						goto _ctr665;
					}
					case 642:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr766;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 643:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr767;
							}
						}
						goto _ctr665;
					}
					case 644:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 100u: {
								goto _ctr768;
							}
						}
						goto _ctr665;
					}
					case 645:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr769;
							}
						}
						goto _ctr665;
					}
					case 646:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 103u: {
								goto _ctr770;
							}
						}
						goto _ctr665;
					}
					case 647:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr771;
							}
						}
						goto _ctr665;
					}
					case 648:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr772;
							}
						}
						goto _ctr665;
					}
					case 649:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr773;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 650:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr774;
							}
						}
						goto _ctr665;
					}
					case 651:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr775;
							}
							case 115u: {
								goto _ctr776;
							}
						}
						goto _ctr665;
					}
					case 652:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr777;
							}
							case 112u: {
								goto _ctr778;
							}
						}
						goto _ctr665;
					}
					case 653:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 117u: {
								goto _ctr779;
							}
						}
						goto _ctr665;
					}
					case 654:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 109u: {
								goto _ctr780;
							}
						}
						goto _ctr665;
					}
					case 655:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr781;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 656:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr782;
							}
						}
						goto _ctr665;
					}
					case 657:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 104u: {
								goto _ctr783;
							}
						}
						goto _ctr665;
					}
					case 658:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr784;
							}
						}
						goto _ctr665;
					}
					case 659:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr785;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 660:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr786;
							}
						}
						goto _ctr665;
					}
					case 661:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr787;
							}
						}
						goto _ctr665;
					}
					case 662:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr788;
							}
						}
						goto _ctr665;
					}
					case 663:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr789;
							}
						}
						goto _ctr665;
					}
					case 664:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr790;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 665:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr791;
							}
						}
						goto _ctr665;
					}
					case 666:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr792;
							}
						}
						goto _ctr665;
					}
					case 667:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr793;
							}
						}
						goto _ctr665;
					}
					case 668:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr794;
							}
						}
						goto _ctr665;
					}
					case 669:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 107u: {
								goto _ctr795;
							}
						}
						goto _ctr665;
					}
					case 670:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr796;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 671:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr797;
							}
						}
						goto _ctr665;
					}
					case 672:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr798;
							}
						}
						goto _ctr665;
					}
					case 673:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr799;
							}
						}
						goto _ctr665;
					}
					case 674:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr800;
							}
						}
						goto _ctr665;
					}
					case 675:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 108u: {
								goto _ctr801;
							}
						}
						goto _ctr665;
					}
					case 676:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr802;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 677:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr803;
							}
						}
						goto _ctr665;
					}
					case 678:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr804;
							}
						}
						goto _ctr665;
					}
					case 679:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 103u: {
								goto _ctr805;
							}
						}
						goto _ctr665;
					}
					case 680:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr806;
							}
						}
						goto _ctr665;
					}
					case 681:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr807;
							}
						}
						goto _ctr665;
					}
					case 682:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr808;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 683:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr809;
							}
						}
						goto _ctr665;
					}
					case 684:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr810;
							}
						}
						goto _ctr665;
					}
					case 685:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr811;
							}
						}
						goto _ctr665;
					}
					case 686:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr812;
							}
						}
						goto _ctr665;
					}
					case 687:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 104u: {
								goto _ctr813;
							}
						}
						goto _ctr665;
					}
					case 688:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr814;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 689:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr815;
							}
						}
						goto _ctr665;
					}
					case 690:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 111u: {
								goto _ctr816;
							}
						}
						goto _ctr665;
					}
					case 691:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 119u: {
								goto _ctr817;
							}
						}
						goto _ctr665;
					}
					case 692:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr818;
							}
						}
						goto _ctr665;
					}
					case 693:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr819;
							}
						}
						goto _ctr665;
					}
					case 694:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr820;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 695:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr821;
							}
						}
						goto _ctr665;
					}
					case 696:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr822;
							}
							case 117u: {
								goto _ctr823;
							}
						}
						goto _ctr665;
					}
					case 697:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr824;
							}
						}
						goto _ctr665;
					}
					case 698:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr825;
							}
						}
						goto _ctr665;
					}
					case 699:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr826;
							}
						}
						goto _ctr665;
					}
					case 700:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr827;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 701:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr828;
							}
						}
						goto _ctr665;
					}
					case 702:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 110u: {
								goto _ctr829;
							}
						}
						goto _ctr665;
					}
					case 703:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr830;
							}
						}
						goto _ctr665;
					}
					case 704:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr831;
							}
						}
						goto _ctr665;
					}
					case 705:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr832;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 706:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr833;
							}
						}
						goto _ctr665;
					}
					case 707:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr834;
							}
						}
						goto _ctr665;
					}
					case 708:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 97u: {
								goto _ctr835;
							}
						}
						goto _ctr665;
					}
					case 709:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 99u: {
								goto _ctr836;
							}
						}
						goto _ctr665;
					}
					case 710:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr837;
							}
						}
						goto _ctr665;
					}
					case 711:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr838;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 712:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr839;
							}
						}
						goto _ctr665;
					}
					case 713:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr840;
							}
						}
						goto _ctr665;
					}
					case 714:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 112u: {
								goto _ctr841;
							}
						}
						goto _ctr665;
					}
					case 715:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 101u: {
								goto _ctr842;
							}
						}
						goto _ctr665;
					}
					case 716:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr843;
							}
						}
						goto _ctr665;
					}
					case 717:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr844;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 718:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr845;
							}
						}
						goto _ctr665;
					}
					case 719:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 111u: {
								goto _ctr846;
							}
						}
						goto _ctr665;
					}
					case 720:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 114u: {
								goto _ctr847;
							}
						}
						goto _ctr665;
					}
					case 721:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 100u: {
								goto _ctr848;
							}
						}
						goto _ctr665;
					}
					case 722:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr849;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 723:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr850;
							}
						}
						goto _ctr665;
					}
					case 724:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 100u: {
								goto _ctr851;
							}
						}
						goto _ctr665;
					}
					case 725:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr852;
							}
						}
						goto _ctr665;
					}
					case 726:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 103u: {
								goto _ctr853;
							}
						}
						goto _ctr665;
					}
					case 727:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 105u: {
								goto _ctr854;
							}
						}
						goto _ctr665;
					}
					case 728:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
							case 116u: {
								goto _ctr855;
							}
						}
						goto _ctr665;
					}
					case 729:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr856;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr665;
					}
					case 730:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 58u: {
								goto _ctr666;
							}
							case 92u: {
								goto _ctr667;
							}
							case 93u: {
								goto _ctr857;
							}
						}
						goto _ctr665;
					}
					case 731:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr859;
							}
							case 92u: {
								goto _ctr860;
							}
							case 93u: {
								goto _ctr660;
							}
						}
						goto _ctr858;
					}
					case 732:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr859;
							}
							case 92u: {
								goto _ctr860;
							}
							case 93u: {
								goto _ctr664;
							}
						}
						goto _ctr858;
					}
					case 733:
					if ( p == eof ) {
						goto _ctr660;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 61u: {
								goto _ctr859;
							}
							case 92u: {
								goto _ctr860;
							}
						}
						goto _ctr858;
					}
					case 821:
					if ( p == eof ) {
						goto _ctr1183;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 48u: {
								goto _ctr1186;
							}
							case 68u: {
								goto _ctr1190;
							}
							case 69u: {
								goto _ctr1191;
							}
							case 72u: {
								goto _ctr1192;
							}
							case 76u: {
								goto _ctr1193;
							}
							case 78u: {
								goto _ctr1193;
							}
							case 80u: {
								goto _ctr1194;
							}
							case 81u: {
								goto _ctr1195;
							}
							case 83u: {
								goto _ctr1196;
							}
							case 85u: {
								goto _ctr1193;
							}
							case 86u: {
								goto _ctr1197;
							}
							case 87u: {
								goto _ctr1198;
							}
							case 97u: {
								goto _ctr1199;
							}
							case 98u: {
								goto _ctr1200;
							}
							case 99u: {
								goto _ctr1201;
							}
							case 100u: {
								goto _ctr1202;
							}
							case 101u: {
								goto _ctr1203;
							}
							case 102u: {
								goto _ctr1204;
							}
							case 103u: {
								goto _ctr1205;
							}
							case 104u: {
								goto _ctr1206;
							}
							case 108u: {
								goto _ctr1193;
							}
							case 110u: {
								goto _ctr1207;
							}
							case 111u: {
								goto _ctr1208;
							}
							case 112u: {
								goto _ctr1209;
							}
							case 114u: {
								goto _ctr1210;
							}
							case 115u: {
								goto _ctr1211;
							}
							case 116u: {
								goto _ctr1212;
							}
							case 117u: {
								goto _ctr1193;
							}
							case 118u: {
								goto _ctr1213;
							}
							case 119u: {
								goto _ctr1214;
							}
							case 120u: {
								goto _ctr1215;
							}
						}
						if ( ( (*( p))) < 56u ) {
							if ( 49u <= ( (*( p))) ) {
								goto _ctr1187;
							}
						} else if ( ( (*( p))) > 57u ) {
							if ( ( (*( p))) > 90u ) {
								if ( 105u <= ( (*( p))) && ( (*( p))) <= 122u ) {
									goto _ctr1189;
								}
							} else if ( ( (*( p))) >= 65u ) {
								goto _ctr1189;
							}
						} else {
							goto _ctr1188;
						}
						goto _ctr1185;
					}
					case 822:
					if ( p == eof ) {
						goto _ctr1216;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr1217;
						}
						goto _ctr1216;
					}
					case 823:
					if ( p == eof ) {
						goto _ctr1216;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr1218;
						}
						goto _ctr1216;
					}
					case 824:
					if ( p == eof ) {
						goto _ctr1219;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr1220;
						}
						goto _ctr1219;
					}
					case 825:
					if ( p == eof ) {
						goto _ctr1219;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr1221;
						}
						goto _ctr1219;
					}
					case 826:
					if ( p == eof ) {
						goto _ctr1222;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1224;
						}
						goto _ctr1223;
					}
					case 827:
					if ( p == eof ) {
						goto _ctr1225;	goto _again;
					}
					else {
						goto _ctr1226;
					}
					case 828:
					if ( p == eof ) {
						goto _ctr1227;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1228;
						}
						goto _ctr1227;
					}
					case 734:
					if ( p == eof ) {
						goto _ctr861;	goto _again;
					}
					else {
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr862;
						}
						goto _ctr861;
					}
					case 735:
					if ( p == eof ) {
						goto _ctr861;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr863;
						}
						if ( 48u <= ( (*( p))) && ( (*( p))) <= 55u ) {
							goto _ctr862;
						}
						goto _ctr861;
					}
					case 829:
					if ( p == eof ) {
						goto _ctr1229;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1231;
						}
						goto _ctr1230;
					}
					case 830:
					if ( p == eof ) {
						goto _ctr1232;	goto _again;
					}
					else {
						if ( ( (*( p))) == 123u ) {
							goto _ctr1236;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr1233;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr1235;
							}
						} else {
							goto _ctr1234;
						}
						goto _ctr1232;
					}
					case 831:
					if ( p == eof ) {
						goto _ctr1232;	goto _again;
					}
					else {
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr1237;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr1239;
							}
						} else {
							goto _ctr1238;
						}
						goto _ctr1232;
					}
					case 832:
					if ( p == eof ) {
						goto _ctr1240;	goto _again;
					}
					else {
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr865;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr865;
							}
						} else {
							goto _ctr865;
						}
						goto _ctr1240;
					}
					case 736:
					if ( p == eof ) {
						goto _ctr864;	goto _again;
					}
					else {
						if ( ( (*( p))) == 125u ) {
							goto _ctr866;
						}
						if ( ( (*( p))) < 65u ) {
							if ( 48u <= ( (*( p))) && ( (*( p))) <= 57u ) {
								goto _ctr865;
							}
						} else if ( ( (*( p))) > 70u ) {
							if ( 97u <= ( (*( p))) && ( (*( p))) <= 102u ) {
								goto _ctr865;
							}
						} else {
							goto _ctr865;
						}
						goto _ctr864;
					}
					case 833:
					if ( p == eof ) {
						goto _ctr1241;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12582 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1243;
							} else {
								goto _ctr1242;
							}
							
						}
						goto _ctr1241;
					}
					case 834:
					if ( p == eof ) {
						goto _ctr1241;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1244;
							} else {
								goto _ctr1242;
							}
							
						}
						goto _ctr1241;
					}
					case 737:
					if ( p == eof ) {
						goto _ctr867;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12622 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr869;
							} else {
								goto _ctr868;
							}
							
						}
						goto _ctr867;
					}
					case 835:
					if ( p == eof ) {
						goto _ctr1241;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12642 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1245;
							} else {
								goto _ctr1242;
							}
							
						}
						goto _ctr1241;
					}
					case 738:
					if ( p == eof ) {
						goto _ctr867;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12662 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr870;
							} else {
								goto _ctr868;
							}
							
						}
						goto _ctr867;
					}
					case 739:
					if ( p == eof ) {
						goto _ctr867;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12682 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr871;
							} else {
								goto _ctr868;
							}
							
						}
						goto _ctr867;
					}
					case 836:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 12699 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1246;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 92u: {
								goto _ctr1248;
							}
							case 93u: {
								_ck = 0;
								if ( ( inCharClassEarly )
#line 12712 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1250;
								} else {
									goto _ctr1249;
								}
								
							}
							case 94u: {
								_ck = 0;
								if ( ( inCharClassEarly )
#line 12725 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1251;
								} else {
									goto _ctr1249;
								}
								
							}
						}
						goto _ctr1247;
					}
					case 837:
					if ( p == eof ) {
						goto _ctr1252;	goto _again;
					}
					else {
						switch( ( (*( p))) ) {
							case 69u: {
								goto _ctr1253;
							}
							case 81u: {
								goto _ctr1254;
							}
						}
						goto _ctr1252;
					}
					case 838:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 12758 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1255;	goto _again;
					}
					else {
						if ( ( (*( p))) == 92u ) {
							goto _ctr1257;
						}
						if ( ( (*( p))) < 224u ) {
							if ( ( (*( p))) > 191u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 12772 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1260;
									} else {
										goto _ctr1258;
									}
									
								}
							} else if ( ( (*( p))) >= 128u ) {
								_ck = 0;
								if ( ( mode.utf8 )
#line 12785 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1259;
								} else {
									goto _ctr1258;
								}
								
							}
						} else if ( ( (*( p))) > 239u ) {
							if ( ( (*( p))) > 247u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 12800 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1259;
									} else {
										goto _ctr1258;
									}
									
								}
							} else {
								_ck = 0;
								if ( ( mode.utf8 )
#line 12813 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1262;
								} else {
									goto _ctr1258;
								}
								
							}
						} else {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12826 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1261;
							} else {
								goto _ctr1258;
							}
							
						}
						goto _ctr1256;
					}
					case 839:
					if ( p == eof ) {
						goto _ctr1263;	goto _again;
					}
					else {
						if ( ( (*( p))) == 69u ) {
							goto _ctr1264;
						}
						goto _ctr1263;
					}
					case 840:
					if ( p == eof ) {
						goto _ctr1265;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12856 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1267;
							} else {
								goto _ctr1266;
							}
							
						}
						goto _ctr1265;
					}
					case 841:
					if ( p == eof ) {
						goto _ctr1265;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12876 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1268;
							} else {
								goto _ctr1266;
							}
							
						}
						goto _ctr1265;
					}
					case 740:
					if ( p == eof ) {
						goto _ctr872;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12896 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr874;
							} else {
								goto _ctr873;
							}
							
						}
						goto _ctr872;
					}
					case 842:
					if ( p == eof ) {
						goto _ctr1265;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12916 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1269;
							} else {
								goto _ctr1266;
							}
							
						}
						goto _ctr1265;
					}
					case 741:
					if ( p == eof ) {
						goto _ctr872;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12936 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr875;
							} else {
								goto _ctr873;
							}
							
						}
						goto _ctr872;
					}
					case 742:
					if ( p == eof ) {
						goto _ctr872;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 12956 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr876;
							} else {
								goto _ctr873;
							}
							
						}
						goto _ctr872;
					}
					case 843:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 12973 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1270;	goto _again;
					}
					else {
						if ( ( (*( p))) == 92u ) {
							goto _ctr1272;
						}
						if ( ( (*( p))) < 224u ) {
							if ( ( (*( p))) > 191u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 12987 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1275;
									} else {
										goto _ctr1273;
									}
									
								}
							} else if ( ( (*( p))) >= 128u ) {
								_ck = 0;
								if ( ( mode.utf8 )
#line 13000 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1274;
								} else {
									goto _ctr1273;
								}
								
							}
						} else if ( ( (*( p))) > 239u ) {
							if ( ( (*( p))) > 247u ) {
								{
									_ck = 0;
									if ( ( mode.utf8 )
#line 13015 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
									)
									_ck += 1;
									if ( _ck > 0 ) {
										goto _ctr1274;
									} else {
										goto _ctr1273;
									}
									
								}
							} else {
								_ck = 0;
								if ( ( mode.utf8 )
#line 13028 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
								)
								_ck += 1;
								if ( _ck > 0 ) {
									goto _ctr1277;
								} else {
									goto _ctr1273;
								}
								
							}
						} else {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13041 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1276;
							} else {
								goto _ctr1273;
							}
							
						}
						goto _ctr1271;
					}
					case 844:
					if ( p == eof ) {
						goto _ctr1278;	goto _again;
					}
					else {
						if ( ( (*( p))) == 69u ) {
							goto _ctr1279;
						}
						goto _ctr1278;
					}
					case 845:
					if ( p == eof ) {
						goto _ctr1280;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13071 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1282;
							} else {
								goto _ctr1281;
							}
							
						}
						goto _ctr1280;
					}
					case 846:
					if ( p == eof ) {
						goto _ctr1280;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13091 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1283;
							} else {
								goto _ctr1281;
							}
							
						}
						goto _ctr1280;
					}
					case 743:
					if ( p == eof ) {
						goto _ctr877;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13111 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr879;
							} else {
								goto _ctr878;
							}
							
						}
						goto _ctr877;
					}
					case 847:
					if ( p == eof ) {
						goto _ctr1280;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13131 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr1284;
							} else {
								goto _ctr1281;
							}
							
						}
						goto _ctr1280;
					}
					case 744:
					if ( p == eof ) {
						goto _ctr877;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13151 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr880;
							} else {
								goto _ctr878;
							}
							
						}
						goto _ctr877;
					}
					case 745:
					if ( p == eof ) {
						goto _ctr877;	goto _again;
					}
					else {
						if ( 128u <= ( (*( p))) && ( (*( p))) <= 191u ) {
							_ck = 0;
							if ( ( mode.utf8 )
#line 13171 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							)
							_ck += 1;
							if ( _ck > 0 ) {
								goto _ctr881;
							} else {
								goto _ctr878;
							}
							
						}
						goto _ctr877;
					}
					case 848:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 13188 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1285;	goto _again;
					}
					else {
						if ( ( (*( p))) == 41u ) {
							goto _ctr1287;
						}
						goto _ctr1286;
					}
					case 849:
					{
#line 1 "NONE"
						{ts = p;}}
					
#line 13204 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					if ( p == eof ) {
						goto _ctr1288;	goto _again;
					}
					else {
						if ( ( (*( p))) == 10u ) {
							goto _ctr1290;
						}
						goto _ctr1289;
					}
				}
				
				_ctr0: cs = 0; goto _again;
				_ctr652: cs = 0; goto f174;
				_ctr905: cs = 1; goto _again;
				_ctr906: cs = 2; goto _again;
				_ctr24: cs = 3; goto _again;
				_ctr26: cs = 4; goto f17;
				_ctr909: cs = 5; goto f237;
				_ctr28: cs = 6; goto f17;
				_ctr29: cs = 7; goto _again;
				_ctr31: cs = 8; goto f17;
				_ctr33: cs = 9; goto f21;
				_ctr910: cs = 9; goto f238;
				_ctr911: cs = 10; goto f238;
				_ctr912: cs = 11; goto f237;
				_ctr34: cs = 12; goto f17;
				_ctr913: cs = 13; goto _again;
				_ctr36: cs = 14; goto _again;
				_ctr914: cs = 15; goto f17;
				_ctr40: cs = 16; goto _again;
				_ctr43: cs = 17; goto _again;
				_ctr42: cs = 18; goto f17;
				_ctr45: cs = 19; goto _again;
				_ctr48: cs = 20; goto f30;
				_ctr49: cs = 21; goto f17;
				_ctr51: cs = 22; goto _again;
				_ctr918: cs = 23; goto f237;
				_ctr919: cs = 24; goto f237;
				_ctr52: cs = 25; goto f17;
				_ctr920: cs = 26; goto _again;
				_ctr54: cs = 27; goto _again;
				_ctr55: cs = 28; goto f33;
				_ctr57: cs = 29; goto _again;
				_ctr58: cs = 30; goto _again;
				_ctr59: cs = 31; goto _again;
				_ctr61: cs = 32; goto _again;
				_ctr62: cs = 33; goto _again;
				_ctr63: cs = 34; goto _again;
				_ctr65: cs = 35; goto _again;
				_ctr66: cs = 36; goto _again;
				_ctr67: cs = 37; goto _again;
				_ctr69: cs = 38; goto _again;
				_ctr987: cs = 39; goto f40;
				_ctr72: cs = 40; goto _again;
				_ctr989: cs = 41; goto f302;
				_ctr73: cs = 42; goto f40;
				_ctr76: cs = 43; goto f39;
				_ctr77: cs = 44; goto f39;
				_ctr79: cs = 45; goto f39;
				_ctr74: cs = 46; goto f21;
				_ctr80: cs = 47; goto f21;
				_ctr82: cs = 48; goto f21;
				_ctr75: cs = 49; goto f17;
				_ctr997: cs = 50; goto f237;
				_ctr85: cs = 51; goto f17;
				_ctr998: cs = 52; goto f237;
				_ctr87: cs = 53; goto f17;
				_ctr999: cs = 54; goto f237;
				_ctr89: cs = 55; goto f17;
				_ctr1001: cs = 56; goto _again;
				_ctr92: cs = 57; goto _again;
				_ctr95: cs = 58; goto _again;
				_ctr99: cs = 59; goto f53;
				_ctr98: cs = 60; goto _again;
				_ctr101: cs = 61; goto f55;
				_ctr110: cs = 65; goto _again;
				_ctr1028: cs = 66; goto _again;
				_ctr112: cs = 67; goto _again;
				_ctr1029: cs = 68; goto _again;
				_ctr114: cs = 69; goto _again;
				_ctr116: cs = 70; goto _again;
				_ctr1056: cs = 71; goto _again;
				_ctr1057: cs = 72; goto _again;
				_ctr120: cs = 73; goto _again;
				_ctr122: cs = 74; goto _again;
				_ctr123: cs = 75; goto _again;
				_ctr121: cs = 76; goto _again;
				_ctr125: cs = 77; goto _again;
				_ctr126: cs = 78; goto _again;
				_ctr127: cs = 79; goto _again;
				_ctr128: cs = 80; goto _again;
				_ctr1058: cs = 81; goto _again;
				_ctr130: cs = 82; goto _again;
				_ctr131: cs = 83; goto _again;
				_ctr132: cs = 84; goto _again;
				_ctr133: cs = 85; goto _again;
				_ctr1059: cs = 86; goto _again;
				_ctr135: cs = 87; goto _again;
				_ctr138: cs = 88; goto _again;
				_ctr139: cs = 89; goto _again;
				_ctr140: cs = 90; goto _again;
				_ctr141: cs = 91; goto _again;
				_ctr136: cs = 92; goto _again;
				_ctr143: cs = 93; goto _again;
				_ctr137: cs = 94; goto _again;
				_ctr145: cs = 95; goto _again;
				_ctr1060: cs = 96; goto _again;
				_ctr147: cs = 97; goto _again;
				_ctr148: cs = 98; goto _again;
				_ctr149: cs = 99; goto _again;
				_ctr150: cs = 100; goto _again;
				_ctr1061: cs = 101; goto _again;
				_ctr152: cs = 102; goto _again;
				_ctr153: cs = 103; goto _again;
				_ctr154: cs = 104; goto _again;
				_ctr155: cs = 105; goto _again;
				_ctr156: cs = 106; goto _again;
				_ctr1062: cs = 107; goto _again;
				_ctr158: cs = 108; goto _again;
				_ctr159: cs = 109; goto _again;
				_ctr161: cs = 110; goto _again;
				_ctr160: cs = 111; goto _again;
				_ctr163: cs = 112; goto _again;
				_ctr164: cs = 113; goto _again;
				_ctr1063: cs = 114; goto _again;
				_ctr166: cs = 115; goto _again;
				_ctr168: cs = 116; goto _again;
				_ctr169: cs = 117; goto _again;
				_ctr170: cs = 118; goto _again;
				_ctr171: cs = 119; goto _again;
				_ctr167: cs = 120; goto _again;
				_ctr173: cs = 121; goto _again;
				_ctr1065: cs = 122; goto _again;
				_ctr176: cs = 123; goto _again;
				_ctr178: cs = 124; goto _again;
				_ctr179: cs = 125; goto _again;
				_ctr180: cs = 126; goto _again;
				_ctr181: cs = 127; goto _again;
				_ctr182: cs = 128; goto _again;
				_ctr183: cs = 129; goto _again;
				_ctr184: cs = 130; goto _again;
				_ctr185: cs = 131; goto _again;
				_ctr186: cs = 132; goto _again;
				_ctr187: cs = 133; goto _again;
				_ctr188: cs = 134; goto _again;
				_ctr189: cs = 135; goto _again;
				_ctr190: cs = 136; goto _again;
				_ctr191: cs = 137; goto _again;
				_ctr192: cs = 138; goto _again;
				_ctr177: cs = 139; goto _again;
				_ctr194: cs = 140; goto _again;
				_ctr195: cs = 141; goto _again;
				_ctr1068: cs = 142; goto _again;
				_ctr197: cs = 143; goto _again;
				_ctr198: cs = 144; goto _again;
				_ctr200: cs = 145; goto _again;
				_ctr201: cs = 146; goto _again;
				_ctr202: cs = 147; goto _again;
				_ctr203: cs = 148; goto _again;
				_ctr1075: cs = 149; goto _again;
				_ctr206: cs = 150; goto _again;
				_ctr207: cs = 151; goto _again;
				_ctr1076: cs = 152; goto _again;
				_ctr209: cs = 153; goto _again;
				_ctr210: cs = 154; goto _again;
				_ctr1072: cs = 155; goto _again;
				_ctr212: cs = 156; goto _again;
				_ctr213: cs = 157; goto _again;
				_ctr214: cs = 158; goto _again;
				_ctr215: cs = 159; goto _again;
				_ctr216: cs = 160; goto _again;
				_ctr217: cs = 161; goto _again;
				_ctr1073: cs = 162; goto _again;
				_ctr219: cs = 163; goto _again;
				_ctr221: cs = 164; goto _again;
				_ctr222: cs = 165; goto _again;
				_ctr223: cs = 166; goto _again;
				_ctr220: cs = 167; goto _again;
				_ctr225: cs = 168; goto _again;
				_ctr226: cs = 169; goto _again;
				_ctr227: cs = 170; goto _again;
				_ctr228: cs = 171; goto _again;
				_ctr1077: cs = 172; goto _again;
				_ctr230: cs = 173; goto _again;
				_ctr232: cs = 174; goto _again;
				_ctr233: cs = 175; goto _again;
				_ctr234: cs = 176; goto _again;
				_ctr231: cs = 177; goto _again;
				_ctr236: cs = 178; goto _again;
				_ctr237: cs = 179; goto _again;
				_ctr238: cs = 180; goto _again;
				_ctr239: cs = 181; goto _again;
				_ctr240: cs = 182; goto _again;
				_ctr241: cs = 183; goto _again;
				_ctr1078: cs = 184; goto _again;
				_ctr243: cs = 185; goto _again;
				_ctr244: cs = 186; goto _again;
				_ctr245: cs = 187; goto _again;
				_ctr246: cs = 188; goto _again;
				_ctr247: cs = 189; goto _again;
				_ctr248: cs = 190; goto _again;
				_ctr249: cs = 191; goto _again;
				_ctr250: cs = 192; goto _again;
				_ctr251: cs = 193; goto _again;
				_ctr252: cs = 194; goto _again;
				_ctr253: cs = 195; goto _again;
				_ctr254: cs = 196; goto _again;
				_ctr255: cs = 197; goto _again;
				_ctr256: cs = 198; goto _again;
				_ctr257: cs = 199; goto _again;
				_ctr258: cs = 200; goto _again;
				_ctr259: cs = 201; goto _again;
				_ctr1079: cs = 202; goto _again;
				_ctr261: cs = 203; goto _again;
				_ctr262: cs = 204; goto _again;
				_ctr263: cs = 205; goto _again;
				_ctr264: cs = 206; goto _again;
				_ctr265: cs = 207; goto _again;
				_ctr1080: cs = 208; goto _again;
				_ctr267: cs = 209; goto _again;
				_ctr268: cs = 210; goto _again;
				_ctr269: cs = 211; goto _again;
				_ctr270: cs = 212; goto _again;
				_ctr271: cs = 213; goto _again;
				_ctr1081: cs = 214; goto _again;
				_ctr273: cs = 215; goto _again;
				_ctr274: cs = 216; goto _again;
				_ctr275: cs = 217; goto _again;
				_ctr276: cs = 218; goto _again;
				_ctr277: cs = 219; goto _again;
				_ctr278: cs = 220; goto _again;
				_ctr279: cs = 221; goto _again;
				_ctr1082: cs = 222; goto _again;
				_ctr281: cs = 223; goto _again;
				_ctr282: cs = 224; goto _again;
				_ctr283: cs = 225; goto _again;
				_ctr1083: cs = 226; goto _again;
				_ctr285: cs = 227; goto _again;
				_ctr286: cs = 228; goto _again;
				_ctr1084: cs = 229; goto _again;
				_ctr288: cs = 230; goto _again;
				_ctr290: cs = 231; goto _again;
				_ctr291: cs = 232; goto _again;
				_ctr292: cs = 233; goto _again;
				_ctr293: cs = 234; goto _again;
				_ctr289: cs = 235; goto _again;
				_ctr295: cs = 236; goto _again;
				_ctr296: cs = 237; goto _again;
				_ctr297: cs = 238; goto _again;
				_ctr298: cs = 239; goto _again;
				_ctr1085: cs = 240; goto _again;
				_ctr1089: cs = 241; goto _again;
				_ctr302: cs = 242; goto _again;
				_ctr1090: cs = 243; goto _again;
				_ctr304: cs = 244; goto _again;
				_ctr305: cs = 245; goto _again;
				_ctr1086: cs = 246; goto _again;
				_ctr307: cs = 247; goto _again;
				_ctr308: cs = 248; goto _again;
				_ctr309: cs = 249; goto _again;
				_ctr1087: cs = 250; goto _again;
				_ctr311: cs = 251; goto _again;
				_ctr312: cs = 252; goto _again;
				_ctr313: cs = 253; goto _again;
				_ctr314: cs = 254; goto _again;
				_ctr315: cs = 255; goto _again;
				_ctr1091: cs = 256; goto _again;
				_ctr317: cs = 257; goto _again;
				_ctr318: cs = 258; goto _again;
				_ctr319: cs = 259; goto _again;
				_ctr320: cs = 260; goto _again;
				_ctr321: cs = 261; goto _again;
				_ctr322: cs = 262; goto _again;
				_ctr323: cs = 263; goto _again;
				_ctr324: cs = 264; goto _again;
				_ctr325: cs = 265; goto _again;
				_ctr326: cs = 266; goto _again;
				_ctr327: cs = 267; goto _again;
				_ctr328: cs = 268; goto _again;
				_ctr329: cs = 269; goto _again;
				_ctr1092: cs = 270; goto _again;
				_ctr331: cs = 271; goto _again;
				_ctr333: cs = 272; goto _again;
				_ctr334: cs = 273; goto _again;
				_ctr335: cs = 274; goto _again;
				_ctr336: cs = 275; goto _again;
				_ctr337: cs = 276; goto _again;
				_ctr332: cs = 277; goto _again;
				_ctr339: cs = 278; goto _again;
				_ctr340: cs = 279; goto _again;
				_ctr341: cs = 280; goto _again;
				_ctr342: cs = 281; goto _again;
				_ctr343: cs = 282; goto _again;
				_ctr344: cs = 283; goto _again;
				_ctr345: cs = 284; goto _again;
				_ctr346: cs = 285; goto _again;
				_ctr347: cs = 286; goto _again;
				_ctr348: cs = 287; goto _again;
				_ctr349: cs = 288; goto _again;
				_ctr350: cs = 289; goto _again;
				_ctr351: cs = 290; goto _again;
				_ctr352: cs = 291; goto _again;
				_ctr354: cs = 292; goto _again;
				_ctr355: cs = 293; goto _again;
				_ctr356: cs = 294; goto _again;
				_ctr353: cs = 295; goto _again;
				_ctr358: cs = 296; goto _again;
				_ctr359: cs = 297; goto _again;
				_ctr360: cs = 298; goto _again;
				_ctr361: cs = 299; goto _again;
				_ctr1093: cs = 300; goto _again;
				_ctr363: cs = 301; goto _again;
				_ctr364: cs = 302; goto _again;
				_ctr365: cs = 303; goto _again;
				_ctr366: cs = 304; goto _again;
				_ctr367: cs = 305; goto _again;
				_ctr1094: cs = 306; goto _again;
				_ctr369: cs = 307; goto _again;
				_ctr373: cs = 308; goto _again;
				_ctr374: cs = 309; goto _again;
				_ctr370: cs = 310; goto _again;
				_ctr376: cs = 311; goto _again;
				_ctr377: cs = 312; goto _again;
				_ctr378: cs = 313; goto _again;
				_ctr371: cs = 314; goto _again;
				_ctr380: cs = 315; goto _again;
				_ctr381: cs = 316; goto _again;
				_ctr382: cs = 317; goto _again;
				_ctr383: cs = 318; goto _again;
				_ctr372: cs = 319; goto _again;
				_ctr385: cs = 320; goto _again;
				_ctr386: cs = 321; goto _again;
				_ctr387: cs = 322; goto _again;
				_ctr388: cs = 323; goto _again;
				_ctr1095: cs = 324; goto _again;
				_ctr390: cs = 325; goto _again;
				_ctr392: cs = 326; goto _again;
				_ctr393: cs = 327; goto _again;
				_ctr394: cs = 328; goto _again;
				_ctr395: cs = 329; goto _again;
				_ctr396: cs = 330; goto _again;
				_ctr397: cs = 331; goto _again;
				_ctr391: cs = 332; goto _again;
				_ctr399: cs = 333; goto _again;
				_ctr1098: cs = 334; goto _again;
				_ctr403: cs = 335; goto _again;
				_ctr404: cs = 336; goto _again;
				_ctr1099: cs = 337; goto _again;
				_ctr406: cs = 338; goto _again;
				_ctr407: cs = 339; goto _again;
				_ctr408: cs = 340; goto _again;
				_ctr1100: cs = 341; goto _again;
				_ctr410: cs = 342; goto _again;
				_ctr413: cs = 343; goto _again;
				_ctr411: cs = 344; goto _again;
				_ctr415: cs = 345; goto _again;
				_ctr416: cs = 346; goto _again;
				_ctr417: cs = 347; goto _again;
				_ctr418: cs = 348; goto _again;
				_ctr412: cs = 349; goto _again;
				_ctr1106: cs = 350; goto _again;
				_ctr421: cs = 351; goto _again;
				_ctr423: cs = 352; goto _again;
				_ctr424: cs = 353; goto _again;
				_ctr422: cs = 354; goto _again;
				_ctr426: cs = 355; goto _again;
				_ctr427: cs = 356; goto _again;
				_ctr1108: cs = 357; goto _again;
				_ctr430: cs = 358; goto _again;
				_ctr432: cs = 359; goto _again;
				_ctr433: cs = 360; goto _again;
				_ctr434: cs = 361; goto _again;
				_ctr435: cs = 362; goto _again;
				_ctr436: cs = 363; goto _again;
				_ctr431: cs = 364; goto _again;
				_ctr438: cs = 365; goto _again;
				_ctr439: cs = 366; goto _again;
				_ctr440: cs = 367; goto _again;
				_ctr1115: cs = 368; goto _again;
				_ctr443: cs = 369; goto _again;
				_ctr444: cs = 370; goto _again;
				_ctr445: cs = 371; goto _again;
				_ctr446: cs = 372; goto _again;
				_ctr447: cs = 373; goto _again;
				_ctr448: cs = 374; goto _again;
				_ctr449: cs = 375; goto _again;
				_ctr450: cs = 376; goto _again;
				_ctr1112: cs = 377; goto _again;
				_ctr452: cs = 378; goto _again;
				_ctr453: cs = 379; goto _again;
				_ctr454: cs = 380; goto _again;
				_ctr455: cs = 381; goto _again;
				_ctr456: cs = 382; goto _again;
				_ctr457: cs = 383; goto _again;
				_ctr1113: cs = 384; goto _again;
				_ctr459: cs = 385; goto _again;
				_ctr460: cs = 386; goto _again;
				_ctr461: cs = 387; goto _again;
				_ctr462: cs = 388; goto _again;
				_ctr1118: cs = 389; goto _again;
				_ctr465: cs = 390; goto _again;
				_ctr466: cs = 391; goto _again;
				_ctr467: cs = 392; goto _again;
				_ctr468: cs = 393; goto _again;
				_ctr469: cs = 394; goto _again;
				_ctr470: cs = 395; goto _again;
				_ctr471: cs = 396; goto _again;
				_ctr472: cs = 397; goto _again;
				_ctr1119: cs = 398; goto _again;
				_ctr1122: cs = 399; goto _again;
				_ctr475: cs = 400; goto _again;
				_ctr476: cs = 401; goto _again;
				_ctr1123: cs = 402; goto _again;
				_ctr478: cs = 403; goto _again;
				_ctr480: cs = 404; goto _again;
				_ctr481: cs = 405; goto _again;
				_ctr482: cs = 406; goto _again;
				_ctr483: cs = 407; goto _again;
				_ctr479: cs = 408; goto _again;
				_ctr485: cs = 409; goto _again;
				_ctr486: cs = 410; goto _again;
				_ctr490: cs = 411; goto _again;
				_ctr491: cs = 412; goto _again;
				_ctr492: cs = 413; goto _again;
				_ctr493: cs = 414; goto _again;
				_ctr487: cs = 415; goto _again;
				_ctr495: cs = 416; goto _again;
				_ctr496: cs = 417; goto _again;
				_ctr497: cs = 418; goto _again;
				_ctr498: cs = 419; goto _again;
				_ctr499: cs = 420; goto _again;
				_ctr488: cs = 421; goto _again;
				_ctr501: cs = 422; goto _again;
				_ctr502: cs = 423; goto _again;
				_ctr503: cs = 424; goto _again;
				_ctr504: cs = 425; goto _again;
				_ctr505: cs = 426; goto _again;
				_ctr506: cs = 427; goto _again;
				_ctr507: cs = 428; goto _again;
				_ctr508: cs = 429; goto _again;
				_ctr509: cs = 430; goto _again;
				_ctr510: cs = 431; goto _again;
				_ctr511: cs = 432; goto _again;
				_ctr489: cs = 433; goto _again;
				_ctr513: cs = 434; goto _again;
				_ctr514: cs = 435; goto _again;
				_ctr515: cs = 436; goto _again;
				_ctr516: cs = 437; goto _again;
				_ctr1124: cs = 438; goto _again;
				_ctr518: cs = 439; goto _again;
				_ctr519: cs = 440; goto _again;
				_ctr1125: cs = 441; goto _again;
				_ctr521: cs = 442; goto _again;
				_ctr522: cs = 443; goto _again;
				_ctr523: cs = 444; goto _again;
				_ctr524: cs = 445; goto _again;
				_ctr1131: cs = 446; goto _again;
				_ctr527: cs = 447; goto _again;
				_ctr529: cs = 448; goto _again;
				_ctr530: cs = 449; goto _again;
				_ctr531: cs = 450; goto _again;
				_ctr532: cs = 451; goto _again;
				_ctr528: cs = 452; goto _again;
				_ctr534: cs = 453; goto _again;
				_ctr535: cs = 454; goto _again;
				_ctr536: cs = 455; goto _again;
				_ctr537: cs = 456; goto _again;
				_ctr538: cs = 457; goto _again;
				_ctr539: cs = 458; goto _again;
				_ctr1135: cs = 459; goto _again;
				_ctr541: cs = 460; goto _again;
				_ctr542: cs = 461; goto _again;
				_ctr543: cs = 462; goto _again;
				_ctr1136: cs = 463; goto _again;
				_ctr545: cs = 464; goto _again;
				_ctr546: cs = 465; goto _again;
				_ctr1138: cs = 466; goto _again;
				_ctr549: cs = 467; goto _again;
				_ctr551: cs = 468; goto _again;
				_ctr552: cs = 469; goto _again;
				_ctr553: cs = 470; goto _again;
				_ctr554: cs = 471; goto _again;
				_ctr555: cs = 472; goto _again;
				_ctr550: cs = 473; goto _again;
				_ctr557: cs = 474; goto _again;
				_ctr558: cs = 475; goto _again;
				_ctr559: cs = 476; goto _again;
				_ctr560: cs = 477; goto _again;
				_ctr561: cs = 478; goto _again;
				_ctr562: cs = 479; goto _again;
				_ctr1140: cs = 480; goto _again;
				_ctr564: cs = 481; goto _again;
				_ctr565: cs = 482; goto _again;
				_ctr566: cs = 483; goto _again;
				_ctr567: cs = 484; goto _again;
				_ctr1141: cs = 485; goto _again;
				_ctr569: cs = 486; goto _again;
				_ctr570: cs = 487; goto _again;
				_ctr571: cs = 488; goto _again;
				_ctr572: cs = 489; goto _again;
				_ctr1145: cs = 490; goto _again;
				_ctr574: cs = 491; goto _again;
				_ctr575: cs = 492; goto _again;
				_ctr576: cs = 493; goto _again;
				_ctr577: cs = 494; goto _again;
				_ctr578: cs = 495; goto _again;
				_ctr579: cs = 496; goto _again;
				_ctr1146: cs = 497; goto _again;
				_ctr581: cs = 498; goto _again;
				_ctr583: cs = 499; goto _again;
				_ctr584: cs = 500; goto _again;
				_ctr585: cs = 501; goto _again;
				_ctr586: cs = 502; goto _again;
				_ctr587: cs = 503; goto _again;
				_ctr588: cs = 504; goto _again;
				_ctr589: cs = 505; goto _again;
				_ctr590: cs = 506; goto _again;
				_ctr582: cs = 507; goto _again;
				_ctr592: cs = 508; goto _again;
				_ctr593: cs = 509; goto _again;
				_ctr1147: cs = 510; goto _again;
				_ctr595: cs = 511; goto _again;
				_ctr598: cs = 512; goto _again;
				_ctr600: cs = 513; goto _again;
				_ctr601: cs = 514; goto _again;
				_ctr599: cs = 515; goto _again;
				_ctr603: cs = 516; goto _again;
				_ctr604: cs = 517; goto _again;
				_ctr605: cs = 518; goto _again;
				_ctr596: cs = 519; goto _again;
				_ctr607: cs = 520; goto _again;
				_ctr608: cs = 521; goto _again;
				_ctr609: cs = 522; goto _again;
				_ctr612: cs = 523; goto _again;
				_ctr613: cs = 524; goto _again;
				_ctr610: cs = 525; goto _again;
				_ctr615: cs = 526; goto _again;
				_ctr616: cs = 527; goto _again;
				_ctr597: cs = 528; goto _again;
				_ctr618: cs = 529; goto _again;
				_ctr1148: cs = 530; goto _again;
				_ctr620: cs = 531; goto _again;
				_ctr621: cs = 532; goto _again;
				_ctr622: cs = 533; goto _again;
				_ctr1149: cs = 534; goto _again;
				_ctr624: cs = 535; goto _again;
				_ctr625: cs = 536; goto _again;
				_ctr627: cs = 537; goto _again;
				_ctr1150: cs = 538; goto _again;
				_ctr629: cs = 539; goto _again;
				_ctr631: cs = 540; goto _again;
				_ctr632: cs = 541; goto _again;
				_ctr633: cs = 542; goto _again;
				_ctr630: cs = 543; goto _again;
				_ctr635: cs = 544; goto _again;
				_ctr636: cs = 545; goto _again;
				_ctr637: cs = 546; goto _again;
				_ctr638: cs = 547; goto _again;
				_ctr1151: cs = 548; goto _again;
				_ctr640: cs = 549; goto _again;
				_ctr641: cs = 550; goto _again;
				_ctr642: cs = 551; goto _again;
				_ctr643: cs = 552; goto _again;
				_ctr644: cs = 553; goto _again;
				_ctr1152: cs = 554; goto _again;
				_ctr1153: cs = 555; goto _again;
				_ctr1154: cs = 556; goto _again;
				_ctr1155: cs = 557; goto _again;
				_ctr1156: cs = 558; goto _again;
				_ctr651: cs = 559; goto f174;
				_ctr653: cs = 560; goto _again;
				_ctr654: cs = 560; goto f174;
				_ctr657: cs = 561; goto f174;
				_ctr655: cs = 561; goto f175;
				_ctr659: cs = 562; goto f174;
				_ctr656: cs = 562; goto f176;
				_ctr661: cs = 563; goto _again;
				_ctr662: cs = 564; goto _again;
				_ctr663: cs = 565; goto _again;
				_ctr1184: cs = 566; goto _again;
				_ctr665: cs = 567; goto _again;
				_ctr666: cs = 568; goto _again;
				_ctr667: cs = 569; goto _again;
				_ctr668: cs = 570; goto _again;
				_ctr681: cs = 571; goto _again;
				_ctr692: cs = 572; goto _again;
				_ctr694: cs = 573; goto _again;
				_ctr696: cs = 574; goto _again;
				_ctr697: cs = 575; goto _again;
				_ctr698: cs = 576; goto _again;
				_ctr695: cs = 577; goto _again;
				_ctr700: cs = 578; goto _again;
				_ctr701: cs = 579; goto _again;
				_ctr702: cs = 580; goto _again;
				_ctr693: cs = 581; goto _again;
				_ctr704: cs = 582; goto _again;
				_ctr705: cs = 583; goto _again;
				_ctr706: cs = 584; goto _again;
				_ctr707: cs = 585; goto _again;
				_ctr682: cs = 586; goto _again;
				_ctr709: cs = 587; goto _again;
				_ctr710: cs = 588; goto _again;
				_ctr711: cs = 589; goto _again;
				_ctr712: cs = 590; goto _again;
				_ctr713: cs = 591; goto _again;
				_ctr683: cs = 592; goto _again;
				_ctr715: cs = 593; goto _again;
				_ctr716: cs = 594; goto _again;
				_ctr717: cs = 595; goto _again;
				_ctr718: cs = 596; goto _again;
				_ctr719: cs = 597; goto _again;
				_ctr684: cs = 598; goto _again;
				_ctr721: cs = 599; goto _again;
				_ctr722: cs = 600; goto _again;
				_ctr723: cs = 601; goto _again;
				_ctr724: cs = 602; goto _again;
				_ctr725: cs = 603; goto _again;
				_ctr685: cs = 604; goto _again;
				_ctr727: cs = 605; goto _again;
				_ctr728: cs = 606; goto _again;
				_ctr729: cs = 607; goto _again;
				_ctr730: cs = 608; goto _again;
				_ctr731: cs = 609; goto _again;
				_ctr686: cs = 610; goto _again;
				_ctr733: cs = 611; goto _again;
				_ctr734: cs = 612; goto _again;
				_ctr735: cs = 613; goto _again;
				_ctr736: cs = 614; goto _again;
				_ctr737: cs = 615; goto _again;
				_ctr687: cs = 616; goto _again;
				_ctr739: cs = 617; goto _again;
				_ctr741: cs = 618; goto _again;
				_ctr742: cs = 619; goto _again;
				_ctr743: cs = 620; goto _again;
				_ctr744: cs = 621; goto _again;
				_ctr740: cs = 622; goto _again;
				_ctr746: cs = 623; goto _again;
				_ctr747: cs = 624; goto _again;
				_ctr748: cs = 625; goto _again;
				_ctr749: cs = 626; goto _again;
				_ctr688: cs = 627; goto _again;
				_ctr751: cs = 628; goto _again;
				_ctr752: cs = 629; goto _again;
				_ctr753: cs = 630; goto _again;
				_ctr754: cs = 631; goto _again;
				_ctr755: cs = 632; goto _again;
				_ctr689: cs = 633; goto _again;
				_ctr757: cs = 634; goto _again;
				_ctr758: cs = 635; goto _again;
				_ctr759: cs = 636; goto _again;
				_ctr760: cs = 637; goto _again;
				_ctr761: cs = 638; goto _again;
				_ctr690: cs = 639; goto _again;
				_ctr763: cs = 640; goto _again;
				_ctr764: cs = 641; goto _again;
				_ctr765: cs = 642; goto _again;
				_ctr766: cs = 643; goto _again;
				_ctr691: cs = 644; goto _again;
				_ctr768: cs = 645; goto _again;
				_ctr769: cs = 646; goto _again;
				_ctr770: cs = 647; goto _again;
				_ctr771: cs = 648; goto _again;
				_ctr772: cs = 649; goto _again;
				_ctr773: cs = 650; goto _again;
				_ctr669: cs = 651; goto _again;
				_ctr775: cs = 652; goto _again;
				_ctr777: cs = 653; goto _again;
				_ctr779: cs = 654; goto _again;
				_ctr780: cs = 655; goto _again;
				_ctr781: cs = 656; goto _again;
				_ctr778: cs = 657; goto _again;
				_ctr783: cs = 658; goto _again;
				_ctr784: cs = 659; goto _again;
				_ctr785: cs = 660; goto _again;
				_ctr776: cs = 661; goto _again;
				_ctr787: cs = 662; goto _again;
				_ctr788: cs = 663; goto _again;
				_ctr789: cs = 664; goto _again;
				_ctr790: cs = 665; goto _again;
				_ctr670: cs = 666; goto _again;
				_ctr792: cs = 667; goto _again;
				_ctr793: cs = 668; goto _again;
				_ctr794: cs = 669; goto _again;
				_ctr795: cs = 670; goto _again;
				_ctr796: cs = 671; goto _again;
				_ctr671: cs = 672; goto _again;
				_ctr798: cs = 673; goto _again;
				_ctr799: cs = 674; goto _again;
				_ctr800: cs = 675; goto _again;
				_ctr801: cs = 676; goto _again;
				_ctr802: cs = 677; goto _again;
				_ctr672: cs = 678; goto _again;
				_ctr804: cs = 679; goto _again;
				_ctr805: cs = 680; goto _again;
				_ctr806: cs = 681; goto _again;
				_ctr807: cs = 682; goto _again;
				_ctr808: cs = 683; goto _again;
				_ctr673: cs = 684; goto _again;
				_ctr810: cs = 685; goto _again;
				_ctr811: cs = 686; goto _again;
				_ctr812: cs = 687; goto _again;
				_ctr813: cs = 688; goto _again;
				_ctr814: cs = 689; goto _again;
				_ctr674: cs = 690; goto _again;
				_ctr816: cs = 691; goto _again;
				_ctr817: cs = 692; goto _again;
				_ctr818: cs = 693; goto _again;
				_ctr819: cs = 694; goto _again;
				_ctr820: cs = 695; goto _again;
				_ctr675: cs = 696; goto _again;
				_ctr822: cs = 697; goto _again;
				_ctr824: cs = 698; goto _again;
				_ctr825: cs = 699; goto _again;
				_ctr826: cs = 700; goto _again;
				_ctr827: cs = 701; goto _again;
				_ctr823: cs = 702; goto _again;
				_ctr829: cs = 703; goto _again;
				_ctr830: cs = 704; goto _again;
				_ctr831: cs = 705; goto _again;
				_ctr832: cs = 706; goto _again;
				_ctr676: cs = 707; goto _again;
				_ctr834: cs = 708; goto _again;
				_ctr835: cs = 709; goto _again;
				_ctr836: cs = 710; goto _again;
				_ctr837: cs = 711; goto _again;
				_ctr838: cs = 712; goto _again;
				_ctr677: cs = 713; goto _again;
				_ctr840: cs = 714; goto _again;
				_ctr841: cs = 715; goto _again;
				_ctr842: cs = 716; goto _again;
				_ctr843: cs = 717; goto _again;
				_ctr844: cs = 718; goto _again;
				_ctr678: cs = 719; goto _again;
				_ctr846: cs = 720; goto _again;
				_ctr847: cs = 721; goto _again;
				_ctr848: cs = 722; goto _again;
				_ctr849: cs = 723; goto _again;
				_ctr679: cs = 724; goto _again;
				_ctr851: cs = 725; goto _again;
				_ctr852: cs = 726; goto _again;
				_ctr853: cs = 727; goto _again;
				_ctr854: cs = 728; goto _again;
				_ctr855: cs = 729; goto _again;
				_ctr856: cs = 730; goto _again;
				_ctr858: cs = 731; goto _again;
				_ctr859: cs = 732; goto _again;
				_ctr860: cs = 733; goto _again;
				_ctr1228: cs = 734; goto _again;
				_ctr862: cs = 735; goto _again;
				_ctr865: cs = 736; goto _again;
				_ctr882: cs = 746; goto _again;
				_ctr1: cs = 746; goto f0;
				_ctr2: cs = 746; goto f1;
				_ctr3: cs = 746; goto f2;
				_ctr4: cs = 746; goto f3;
				_ctr5: cs = 746; goto f4;
				_ctr9: cs = 746; goto f7;
				_ctr13: cs = 746; goto f9;
				_ctr15: cs = 746; goto f10;
				_ctr16: cs = 746; goto f11;
				_ctr21: cs = 746; goto f13;
				_ctr22: cs = 746; goto f14;
				_ctr23: cs = 746; goto f15;
				_ctr25: cs = 746; goto f16;
				_ctr27: cs = 746; goto f18;
				_ctr30: cs = 746; goto f19;
				_ctr32: cs = 746; goto f20;
				_ctr35: cs = 746; goto f22;
				_ctr37: cs = 746; goto f23;
				_ctr38: cs = 746; goto f24;
				_ctr39: cs = 746; goto f25;
				_ctr41: cs = 746; goto f26;
				_ctr44: cs = 746; goto f27;
				_ctr46: cs = 746; goto f28;
				_ctr47: cs = 746; goto f29;
				_ctr50: cs = 746; goto f31;
				_ctr53: cs = 746; goto f32;
				_ctr56: cs = 746; goto f34;
				_ctr60: cs = 746; goto f35;
				_ctr64: cs = 746; goto f36;
				_ctr68: cs = 746; goto f37;
				_ctr70: cs = 746; goto f38;
				_ctr78: cs = 746; goto f41;
				_ctr81: cs = 746; goto f42;
				_ctr83: cs = 746; goto f43;
				_ctr84: cs = 746; goto f44;
				_ctr86: cs = 746; goto f45;
				_ctr88: cs = 746; goto f46;
				_ctr90: cs = 746; goto f47;
				_ctr91: cs = 746; goto f48;
				_ctr93: cs = 746; goto f49;
				_ctr94: cs = 746; goto f50;
				_ctr96: cs = 746; goto f51;
				_ctr97: cs = 746; goto f52;
				_ctr104: cs = 746; goto f57;
				_ctr883: cs = 746; goto f223;
				_ctr884: cs = 746; goto f224;
				_ctr885: cs = 746; goto f225;
				_ctr888: cs = 746; goto f227;
				_ctr890: cs = 746; goto f228;
				_ctr893: cs = 746; goto f229;
				_ctr897: cs = 746; goto f230;
				_ctr899: cs = 746; goto f232;
				_ctr904: cs = 746; goto f234;
				_ctr907: cs = 746; goto f235;
				_ctr908: cs = 746; goto f236;
				_ctr915: cs = 746; goto f239;
				_ctr916: cs = 746; goto f240;
				_ctr917: cs = 746; goto f241;
				_ctr921: cs = 746; goto f242;
				_ctr922: cs = 746; goto f243;
				_ctr923: cs = 746; goto f244;
				_ctr924: cs = 746; goto f245;
				_ctr925: cs = 746; goto f246;
				_ctr926: cs = 746; goto f247;
				_ctr927: cs = 746; goto f248;
				_ctr928: cs = 746; goto f249;
				_ctr929: cs = 746; goto f250;
				_ctr930: cs = 746; goto f251;
				_ctr931: cs = 746; goto f252;
				_ctr932: cs = 746; goto f253;
				_ctr936: cs = 746; goto f257;
				_ctr937: cs = 746; goto f258;
				_ctr938: cs = 746; goto f259;
				_ctr939: cs = 746; goto f260;
				_ctr940: cs = 746; goto f261;
				_ctr941: cs = 746; goto f262;
				_ctr942: cs = 746; goto f263;
				_ctr943: cs = 746; goto f264;
				_ctr944: cs = 746; goto f265;
				_ctr946: cs = 746; goto f266;
				_ctr947: cs = 746; goto f267;
				_ctr948: cs = 746; goto f268;
				_ctr949: cs = 746; goto f269;
				_ctr950: cs = 746; goto f270;
				_ctr951: cs = 746; goto f271;
				_ctr952: cs = 746; goto f272;
				_ctr953: cs = 746; goto f273;
				_ctr954: cs = 746; goto f274;
				_ctr956: cs = 746; goto f275;
				_ctr957: cs = 746; goto f276;
				_ctr958: cs = 746; goto f277;
				_ctr960: cs = 746; goto f279;
				_ctr962: cs = 746; goto f280;
				_ctr965: cs = 746; goto f281;
				_ctr966: cs = 746; goto f282;
				_ctr967: cs = 746; goto f283;
				_ctr968: cs = 746; goto f284;
				_ctr969: cs = 746; goto f285;
				_ctr971: cs = 746; goto f286;
				_ctr972: cs = 746; goto f287;
				_ctr974: cs = 746; goto f289;
				_ctr975: cs = 746; goto f290;
				_ctr978: cs = 746; goto f293;
				_ctr980: cs = 746; goto f295;
				_ctr981: cs = 746; goto f296;
				_ctr982: cs = 746; goto f297;
				_ctr983: cs = 746; goto f298;
				_ctr984: cs = 746; goto f299;
				_ctr985: cs = 746; goto f300;
				_ctr986: cs = 746; goto f301;
				_ctr990: cs = 746; goto f303;
				_ctr992: cs = 746; goto f304;
				_ctr993: cs = 746; goto f305;
				_ctr995: cs = 746; goto f306;
				_ctr996: cs = 746; goto f307;
				_ctr1000: cs = 746; goto f308;
				_ctr1002: cs = 746; goto f309;
				_ctr1003: cs = 746; goto f310;
				_ctr1004: cs = 746; goto f311;
				_ctr1005: cs = 746; goto f312;
				_ctr1010: cs = 746; goto f316;
				_ctr1011: cs = 746; goto f317;
				_ctr1012: cs = 746; goto f318;
				_ctr1013: cs = 746; goto f319;
				_ctr1014: cs = 746; goto f320;
				_ctr1015: cs = 746; goto f321;
				_ctr1016: cs = 746; goto f322;
				_ctr1017: cs = 746; goto f323;
				_ctr1018: cs = 746; goto f324;
				_ctr889: cs = 747; goto f5;
				_ctr6: cs = 748; goto f5;
				_ctr7: cs = 749; goto f6;
				_ctr8: cs = 750; goto f6;
				_ctr10: cs = 751; goto f5;
				_ctr11: cs = 752; goto f8;
				_ctr12: cs = 753; goto f5;
				_ctr14: cs = 754; goto f6;
				_ctr17: cs = 755; goto _again;
				_ctr18: cs = 756; goto f5;
				_ctr19: cs = 757; goto f5;
				_ctr20: cs = 758; goto f12;
				_ctr891: cs = 759; goto _again;
				_ctr892: cs = 760; goto _again;
				_ctr894: cs = 761; goto _again;
				_ctr895: cs = 762; goto f5;
				_ctr896: cs = 763; goto _again;
				_ctr933: cs = 764; goto f254;
				_ctr973: cs = 765; goto f288;
				_ctr934: cs = 766; goto f255;
				_ctr976: cs = 767; goto f291;
				_ctr935: cs = 768; goto f256;
				_ctr977: cs = 768; goto f292;
				_ctr979: cs = 768; goto f294;
				_ctr945: cs = 769; goto _again;
				_ctr955: cs = 770; goto _again;
				_ctr959: cs = 771; goto f278;
				_ctr71: cs = 772; goto f39;
				_ctr991: cs = 773; goto f39;
				_ctr988: cs = 774; goto f39;
				_ctr994: cs = 775; goto f39;
				_ctr961: cs = 776; goto f5;
				_ctr963: cs = 777; goto f5;
				_ctr964: cs = 778; goto _again;
				_ctr970: cs = 779; goto f40;
				_ctr1006: cs = 780; goto f313;
				_ctr1007: cs = 780; goto f314;
				_ctr1008: cs = 780; goto f315;
				_ctr1009: cs = 781; goto f5;
				_ctr898: cs = 782; goto f231;
				_ctr103: cs = 783; goto _again;
				_ctr100: cs = 783; goto f54;
				_ctr102: cs = 783; goto f56;
				_ctr1023: cs = 787; goto _again;
				_ctr109: cs = 787; goto f60;
				_ctr111: cs = 787; goto f61;
				_ctr113: cs = 787; goto f62;
				_ctr115: cs = 787; goto f63;
				_ctr117: cs = 787; goto f64;
				_ctr1025: cs = 787; goto f326;
				_ctr1027: cs = 787; goto f327;
				_ctr1024: cs = 788; goto f5;
				_ctr1026: cs = 789; goto f5;
				_ctr1030: cs = 790; goto _again;
				_ctr118: cs = 790; goto f65;
				_ctr119: cs = 790; goto f66;
				_ctr124: cs = 790; goto f67;
				_ctr129: cs = 790; goto f68;
				_ctr134: cs = 790; goto f69;
				_ctr142: cs = 790; goto f70;
				_ctr144: cs = 790; goto f71;
				_ctr146: cs = 790; goto f72;
				_ctr151: cs = 790; goto f73;
				_ctr157: cs = 790; goto f74;
				_ctr162: cs = 790; goto f75;
				_ctr165: cs = 790; goto f76;
				_ctr172: cs = 790; goto f77;
				_ctr174: cs = 790; goto f78;
				_ctr175: cs = 790; goto f79;
				_ctr193: cs = 790; goto f80;
				_ctr196: cs = 790; goto f81;
				_ctr199: cs = 790; goto f82;
				_ctr204: cs = 790; goto f83;
				_ctr205: cs = 790; goto f84;
				_ctr208: cs = 790; goto f85;
				_ctr211: cs = 790; goto f86;
				_ctr218: cs = 790; goto f87;
				_ctr224: cs = 790; goto f88;
				_ctr229: cs = 790; goto f89;
				_ctr235: cs = 790; goto f90;
				_ctr242: cs = 790; goto f91;
				_ctr260: cs = 790; goto f92;
				_ctr266: cs = 790; goto f93;
				_ctr272: cs = 790; goto f94;
				_ctr280: cs = 790; goto f95;
				_ctr284: cs = 790; goto f96;
				_ctr287: cs = 790; goto f97;
				_ctr294: cs = 790; goto f98;
				_ctr299: cs = 790; goto f99;
				_ctr301: cs = 790; goto f100;
				_ctr303: cs = 790; goto f101;
				_ctr306: cs = 790; goto f102;
				_ctr310: cs = 790; goto f103;
				_ctr316: cs = 790; goto f104;
				_ctr330: cs = 790; goto f105;
				_ctr338: cs = 790; goto f106;
				_ctr357: cs = 790; goto f107;
				_ctr362: cs = 790; goto f108;
				_ctr368: cs = 790; goto f109;
				_ctr375: cs = 790; goto f110;
				_ctr379: cs = 790; goto f111;
				_ctr384: cs = 790; goto f112;
				_ctr389: cs = 790; goto f113;
				_ctr398: cs = 790; goto f114;
				_ctr400: cs = 790; goto f115;
				_ctr401: cs = 790; goto f116;
				_ctr402: cs = 790; goto f117;
				_ctr405: cs = 790; goto f118;
				_ctr409: cs = 790; goto f119;
				_ctr414: cs = 790; goto f120;
				_ctr419: cs = 790; goto f121;
				_ctr420: cs = 790; goto f122;
				_ctr425: cs = 790; goto f123;
				_ctr428: cs = 790; goto f124;
				_ctr429: cs = 790; goto f125;
				_ctr437: cs = 790; goto f126;
				_ctr441: cs = 790; goto f127;
				_ctr442: cs = 790; goto f128;
				_ctr451: cs = 790; goto f129;
				_ctr458: cs = 790; goto f130;
				_ctr463: cs = 790; goto f131;
				_ctr464: cs = 790; goto f132;
				_ctr473: cs = 790; goto f133;
				_ctr474: cs = 790; goto f134;
				_ctr477: cs = 790; goto f135;
				_ctr484: cs = 790; goto f136;
				_ctr494: cs = 790; goto f137;
				_ctr500: cs = 790; goto f138;
				_ctr512: cs = 790; goto f139;
				_ctr517: cs = 790; goto f140;
				_ctr520: cs = 790; goto f141;
				_ctr525: cs = 790; goto f142;
				_ctr526: cs = 790; goto f143;
				_ctr533: cs = 790; goto f144;
				_ctr540: cs = 790; goto f145;
				_ctr544: cs = 790; goto f146;
				_ctr547: cs = 790; goto f147;
				_ctr548: cs = 790; goto f148;
				_ctr556: cs = 790; goto f149;
				_ctr563: cs = 790; goto f150;
				_ctr568: cs = 790; goto f151;
				_ctr573: cs = 790; goto f152;
				_ctr580: cs = 790; goto f153;
				_ctr591: cs = 790; goto f154;
				_ctr594: cs = 790; goto f155;
				_ctr602: cs = 790; goto f156;
				_ctr606: cs = 790; goto f157;
				_ctr611: cs = 790; goto f158;
				_ctr614: cs = 790; goto f159;
				_ctr617: cs = 790; goto f160;
				_ctr619: cs = 790; goto f161;
				_ctr623: cs = 790; goto f162;
				_ctr626: cs = 790; goto f163;
				_ctr628: cs = 790; goto f164;
				_ctr634: cs = 790; goto f165;
				_ctr639: cs = 790; goto f166;
				_ctr645: cs = 790; goto f167;
				_ctr646: cs = 790; goto f168;
				_ctr647: cs = 790; goto f169;
				_ctr648: cs = 790; goto f170;
				_ctr649: cs = 790; goto f171;
				_ctr650: cs = 790; goto f172;
				_ctr1031: cs = 790; goto f328;
				_ctr1055: cs = 790; goto f329;
				_ctr1064: cs = 790; goto f330;
				_ctr1066: cs = 790; goto f331;
				_ctr1067: cs = 790; goto f332;
				_ctr1069: cs = 790; goto f333;
				_ctr1071: cs = 790; goto f334;
				_ctr1074: cs = 790; goto f335;
				_ctr1088: cs = 790; goto f336;
				_ctr1096: cs = 790; goto f337;
				_ctr1097: cs = 790; goto f338;
				_ctr1101: cs = 790; goto f339;
				_ctr1102: cs = 790; goto f340;
				_ctr1103: cs = 790; goto f341;
				_ctr1104: cs = 790; goto f342;
				_ctr1105: cs = 790; goto f343;
				_ctr1107: cs = 790; goto f344;
				_ctr1109: cs = 790; goto f345;
				_ctr1111: cs = 790; goto f346;
				_ctr1114: cs = 790; goto f347;
				_ctr1116: cs = 790; goto f348;
				_ctr1117: cs = 790; goto f349;
				_ctr1120: cs = 790; goto f350;
				_ctr1121: cs = 790; goto f351;
				_ctr1126: cs = 790; goto f352;
				_ctr1127: cs = 790; goto f353;
				_ctr1128: cs = 790; goto f354;
				_ctr1129: cs = 790; goto f355;
				_ctr1130: cs = 790; goto f356;
				_ctr1132: cs = 790; goto f357;
				_ctr1133: cs = 790; goto f358;
				_ctr1134: cs = 790; goto f359;
				_ctr1137: cs = 790; goto f360;
				_ctr1139: cs = 790; goto f361;
				_ctr1142: cs = 790; goto f362;
				_ctr1143: cs = 790; goto f363;
				_ctr1144: cs = 790; goto f364;
				_ctr1157: cs = 790; goto f365;
				_ctr1158: cs = 790; goto f366;
				_ctr1159: cs = 790; goto f367;
				_ctr1160: cs = 790; goto f368;
				_ctr1161: cs = 790; goto f369;
				_ctr1032: cs = 791; goto f5;
				_ctr1033: cs = 792; goto f5;
				_ctr1034: cs = 793; goto f5;
				_ctr1070: cs = 794; goto f5;
				_ctr1035: cs = 795; goto f5;
				_ctr1036: cs = 796; goto f5;
				_ctr1037: cs = 797; goto f5;
				_ctr1038: cs = 798; goto f5;
				_ctr300: cs = 799; goto f5;
				_ctr1039: cs = 800; goto f5;
				_ctr1040: cs = 801; goto f5;
				_ctr1041: cs = 802; goto f5;
				_ctr1042: cs = 803; goto f5;
				_ctr1043: cs = 804; goto f5;
				_ctr1110: cs = 805; goto f5;
				_ctr1044: cs = 806; goto f5;
				_ctr1045: cs = 807; goto f5;
				_ctr1046: cs = 808; goto f5;
				_ctr1047: cs = 809; goto f5;
				_ctr1048: cs = 810; goto f5;
				_ctr1049: cs = 811; goto f5;
				_ctr1050: cs = 812; goto f5;
				_ctr1051: cs = 813; goto f5;
				_ctr1052: cs = 814; goto f5;
				_ctr1053: cs = 815; goto _again;
				_ctr1054: cs = 816; goto _again;
				_ctr1162: cs = 817; goto _again;
				_ctr658: cs = 817; goto f177;
				_ctr1163: cs = 818; goto _again;
				_ctr1164: cs = 818; goto f370;
				_ctr1165: cs = 818; goto f371;
				_ctr1166: cs = 818; goto f372;
				_ctr1167: cs = 818; goto f373;
				_ctr1168: cs = 818; goto f374;
				_ctr1169: cs = 818; goto f375;
				_ctr1170: cs = 818; goto f376;
				_ctr1171: cs = 818; goto f377;
				_ctr1172: cs = 819; goto _again;
				_ctr660: cs = 819; goto f178;
				_ctr664: cs = 819; goto f179;
				_ctr680: cs = 819; goto f180;
				_ctr699: cs = 819; goto f181;
				_ctr703: cs = 819; goto f182;
				_ctr708: cs = 819; goto f183;
				_ctr714: cs = 819; goto f184;
				_ctr720: cs = 819; goto f185;
				_ctr726: cs = 819; goto f186;
				_ctr732: cs = 819; goto f187;
				_ctr738: cs = 819; goto f188;
				_ctr745: cs = 819; goto f189;
				_ctr750: cs = 819; goto f190;
				_ctr756: cs = 819; goto f191;
				_ctr762: cs = 819; goto f192;
				_ctr767: cs = 819; goto f193;
				_ctr774: cs = 819; goto f194;
				_ctr782: cs = 819; goto f195;
				_ctr786: cs = 819; goto f196;
				_ctr791: cs = 819; goto f197;
				_ctr797: cs = 819; goto f198;
				_ctr803: cs = 819; goto f199;
				_ctr809: cs = 819; goto f200;
				_ctr815: cs = 819; goto f201;
				_ctr821: cs = 819; goto f202;
				_ctr828: cs = 819; goto f203;
				_ctr833: cs = 819; goto f204;
				_ctr839: cs = 819; goto f205;
				_ctr845: cs = 819; goto f206;
				_ctr850: cs = 819; goto f207;
				_ctr857: cs = 819; goto f208;
				_ctr861: cs = 819; goto f209;
				_ctr863: cs = 819; goto f210;
				_ctr864: cs = 819; goto f211;
				_ctr866: cs = 819; goto f212;
				_ctr867: cs = 819; goto f213;
				_ctr1173: cs = 819; goto f378;
				_ctr1174: cs = 819; goto f379;
				_ctr1177: cs = 819; goto f380;
				_ctr1183: cs = 819; goto f382;
				_ctr1185: cs = 819; goto f383;
				_ctr1188: cs = 819; goto f385;
				_ctr1189: cs = 819; goto f386;
				_ctr1190: cs = 819; goto f387;
				_ctr1191: cs = 819; goto f388;
				_ctr1192: cs = 819; goto f389;
				_ctr1193: cs = 819; goto f390;
				_ctr1195: cs = 819; goto f391;
				_ctr1196: cs = 819; goto f392;
				_ctr1197: cs = 819; goto f393;
				_ctr1198: cs = 819; goto f394;
				_ctr1199: cs = 819; goto f395;
				_ctr1200: cs = 819; goto f396;
				_ctr1202: cs = 819; goto f397;
				_ctr1203: cs = 819; goto f398;
				_ctr1204: cs = 819; goto f399;
				_ctr1205: cs = 819; goto f400;
				_ctr1206: cs = 819; goto f401;
				_ctr1207: cs = 819; goto f402;
				_ctr1210: cs = 819; goto f403;
				_ctr1211: cs = 819; goto f404;
				_ctr1212: cs = 819; goto f405;
				_ctr1213: cs = 819; goto f406;
				_ctr1214: cs = 819; goto f407;
				_ctr1216: cs = 819; goto f408;
				_ctr1218: cs = 819; goto f409;
				_ctr1219: cs = 819; goto f410;
				_ctr1221: cs = 819; goto f411;
				_ctr1222: cs = 819; goto f412;
				_ctr1223: cs = 819; goto f413;
				_ctr1224: cs = 819; goto f414;
				_ctr1225: cs = 819; goto f415;
				_ctr1226: cs = 819; goto f416;
				_ctr1227: cs = 819; goto f417;
				_ctr1229: cs = 819; goto f418;
				_ctr1230: cs = 819; goto f419;
				_ctr1231: cs = 819; goto f420;
				_ctr1232: cs = 819; goto f421;
				_ctr1237: cs = 819; goto f422;
				_ctr1238: cs = 819; goto f423;
				_ctr1239: cs = 819; goto f424;
				_ctr1240: cs = 819; goto f425;
				_ctr1241: cs = 819; goto f426;
				_ctr1175: cs = 820; goto f5;
				_ctr1176: cs = 821; goto _again;
				_ctr1186: cs = 822; goto f254;
				_ctr1217: cs = 823; goto f288;
				_ctr1187: cs = 824; goto f384;
				_ctr1220: cs = 825; goto f288;
				_ctr1194: cs = 826; goto _again;
				_ctr1201: cs = 827; goto _again;
				_ctr1208: cs = 828; goto f5;
				_ctr1209: cs = 829; goto _again;
				_ctr1215: cs = 830; goto f40;
				_ctr1233: cs = 831; goto f313;
				_ctr1234: cs = 831; goto f314;
				_ctr1235: cs = 831; goto f315;
				_ctr1236: cs = 832; goto f5;
				_ctr1246: cs = 836; goto _again;
				_ctr1247: cs = 836; goto f428;
				_ctr1252: cs = 836; goto f431;
				_ctr1253: cs = 836; goto f432;
				_ctr1254: cs = 836; goto f433;
				_ctr1248: cs = 837; goto _again;
				_ctr1255: cs = 838; goto _again;
				_ctr872: cs = 838; goto f216;
				_ctr1256: cs = 838; goto f434;
				_ctr1263: cs = 838; goto f436;
				_ctr1264: cs = 838; goto f437;
				_ctr1265: cs = 838; goto f438;
				_ctr1257: cs = 839; goto _again;
				_ctr1270: cs = 843; goto _again;
				_ctr877: cs = 843; goto f219;
				_ctr1271: cs = 843; goto f440;
				_ctr1278: cs = 843; goto f442;
				_ctr1279: cs = 843; goto f443;
				_ctr1280: cs = 843; goto f444;
				_ctr1272: cs = 844; goto _again;
				_ctr1285: cs = 848; goto _again;
				_ctr1286: cs = 848; goto f446;
				_ctr1287: cs = 848; goto f447;
				_ctr1288: cs = 849; goto _again;
				_ctr1289: cs = 849; goto f448;
				_ctr1290: cs = 849; goto f449;
				_ctr1021: cs = 62; goto _again;
				_ctr1022: cs = 63; goto _again;
				_ctr107: cs = 64; goto _again;
				_ctr1244: cs = 737; goto _again;
				_ctr1245: cs = 738; goto _again;
				_ctr870: cs = 739; goto _again;
				_ctr1268: cs = 740; goto _again;
				_ctr1269: cs = 741; goto _again;
				_ctr875: cs = 742; goto _again;
				_ctr1283: cs = 743; goto _again;
				_ctr1284: cs = 744; goto _again;
				_ctr880: cs = 745; goto _again;
				_ctr105: cs = 746; goto f57;
				_ctr106: cs = 746; goto f58;
				_ctr108: cs = 746; goto f59;
				_ctr886: cs = 746; goto f224;
				_ctr887: cs = 746; goto f226;
				_ctr900: cs = 746; goto f233;
				_ctr1019: cs = 746; goto f324;
				_ctr1020: cs = 746; goto f325;
				_ctr901: cs = 784; goto _again;
				_ctr902: cs = 785; goto f5;
				_ctr903: cs = 786; goto f5;
				_ctr868: cs = 819; goto f213;
				_ctr869: cs = 819; goto f214;
				_ctr871: cs = 819; goto f215;
				_ctr1178: cs = 819; goto f378;
				_ctr1179: cs = 819; goto f381;
				_ctr1242: cs = 819; goto f426;
				_ctr1243: cs = 819; goto f427;
				_ctr1180: cs = 833; goto _again;
				_ctr1181: cs = 834; goto f5;
				_ctr1182: cs = 835; goto f5;
				_ctr1249: cs = 836; goto f428;
				_ctr1250: cs = 836; goto f429;
				_ctr1251: cs = 836; goto f430;
				_ctr873: cs = 838; goto f216;
				_ctr874: cs = 838; goto f217;
				_ctr876: cs = 838; goto f218;
				_ctr1258: cs = 838; goto f434;
				_ctr1259: cs = 838; goto f435;
				_ctr1266: cs = 838; goto f438;
				_ctr1267: cs = 838; goto f439;
				_ctr1260: cs = 840; goto _again;
				_ctr1261: cs = 841; goto f5;
				_ctr1262: cs = 842; goto f5;
				_ctr878: cs = 843; goto f219;
				_ctr879: cs = 843; goto f220;
				_ctr881: cs = 843; goto f221;
				_ctr1273: cs = 843; goto f440;
				_ctr1274: cs = 843; goto f441;
				_ctr1281: cs = 843; goto f444;
				_ctr1282: cs = 843; goto f445;
				_ctr1275: cs = 845; goto _again;
				_ctr1276: cs = 846; goto f5;
				_ctr1277: cs = 847; goto f5;
				
				f237: _acts = ( _regex_actions + (1)); goto execFuncs;
				f17: _acts = ( _regex_actions + (3)); goto execFuncs;
				f254: _acts = ( _regex_actions + (5)); goto execFuncs;
				f40: _acts = ( _regex_actions + (7)); goto execFuncs;
				f384: _acts = ( _regex_actions + (9)); goto execFuncs;
				f53: _acts = ( _regex_actions + (11)); goto execFuncs;
				f55: _acts = ( _regex_actions + (13)); goto execFuncs;
				f288: _acts = ( _regex_actions + (15)); goto execFuncs;
				f39: _acts = ( _regex_actions + (17)); goto execFuncs;
				f313: _acts = ( _regex_actions + (19)); goto execFuncs;
				f315: _acts = ( _regex_actions + (21)); goto execFuncs;
				f314: _acts = ( _regex_actions + (23)); goto execFuncs;
				f33: _acts = ( _regex_actions + (25)); goto execFuncs;
				f30: _acts = ( _regex_actions + (27)); goto execFuncs;
				f54: _acts = ( _regex_actions + (29)); goto execFuncs;
				f56: _acts = ( _regex_actions + (31)); goto execFuncs;
				f176: _acts = ( _regex_actions + (33)); goto execFuncs;
				f175: _acts = ( _regex_actions + (35)); goto execFuncs;
				f177: _acts = ( _regex_actions + (37)); goto execFuncs;
				f174: _acts = ( _regex_actions + (39)); goto execFuncs;
				f5: _acts = ( _regex_actions + (45)); goto execFuncs;
				f64: _acts = ( _regex_actions + (47)); goto execFuncs;
				f63: _acts = ( _regex_actions + (49)); goto execFuncs;
				f62: _acts = ( _regex_actions + (51)); goto execFuncs;
				f61: _acts = ( _regex_actions + (53)); goto execFuncs;
				f326: _acts = ( _regex_actions + (55)); goto execFuncs;
				f327: _acts = ( _regex_actions + (57)); goto execFuncs;
				f60: _acts = ( _regex_actions + (59)); goto execFuncs;
				f331: _acts = ( _regex_actions + (61)); goto execFuncs;
				f332: _acts = ( _regex_actions + (63)); goto execFuncs;
				f333: _acts = ( _regex_actions + (65)); goto execFuncs;
				f334: _acts = ( _regex_actions + (67)); goto execFuncs;
				f339: _acts = ( _regex_actions + (69)); goto execFuncs;
				f340: _acts = ( _regex_actions + (71)); goto execFuncs;
				f341: _acts = ( _regex_actions + (73)); goto execFuncs;
				f342: _acts = ( _regex_actions + (75)); goto execFuncs;
				f343: _acts = ( _regex_actions + (77)); goto execFuncs;
				f338: _acts = ( _regex_actions + (79)); goto execFuncs;
				f345: _acts = ( _regex_actions + (81)); goto execFuncs;
				f346: _acts = ( _regex_actions + (83)); goto execFuncs;
				f349: _acts = ( _regex_actions + (85)); goto execFuncs;
				f350: _acts = ( _regex_actions + (87)); goto execFuncs;
				f351: _acts = ( _regex_actions + (89)); goto execFuncs;
				f353: _acts = ( _regex_actions + (91)); goto execFuncs;
				f354: _acts = ( _regex_actions + (93)); goto execFuncs;
				f355: _acts = ( _regex_actions + (95)); goto execFuncs;
				f356: _acts = ( _regex_actions + (97)); goto execFuncs;
				f357: _acts = ( _regex_actions + (99)); goto execFuncs;
				f358: _acts = ( _regex_actions + (101)); goto execFuncs;
				f359: _acts = ( _regex_actions + (103)); goto execFuncs;
				f361: _acts = ( _regex_actions + (105)); goto execFuncs;
				f362: _acts = ( _regex_actions + (107)); goto execFuncs;
				f363: _acts = ( _regex_actions + (109)); goto execFuncs;
				f364: _acts = ( _regex_actions + (111)); goto execFuncs;
				f367: _acts = ( _regex_actions + (113)); goto execFuncs;
				f368: _acts = ( _regex_actions + (115)); goto execFuncs;
				f369: _acts = ( _regex_actions + (117)); goto execFuncs;
				f169: _acts = ( _regex_actions + (119)); goto execFuncs;
				f170: _acts = ( _regex_actions + (121)); goto execFuncs;
				f171: _acts = ( _regex_actions + (123)); goto execFuncs;
				f172: _acts = ( _regex_actions + (125)); goto execFuncs;
				f67: _acts = ( _regex_actions + (127)); goto execFuncs;
				f68: _acts = ( _regex_actions + (129)); goto execFuncs;
				f69: _acts = ( _regex_actions + (131)); goto execFuncs;
				f70: _acts = ( _regex_actions + (133)); goto execFuncs;
				f71: _acts = ( _regex_actions + (135)); goto execFuncs;
				f72: _acts = ( _regex_actions + (137)); goto execFuncs;
				f73: _acts = ( _regex_actions + (139)); goto execFuncs;
				f74: _acts = ( _regex_actions + (141)); goto execFuncs;
				f75: _acts = ( _regex_actions + (143)); goto execFuncs;
				f76: _acts = ( _regex_actions + (145)); goto execFuncs;
				f77: _acts = ( _regex_actions + (147)); goto execFuncs;
				f78: _acts = ( _regex_actions + (149)); goto execFuncs;
				f80: _acts = ( _regex_actions + (151)); goto execFuncs;
				f81: _acts = ( _regex_actions + (153)); goto execFuncs;
				f82: _acts = ( _regex_actions + (155)); goto execFuncs;
				f83: _acts = ( _regex_actions + (157)); goto execFuncs;
				f85: _acts = ( _regex_actions + (159)); goto execFuncs;
				f86: _acts = ( _regex_actions + (161)); goto execFuncs;
				f87: _acts = ( _regex_actions + (163)); goto execFuncs;
				f88: _acts = ( _regex_actions + (165)); goto execFuncs;
				f89: _acts = ( _regex_actions + (167)); goto execFuncs;
				f90: _acts = ( _regex_actions + (169)); goto execFuncs;
				f91: _acts = ( _regex_actions + (171)); goto execFuncs;
				f92: _acts = ( _regex_actions + (173)); goto execFuncs;
				f93: _acts = ( _regex_actions + (175)); goto execFuncs;
				f94: _acts = ( _regex_actions + (177)); goto execFuncs;
				f95: _acts = ( _regex_actions + (179)); goto execFuncs;
				f96: _acts = ( _regex_actions + (181)); goto execFuncs;
				f97: _acts = ( _regex_actions + (183)); goto execFuncs;
				f98: _acts = ( _regex_actions + (185)); goto execFuncs;
				f99: _acts = ( _regex_actions + (187)); goto execFuncs;
				f101: _acts = ( _regex_actions + (189)); goto execFuncs;
				f102: _acts = ( _regex_actions + (191)); goto execFuncs;
				f103: _acts = ( _regex_actions + (193)); goto execFuncs;
				f104: _acts = ( _regex_actions + (195)); goto execFuncs;
				f105: _acts = ( _regex_actions + (197)); goto execFuncs;
				f106: _acts = ( _regex_actions + (199)); goto execFuncs;
				f107: _acts = ( _regex_actions + (201)); goto execFuncs;
				f108: _acts = ( _regex_actions + (203)); goto execFuncs;
				f109: _acts = ( _regex_actions + (205)); goto execFuncs;
				f110: _acts = ( _regex_actions + (207)); goto execFuncs;
				f111: _acts = ( _regex_actions + (209)); goto execFuncs;
				f112: _acts = ( _regex_actions + (211)); goto execFuncs;
				f113: _acts = ( _regex_actions + (213)); goto execFuncs;
				f114: _acts = ( _regex_actions + (215)); goto execFuncs;
				f115: _acts = ( _regex_actions + (217)); goto execFuncs;
				f117: _acts = ( _regex_actions + (219)); goto execFuncs;
				f118: _acts = ( _regex_actions + (221)); goto execFuncs;
				f119: _acts = ( _regex_actions + (223)); goto execFuncs;
				f120: _acts = ( _regex_actions + (225)); goto execFuncs;
				f121: _acts = ( _regex_actions + (227)); goto execFuncs;
				f122: _acts = ( _regex_actions + (229)); goto execFuncs;
				f123: _acts = ( _regex_actions + (231)); goto execFuncs;
				f124: _acts = ( _regex_actions + (233)); goto execFuncs;
				f126: _acts = ( _regex_actions + (235)); goto execFuncs;
				f127: _acts = ( _regex_actions + (237)); goto execFuncs;
				f129: _acts = ( _regex_actions + (239)); goto execFuncs;
				f130: _acts = ( _regex_actions + (241)); goto execFuncs;
				f131: _acts = ( _regex_actions + (243)); goto execFuncs;
				f133: _acts = ( _regex_actions + (245)); goto execFuncs;
				f134: _acts = ( _regex_actions + (247)); goto execFuncs;
				f135: _acts = ( _regex_actions + (249)); goto execFuncs;
				f136: _acts = ( _regex_actions + (251)); goto execFuncs;
				f137: _acts = ( _regex_actions + (253)); goto execFuncs;
				f138: _acts = ( _regex_actions + (255)); goto execFuncs;
				f139: _acts = ( _regex_actions + (257)); goto execFuncs;
				f140: _acts = ( _regex_actions + (259)); goto execFuncs;
				f141: _acts = ( _regex_actions + (261)); goto execFuncs;
				f142: _acts = ( _regex_actions + (263)); goto execFuncs;
				f144: _acts = ( _regex_actions + (265)); goto execFuncs;
				f145: _acts = ( _regex_actions + (267)); goto execFuncs;
				f146: _acts = ( _regex_actions + (269)); goto execFuncs;
				f147: _acts = ( _regex_actions + (271)); goto execFuncs;
				f149: _acts = ( _regex_actions + (273)); goto execFuncs;
				f150: _acts = ( _regex_actions + (275)); goto execFuncs;
				f151: _acts = ( _regex_actions + (277)); goto execFuncs;
				f152: _acts = ( _regex_actions + (279)); goto execFuncs;
				f153: _acts = ( _regex_actions + (281)); goto execFuncs;
				f154: _acts = ( _regex_actions + (283)); goto execFuncs;
				f155: _acts = ( _regex_actions + (285)); goto execFuncs;
				f156: _acts = ( _regex_actions + (287)); goto execFuncs;
				f157: _acts = ( _regex_actions + (289)); goto execFuncs;
				f158: _acts = ( _regex_actions + (291)); goto execFuncs;
				f159: _acts = ( _regex_actions + (293)); goto execFuncs;
				f160: _acts = ( _regex_actions + (295)); goto execFuncs;
				f161: _acts = ( _regex_actions + (297)); goto execFuncs;
				f162: _acts = ( _regex_actions + (299)); goto execFuncs;
				f164: _acts = ( _regex_actions + (301)); goto execFuncs;
				f163: _acts = ( _regex_actions + (303)); goto execFuncs;
				f165: _acts = ( _regex_actions + (305)); goto execFuncs;
				f166: _acts = ( _regex_actions + (307)); goto execFuncs;
				f167: _acts = ( _regex_actions + (309)); goto execFuncs;
				f168: _acts = ( _regex_actions + (311)); goto execFuncs;
				f365: _acts = ( _regex_actions + (313)); goto execFuncs;
				f66: _acts = ( _regex_actions + (315)); goto execFuncs;
				f328: _acts = ( _regex_actions + (317)); goto execFuncs;
				f330: _acts = ( _regex_actions + (319)); goto execFuncs;
				f335: _acts = ( _regex_actions + (321)); goto execFuncs;
				f337: _acts = ( _regex_actions + (323)); goto execFuncs;
				f344: _acts = ( _regex_actions + (325)); goto execFuncs;
				f347: _acts = ( _regex_actions + (327)); goto execFuncs;
				f348: _acts = ( _regex_actions + (329)); goto execFuncs;
				f352: _acts = ( _regex_actions + (331)); goto execFuncs;
				f360: _acts = ( _regex_actions + (333)); goto execFuncs;
				f366: _acts = ( _regex_actions + (335)); goto execFuncs;
				f336: _acts = ( _regex_actions + (337)); goto execFuncs;
				f329: _acts = ( _regex_actions + (339)); goto execFuncs;
				f79: _acts = ( _regex_actions + (341)); goto execFuncs;
				f84: _acts = ( _regex_actions + (343)); goto execFuncs;
				f116: _acts = ( _regex_actions + (345)); goto execFuncs;
				f125: _acts = ( _regex_actions + (347)); goto execFuncs;
				f128: _acts = ( _regex_actions + (349)); goto execFuncs;
				f132: _acts = ( _regex_actions + (351)); goto execFuncs;
				f143: _acts = ( _regex_actions + (353)); goto execFuncs;
				f148: _acts = ( _regex_actions + (355)); goto execFuncs;
				f100: _acts = ( _regex_actions + (357)); goto execFuncs;
				f65: _acts = ( _regex_actions + (359)); goto execFuncs;
				f371: _acts = ( _regex_actions + (361)); goto execFuncs;
				f372: _acts = ( _regex_actions + (363)); goto execFuncs;
				f373: _acts = ( _regex_actions + (365)); goto execFuncs;
				f374: _acts = ( _regex_actions + (367)); goto execFuncs;
				f375: _acts = ( _regex_actions + (369)); goto execFuncs;
				f376: _acts = ( _regex_actions + (371)); goto execFuncs;
				f377: _acts = ( _regex_actions + (373)); goto execFuncs;
				f370: _acts = ( _regex_actions + (375)); goto execFuncs;
				f179: _acts = ( _regex_actions + (377)); goto execFuncs;
				f195: _acts = ( _regex_actions + (379)); goto execFuncs;
				f181: _acts = ( _regex_actions + (381)); goto execFuncs;
				f196: _acts = ( _regex_actions + (383)); goto execFuncs;
				f182: _acts = ( _regex_actions + (385)); goto execFuncs;
				f197: _acts = ( _regex_actions + (387)); goto execFuncs;
				f183: _acts = ( _regex_actions + (389)); goto execFuncs;
				f198: _acts = ( _regex_actions + (391)); goto execFuncs;
				f184: _acts = ( _regex_actions + (393)); goto execFuncs;
				f199: _acts = ( _regex_actions + (395)); goto execFuncs;
				f185: _acts = ( _regex_actions + (397)); goto execFuncs;
				f200: _acts = ( _regex_actions + (399)); goto execFuncs;
				f186: _acts = ( _regex_actions + (401)); goto execFuncs;
				f201: _acts = ( _regex_actions + (403)); goto execFuncs;
				f187: _acts = ( _regex_actions + (405)); goto execFuncs;
				f202: _acts = ( _regex_actions + (407)); goto execFuncs;
				f188: _acts = ( _regex_actions + (409)); goto execFuncs;
				f203: _acts = ( _regex_actions + (411)); goto execFuncs;
				f189: _acts = ( _regex_actions + (413)); goto execFuncs;
				f204: _acts = ( _regex_actions + (415)); goto execFuncs;
				f190: _acts = ( _regex_actions + (417)); goto execFuncs;
				f205: _acts = ( _regex_actions + (419)); goto execFuncs;
				f191: _acts = ( _regex_actions + (421)); goto execFuncs;
				f206: _acts = ( _regex_actions + (423)); goto execFuncs;
				f192: _acts = ( _regex_actions + (425)); goto execFuncs;
				f207: _acts = ( _regex_actions + (427)); goto execFuncs;
				f193: _acts = ( _regex_actions + (429)); goto execFuncs;
				f208: _acts = ( _regex_actions + (431)); goto execFuncs;
				f194: _acts = ( _regex_actions + (433)); goto execFuncs;
				f180: _acts = ( _regex_actions + (435)); goto execFuncs;
				f391: _acts = ( _regex_actions + (437)); goto execFuncs;
				f388: _acts = ( _regex_actions + (439)); goto execFuncs;
				f396: _acts = ( _regex_actions + (441)); goto execFuncs;
				f405: _acts = ( _regex_actions + (443)); goto execFuncs;
				f402: _acts = ( _regex_actions + (445)); goto execFuncs;
				f403: _acts = ( _regex_actions + (447)); goto execFuncs;
				f399: _acts = ( _regex_actions + (449)); goto execFuncs;
				f395: _acts = ( _regex_actions + (451)); goto execFuncs;
				f398: _acts = ( _regex_actions + (453)); goto execFuncs;
				f401: _acts = ( _regex_actions + (455)); goto execFuncs;
				f389: _acts = ( _regex_actions + (457)); goto execFuncs;
				f406: _acts = ( _regex_actions + (459)); goto execFuncs;
				f393: _acts = ( _regex_actions + (461)); goto execFuncs;
				f420: _acts = ( _regex_actions + (463)); goto execFuncs;
				f419: _acts = ( _regex_actions + (465)); goto execFuncs;
				f414: _acts = ( _regex_actions + (467)); goto execFuncs;
				f413: _acts = ( _regex_actions + (469)); goto execFuncs;
				f210: _acts = ( _regex_actions + (471)); goto execFuncs;
				f385: _acts = ( _regex_actions + (473)); goto execFuncs;
				f212: _acts = ( _regex_actions + (475)); goto execFuncs;
				f416: _acts = ( _regex_actions + (477)); goto execFuncs;
				f407: _acts = ( _regex_actions + (479)); goto execFuncs;
				f394: _acts = ( _regex_actions + (481)); goto execFuncs;
				f404: _acts = ( _regex_actions + (483)); goto execFuncs;
				f392: _acts = ( _regex_actions + (485)); goto execFuncs;
				f397: _acts = ( _regex_actions + (487)); goto execFuncs;
				f387: _acts = ( _regex_actions + (489)); goto execFuncs;
				f379: _acts = ( _regex_actions + (491)); goto execFuncs;
				f390: _acts = ( _regex_actions + (493)); goto execFuncs;
				f400: _acts = ( _regex_actions + (495)); goto execFuncs;
				f386: _acts = ( _regex_actions + (497)); goto execFuncs;
				f383: _acts = ( _regex_actions + (499)); goto execFuncs;
				f427: _acts = ( _regex_actions + (501)); goto execFuncs;
				f214: _acts = ( _regex_actions + (503)); goto execFuncs;
				f215: _acts = ( _regex_actions + (505)); goto execFuncs;
				f381: _acts = ( _regex_actions + (507)); goto execFuncs;
				f378: _acts = ( _regex_actions + (509)); goto execFuncs;
				f380: _acts = ( _regex_actions + (511)); goto execFuncs;
				f412: _acts = ( _regex_actions + (513)); goto execFuncs;
				f418: _acts = ( _regex_actions + (515)); goto execFuncs;
				f408: _acts = ( _regex_actions + (517)); goto execFuncs;
				f410: _acts = ( _regex_actions + (519)); goto execFuncs;
				f417: _acts = ( _regex_actions + (521)); goto execFuncs;
				f421: _acts = ( _regex_actions + (523)); goto execFuncs;
				f425: _acts = ( _regex_actions + (525)); goto execFuncs;
				f415: _acts = ( _regex_actions + (527)); goto execFuncs;
				f426: _acts = ( _regex_actions + (529)); goto execFuncs;
				f382: _acts = ( _regex_actions + (531)); goto execFuncs;
				f209: _acts = ( _regex_actions + (533)); goto execFuncs;
				f211: _acts = ( _regex_actions + (535)); goto execFuncs;
				f213: _acts = ( _regex_actions + (537)); goto execFuncs;
				f178: _acts = ( _regex_actions + (539)); goto execFuncs;
				f430: _acts = ( _regex_actions + (541)); goto execFuncs;
				f429: _acts = ( _regex_actions + (543)); goto execFuncs;
				f433: _acts = ( _regex_actions + (545)); goto execFuncs;
				f432: _acts = ( _regex_actions + (547)); goto execFuncs;
				f428: _acts = ( _regex_actions + (549)); goto execFuncs;
				f431: _acts = ( _regex_actions + (551)); goto execFuncs;
				f437: _acts = ( _regex_actions + (553)); goto execFuncs;
				f439: _acts = ( _regex_actions + (555)); goto execFuncs;
				f217: _acts = ( _regex_actions + (557)); goto execFuncs;
				f218: _acts = ( _regex_actions + (559)); goto execFuncs;
				f435: _acts = ( _regex_actions + (561)); goto execFuncs;
				f434: _acts = ( _regex_actions + (563)); goto execFuncs;
				f438: _acts = ( _regex_actions + (565)); goto execFuncs;
				f436: _acts = ( _regex_actions + (567)); goto execFuncs;
				f216: _acts = ( _regex_actions + (569)); goto execFuncs;
				f443: _acts = ( _regex_actions + (571)); goto execFuncs;
				f445: _acts = ( _regex_actions + (573)); goto execFuncs;
				f220: _acts = ( _regex_actions + (575)); goto execFuncs;
				f221: _acts = ( _regex_actions + (577)); goto execFuncs;
				f441: _acts = ( _regex_actions + (579)); goto execFuncs;
				f440: _acts = ( _regex_actions + (581)); goto execFuncs;
				f444: _acts = ( _regex_actions + (583)); goto execFuncs;
				f442: _acts = ( _regex_actions + (585)); goto execFuncs;
				f219: _acts = ( _regex_actions + (587)); goto execFuncs;
				f447: _acts = ( _regex_actions + (589)); goto execFuncs;
				f446: _acts = ( _regex_actions + (591)); goto execFuncs;
				f449: _acts = ( _regex_actions + (593)); goto execFuncs;
				f448: _acts = ( _regex_actions + (595)); goto execFuncs;
				f228: _acts = ( _regex_actions + (597)); goto execFuncs;
				f232: _acts = ( _regex_actions + (599)); goto execFuncs;
				f36: _acts = ( _regex_actions + (601)); goto execFuncs;
				f35: _acts = ( _regex_actions + (603)); goto execFuncs;
				f266: _acts = ( _regex_actions + (605)); goto execFuncs;
				f261: _acts = ( _regex_actions + (607)); goto execFuncs;
				f229: _acts = ( _regex_actions + (609)); goto execFuncs;
				f259: _acts = ( _regex_actions + (611)); goto execFuncs;
				f244: _acts = ( _regex_actions + (613)); goto execFuncs;
				f243: _acts = ( _regex_actions + (615)); goto execFuncs;
				f247: _acts = ( _regex_actions + (617)); goto execFuncs;
				f246: _acts = ( _regex_actions + (619)); goto execFuncs;
				f250: _acts = ( _regex_actions + (621)); goto execFuncs;
				f249: _acts = ( _regex_actions + (623)); goto execFuncs;
				f323: _acts = ( _regex_actions + (625)); goto execFuncs;
				f322: _acts = ( _regex_actions + (627)); goto execFuncs;
				f226: _acts = ( _regex_actions + (629)); goto execFuncs;
				f1: _acts = ( _regex_actions + (631)); goto execFuncs;
				f223: _acts = ( _regex_actions + (633)); goto execFuncs;
				f230: _acts = ( _regex_actions + (635)); goto execFuncs;
				f227: _acts = ( _regex_actions + (637)); goto execFuncs;
				f257: _acts = ( _regex_actions + (639)); goto execFuncs;
				f272: _acts = ( _regex_actions + (641)); goto execFuncs;
				f286: _acts = ( _regex_actions + (643)); goto execFuncs;
				f274: _acts = ( _regex_actions + (645)); goto execFuncs;
				f258: _acts = ( _regex_actions + (647)); goto execFuncs;
				f283: _acts = ( _regex_actions + (649)); goto execFuncs;
				f280: _acts = ( _regex_actions + (651)); goto execFuncs;
				f281: _acts = ( _regex_actions + (653)); goto execFuncs;
				f277: _acts = ( _regex_actions + (655)); goto execFuncs;
				f273: _acts = ( _regex_actions + (657)); goto execFuncs;
				f276: _acts = ( _regex_actions + (659)); goto execFuncs;
				f42: _acts = ( _regex_actions + (661)); goto execFuncs;
				f41: _acts = ( _regex_actions + (663)); goto execFuncs;
				f43: _acts = ( _regex_actions + (665)); goto execFuncs;
				f47: _acts = ( _regex_actions + (667)); goto execFuncs;
				f46: _acts = ( _regex_actions + (669)); goto execFuncs;
				f45: _acts = ( _regex_actions + (671)); goto execFuncs;
				f32: _acts = ( _regex_actions + (673)); goto execFuncs;
				f38: _acts = ( _regex_actions + (675)); goto execFuncs;
				f49: _acts = ( _regex_actions + (677)); goto execFuncs;
				f51: _acts = ( _regex_actions + (679)); goto execFuncs;
				f300: _acts = ( _regex_actions + (681)); goto execFuncs;
				f265: _acts = ( _regex_actions + (683)); goto execFuncs;
				f285: _acts = ( _regex_actions + (685)); goto execFuncs;
				f270: _acts = ( _regex_actions + (687)); goto execFuncs;
				f282: _acts = ( _regex_actions + (689)); goto execFuncs;
				f268: _acts = ( _regex_actions + (691)); goto execFuncs;
				f275: _acts = ( _regex_actions + (693)); goto execFuncs;
				f260: _acts = ( _regex_actions + (695)); goto execFuncs;
				f279: _acts = ( _regex_actions + (697)); goto execFuncs;
				f263: _acts = ( _regex_actions + (699)); goto execFuncs;
				f284: _acts = ( _regex_actions + (701)); goto execFuncs;
				f269: _acts = ( _regex_actions + (703)); goto execFuncs;
				f311: _acts = ( _regex_actions + (705)); goto execFuncs;
				f310: _acts = ( _regex_actions + (707)); goto execFuncs;
				f298: _acts = ( _regex_actions + (709)); goto execFuncs;
				f297: _acts = ( _regex_actions + (711)); goto execFuncs;
				f267: _acts = ( _regex_actions + (713)); goto execFuncs;
				f264: _acts = ( _regex_actions + (715)); goto execFuncs;
				f262: _acts = ( _regex_actions + (717)); goto execFuncs;
				f271: _acts = ( _regex_actions + (719)); goto execFuncs;
				f253: _acts = ( _regex_actions + (721)); goto execFuncs;
				f4: _acts = ( _regex_actions + (723)); goto execFuncs;
				f28: _acts = ( _regex_actions + (725)); goto execFuncs;
				f29: _acts = ( _regex_actions + (727)); goto execFuncs;
				f10: _acts = ( _regex_actions + (729)); goto execFuncs;
				f3: _acts = ( _regex_actions + (731)); goto execFuncs;
				f240: _acts = ( _regex_actions + (733)); goto execFuncs;
				f239: _acts = ( _regex_actions + (735)); goto execFuncs;
				f13: _acts = ( _regex_actions + (737)); goto execFuncs;
				f241: _acts = ( _regex_actions + (739)); goto execFuncs;
				f11: _acts = ( _regex_actions + (741)); goto execFuncs;
				f16: _acts = ( _regex_actions + (743)); goto execFuncs;
				f27: _acts = ( _regex_actions + (745)); goto execFuncs;
				f15: _acts = ( _regex_actions + (747)); goto execFuncs;
				f23: _acts = ( _regex_actions + (749)); goto execFuncs;
				f22: _acts = ( _regex_actions + (751)); goto execFuncs;
				f25: _acts = ( _regex_actions + (753)); goto execFuncs;
				f24: _acts = ( _regex_actions + (755)); goto execFuncs;
				f26: _acts = ( _regex_actions + (757)); goto execFuncs;
				f20: _acts = ( _regex_actions + (759)); goto execFuncs;
				f19: _acts = ( _regex_actions + (761)); goto execFuncs;
				f31: _acts = ( _regex_actions + (763)); goto execFuncs;
				f2: _acts = ( _regex_actions + (765)); goto execFuncs;
				f325: _acts = ( _regex_actions + (767)); goto execFuncs;
				f58: _acts = ( _regex_actions + (769)); goto execFuncs;
				f59: _acts = ( _regex_actions + (771)); goto execFuncs;
				f233: _acts = ( _regex_actions + (773)); goto execFuncs;
				f225: _acts = ( _regex_actions + (775)); goto execFuncs;
				f224: _acts = ( _regex_actions + (777)); goto execFuncs;
				f234: _acts = ( _regex_actions + (779)); goto execFuncs;
				f251: _acts = ( _regex_actions + (781)); goto execFuncs;
				f242: _acts = ( _regex_actions + (783)); goto execFuncs;
				f245: _acts = ( _regex_actions + (785)); goto execFuncs;
				f248: _acts = ( _regex_actions + (787)); goto execFuncs;
				f321: _acts = ( _regex_actions + (789)); goto execFuncs;
				f287: _acts = ( _regex_actions + (791)); goto execFuncs;
				f293: _acts = ( _regex_actions + (793)); goto execFuncs;
				f290: _acts = ( _regex_actions + (795)); goto execFuncs;
				f305: _acts = ( _regex_actions + (797)); goto execFuncs;
				f303: _acts = ( _regex_actions + (799)); goto execFuncs;
				f301: _acts = ( _regex_actions + (801)); goto execFuncs;
				f308: _acts = ( _regex_actions + (803)); goto execFuncs;
				f312: _acts = ( _regex_actions + (805)); goto execFuncs;
				f319: _acts = ( _regex_actions + (807)); goto execFuncs;
				f299: _acts = ( _regex_actions + (809)); goto execFuncs;
				f296: _acts = ( _regex_actions + (811)); goto execFuncs;
				f309: _acts = ( _regex_actions + (813)); goto execFuncs;
				f307: _acts = ( _regex_actions + (815)); goto execFuncs;
				f252: _acts = ( _regex_actions + (817)); goto execFuncs;
				f236: _acts = ( _regex_actions + (819)); goto execFuncs;
				f235: _acts = ( _regex_actions + (821)); goto execFuncs;
				f324: _acts = ( _regex_actions + (823)); goto execFuncs;
				f320: _acts = ( _regex_actions + (825)); goto execFuncs;
				f0: _acts = ( _regex_actions + (827)); goto execFuncs;
				f34: _acts = ( _regex_actions + (829)); goto execFuncs;
				f37: _acts = ( _regex_actions + (831)); goto execFuncs;
				f48: _acts = ( _regex_actions + (833)); goto execFuncs;
				f50: _acts = ( _regex_actions + (835)); goto execFuncs;
				f44: _acts = ( _regex_actions + (837)); goto execFuncs;
				f18: _acts = ( _regex_actions + (839)); goto execFuncs;
				f14: _acts = ( _regex_actions + (841)); goto execFuncs;
				f57: _acts = ( _regex_actions + (843)); goto execFuncs;
				f52: _acts = ( _regex_actions + (845)); goto execFuncs;
				f295: _acts = ( _regex_actions + (847)); goto execFuncs;
				f302: _acts = ( _regex_actions + (849)); goto execFuncs;
				f255: _acts = ( _regex_actions + (852)); goto execFuncs;
				f238: _acts = ( _regex_actions + (855)); goto execFuncs;
				f291: _acts = ( _regex_actions + (858)); goto execFuncs;
				f409: _acts = ( _regex_actions + (861)); goto execFuncs;
				f411: _acts = ( _regex_actions + (864)); goto execFuncs;
				f289: _acts = ( _regex_actions + (867)); goto execFuncs;
				f21: _acts = ( _regex_actions + (870)); goto execFuncs;
				f306: _acts = ( _regex_actions + (873)); goto execFuncs;
				f304: _acts = ( _regex_actions + (876)); goto execFuncs;
				f422: _acts = ( _regex_actions + (879)); goto execFuncs;
				f316: _acts = ( _regex_actions + (882)); goto execFuncs;
				f424: _acts = ( _regex_actions + (885)); goto execFuncs;
				f318: _acts = ( _regex_actions + (888)); goto execFuncs;
				f423: _acts = ( _regex_actions + (891)); goto execFuncs;
				f317: _acts = ( _regex_actions + (894)); goto execFuncs;
				f7: _acts = ( _regex_actions + (897)); goto execFuncs;
				f9: _acts = ( _regex_actions + (900)); goto execFuncs;
				f6: _acts = ( _regex_actions + (903)); goto execFuncs;
				f278: _acts = ( _regex_actions + (906)); goto execFuncs;
				f231: _acts = ( _regex_actions + (909)); goto execFuncs;
				f8: _acts = ( _regex_actions + (912)); goto execFuncs;
				f256: _acts = ( _regex_actions + (915)); goto execFuncs;
				f292: _acts = ( _regex_actions + (919)); goto execFuncs;
				f12: _acts = ( _regex_actions + (923)); goto execFuncs;
				f294: _acts = ( _regex_actions + (927)); goto execFuncs;
				
				execFuncs:
				_nacts = (unsigned int)(*( _acts));
				_acts += 1;
				while ( _nacts > 0 ) {
					switch ( (*( _acts)) ) {
						case 0: {
							{
#line 286 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								label.clear();}
							
#line 14968 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 1: {
							{
#line 287 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								label.push_back((( (*( p)))));}
							
#line 14977 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 2: {
							{
#line 288 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								octAccumulator = 0;}
							
#line 14986 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 3: {
							{
#line 289 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								accumulator = 0;}
							
#line 14995 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 4: {
							{
#line 290 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								octAccumulator = 0;
								pushOct(&octAccumulator, (( (*( p)))));
							}
							
#line 15007 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 5: {
							{
#line 294 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								accumulator = 0;
								pushDec(&accumulator, (( (*( p)))));
							}
							
#line 15019 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 6: {
							{
#line 298 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								repeatN = 0; repeatM = 0; }
							
#line 15028 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 7: {
							{
#line 299 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								pushDec(&repeatN, (( (*( p))))); }
							
#line 15037 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 8: {
							{
#line 300 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								pushDec(&repeatM, (( (*( p))))); }
							
#line 15046 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 9: {
							{
#line 301 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								pushOct(&octAccumulator, (( (*( p))))); }
							
#line 15055 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 10: {
							{
#line 302 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								pushDec(&accumulator, (( (*( p))))); }
							
#line 15064 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 11: {
							{
#line 303 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								accumulator *= 16;
								accumulator += (( (*( p)))) - '0';
							}
							
#line 15076 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 12: {
							{
#line 307 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								accumulator *= 16;
								accumulator += 10 + (( (*( p)))) - 'a';
							}
							
#line 15088 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 13: {
							{
#line 311 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								accumulator *= 16;
								accumulator += 10 + (( (*( p)))) - 'A';
							}
							
#line 15100 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 14: {
							{
#line 431 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								newMode = mode;
							}
							
#line 15111 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 15: {
							{
#line 438 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								switch ((( (*( p))))) {
									case 'i':
									newMode.caseless = true;
									break;
									case 'm':
									newMode.multiline = true;
									break;
									case 's':
									newMode.dotall = true;
									break;
									case 'x':
									newMode.ignore_space = true;
									break;
									default:
									assert(0); // this action only called for [imsx]
									break;
								}
							}
							
#line 15138 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 16: {
							{
#line 457 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								switch ((( (*( p))))) {
									case 'i':
									newMode.caseless = false;
									break;
									case 'm':
									newMode.multiline = false;
									break;
									case 's':
									newMode.dotall = false;
									break;
									case 'x':
									newMode.ignore_space = false;
									break;
									default:
									assert(0); // this action only called for [imsx]
									break;
								}
							}
							
#line 15165 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 17: {
							{
#line 511 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								repeatM = repeatN;}
							
#line 15174 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 18: {
							{
#line 511 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								repeatM = ComponentRepeat::NoLimit;}
							
#line 15183 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 19: {
							{
#line 723 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								negated = !negated; }
							
#line 15192 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 20: {
							{
#line 724 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = p - 1; } {{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 790; goto _again;}}
							
#line 15209 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 21: {
							{
#line 725 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								if (!inCharClass) { // not inside [..]
									currentCls->finalize();
									currentSeq->addComponent(std::move(currentCls));
								}
								{top-= 1;cs = stack[top];goto _again;} 
							}
							
#line 15223 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 22: {
							{
#line 731 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								throw LocatedParseError("Malformed property"); }
							
#line 15232 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 25: {
							{
#line 1 "NONE"
								{te = p+1;}}
							
#line 15241 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 26: {
							{
#line 551 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 551 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("(*UTF8) must be at start of "
										"expression, encountered");
									}
								}}
							
#line 15256 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 27: {
							{
#line 555 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 555 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("(*UTF) must be at start of "
										"expression, encountered");
									}
								}}
							
#line 15271 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 28: {
							{
#line 559 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 559 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("(*UCP) must be at start of "
										"expression, encountered");
									}
								}}
							
#line 15286 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 29: {
							{
#line 565 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 565 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										ParseMode temp_mode;
										assert(ts - 2 >= ptr); // parser needs the '(*' at the start too.
										read_control_verbs(ts - 2, te, (ts - 2 - ptr), temp_mode);
										assert(0); // Should have thrown a parse error.
										throw LocatedParseError("Unknown control verb");
									}
								}}
							
#line 15304 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 30: {
							{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Unknown control verb");
									}
								}}
							
#line 15318 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 31: {
							{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Unknown control verb");
									}
								}}
							
#line 15332 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 32: {
							{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 572 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Unknown control verb");
									}
								}}
							
#line 15347 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 33: {
							{
#line 582 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 582 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15359 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 34: {
							{
#line 583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CF, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15371 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 35: {
							{
#line 584 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 584 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15383 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 36: {
							{
#line 586 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 586 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CS, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15395 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 37: {
							{
#line 588 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 588 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_LL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15407 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 38: {
							{
#line 589 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 589 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_LM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15419 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 39: {
							{
#line 590 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 590 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_LO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15431 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 40: {
							{
#line 591 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 591 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_LT, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15443 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 41: {
							{
#line 592 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 592 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_LU, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15455 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 42: {
							{
#line 593 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 593 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_L_AND, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15467 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 43: {
							{
#line 595 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 595 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_MC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15479 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 44: {
							{
#line 597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_MN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15491 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 45: {
							{
#line 599 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 599 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ND, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15503 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 46: {
							{
#line 600 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 600 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_NL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15515 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 47: {
							{
#line 601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_NO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15527 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 48: {
							{
#line 603 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 603 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15539 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 49: {
							{
#line 604 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 604 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PD, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15551 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 50: {
							{
#line 605 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 605 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15563 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 51: {
							{
#line 606 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 606 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PF, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15575 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 52: {
							{
#line 607 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 607 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15587 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 53: {
							{
#line 608 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 608 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15599 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 54: {
							{
#line 609 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 609 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_PS, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 55: {
							{
#line 611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_SC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15623 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 56: {
							{
#line 612 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 612 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_SK, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15635 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 57: {
							{
#line 613 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 613 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_SM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15647 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 58: {
							{
#line 614 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 614 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_SO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15659 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 59: {
							{
#line 616 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 616 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ZL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15671 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 60: {
							{
#line 617 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 617 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ZP, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15683 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 61: {
							{
#line 618 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 618 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ZS, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15695 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 62: {
							{
#line 619 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 619 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_XAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15707 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 63: {
							{
#line 620 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 620 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_XPS, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 64: {
							{
#line 621 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 621 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_XSP, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15731 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 65: {
							{
#line 622 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 622 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_XWD, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15743 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 66: {
							{
#line 623 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 623 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_ARABIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15755 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 67: {
							{
#line 624 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 624 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_ARMENIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15767 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 68: {
							{
#line 625 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 625 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_AVESTAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15779 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 69: {
							{
#line 626 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 626 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BALINESE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15791 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 70: {
							{
#line 627 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 627 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BAMUM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15803 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 71: {
							{
#line 628 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 628 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BATAK, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15815 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 72: {
							{
#line 629 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 629 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BENGALI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15827 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 73: {
							{
#line 630 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 630 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BOPOMOFO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15839 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 74: {
							{
#line 631 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 631 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BRAHMI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15851 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 75: {
							{
#line 632 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 632 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BRAILLE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15863 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 76: {
							{
#line 633 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 633 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BUGINESE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15875 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 77: {
							{
#line 634 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 634 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_BUHID, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15887 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 78: {
							{
#line 635 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 635 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CANADIAN_ABORIGINAL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 79: {
							{
#line 636 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 636 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CARIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15911 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 80: {
							{
#line 637 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 637 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CHAM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15923 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 81: {
							{
#line 638 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 638 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CHEROKEE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15935 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 82: {
							{
#line 639 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 639 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_COMMON, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15947 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 83: {
							{
#line 640 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 640 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_COPTIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15959 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 84: {
							{
#line 641 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 641 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CUNEIFORM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15971 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 85: {
							{
#line 642 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 642 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CYPRIOT, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15983 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 86: {
							{
#line 643 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 643 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_CYRILLIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 15995 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 87: {
							{
#line 644 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 644 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_DESERET, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16007 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 88: {
							{
#line 645 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 645 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_DEVANAGARI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16019 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 89: {
							{
#line 646 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 646 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_EGYPTIAN_HIEROGLYPHS, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16031 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 90: {
							{
#line 647 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 647 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_ETHIOPIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16043 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 91: {
							{
#line 648 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 648 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GEORGIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16055 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 92: {
							{
#line 649 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 649 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GLAGOLITIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16067 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 93: {
							{
#line 650 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 650 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GOTHIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16079 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 94: {
							{
#line 651 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 651 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GREEK, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16091 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 95: {
							{
#line 652 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 652 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GUJARATI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 96: {
							{
#line 653 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 653 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_GURMUKHI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16115 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 97: {
							{
#line 655 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 655 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HANGUL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16127 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 98: {
							{
#line 656 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 656 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HANUNOO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16139 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 99: {
							{
#line 657 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 657 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HEBREW, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16151 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 100: {
							{
#line 658 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 658 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HIRAGANA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16163 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 101: {
							{
#line 659 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 659 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_IMPERIAL_ARAMAIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16175 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 102: {
							{
#line 660 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 660 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_INHERITED, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 103: {
							{
#line 661 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 661 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_INSCRIPTIONAL_PAHLAVI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16199 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 104: {
							{
#line 662 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 662 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_INSCRIPTIONAL_PARTHIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16211 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 105: {
							{
#line 663 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 663 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_JAVANESE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16223 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 106: {
							{
#line 664 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 664 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KAITHI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16235 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 107: {
							{
#line 665 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 665 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KANNADA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16247 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 108: {
							{
#line 666 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 666 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KATAKANA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16259 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 109: {
							{
#line 667 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 667 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KAYAH_LI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16271 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 110: {
							{
#line 668 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 668 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KHAROSHTHI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16283 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 111: {
							{
#line 669 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 669 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_KHMER, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16295 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 112: {
							{
#line 670 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 670 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LAO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16307 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 113: {
							{
#line 671 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 671 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LATIN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16319 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 114: {
							{
#line 672 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 672 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LEPCHA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16331 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 115: {
							{
#line 673 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 673 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LIMBU, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16343 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 116: {
							{
#line 674 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 674 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LINEAR_B, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16355 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 117: {
							{
#line 675 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 675 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LISU, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16367 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 118: {
							{
#line 676 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 676 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LYCIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16379 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 119: {
							{
#line 677 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 677 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_LYDIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16391 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 120: {
							{
#line 678 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 678 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_MALAYALAM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16403 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 121: {
							{
#line 679 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 679 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_MANDAIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16415 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 122: {
							{
#line 680 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 680 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_MEETEI_MAYEK, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16427 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 123: {
							{
#line 681 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 681 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_MONGOLIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16439 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 124: {
							{
#line 682 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 682 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_MYANMAR, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16451 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 125: {
							{
#line 683 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 683 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_NEW_TAI_LUE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16463 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 126: {
							{
#line 684 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 684 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_NKO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16475 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 127: {
							{
#line 685 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 685 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OGHAM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 128: {
							{
#line 686 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 686 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OL_CHIKI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 129: {
							{
#line 687 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 687 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OLD_ITALIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16511 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 130: {
							{
#line 688 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 688 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OLD_PERSIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16523 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 131: {
							{
#line 689 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 689 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OLD_SOUTH_ARABIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16535 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 132: {
							{
#line 690 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 690 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OLD_TURKIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16547 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 133: {
							{
#line 691 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 691 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_ORIYA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16559 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 134: {
							{
#line 692 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 692 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_OSMANYA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16571 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 135: {
							{
#line 693 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 693 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_PHAGS_PA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 136: {
							{
#line 694 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 694 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_PHOENICIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16595 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 137: {
							{
#line 695 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 695 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_REJANG, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16607 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 138: {
							{
#line 696 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 696 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_RUNIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16619 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 139: {
							{
#line 697 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 697 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SAMARITAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16631 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 140: {
							{
#line 698 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 698 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SAURASHTRA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16643 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 141: {
							{
#line 699 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 699 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SHAVIAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16655 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 142: {
							{
#line 700 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 700 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SINHALA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16667 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 143: {
							{
#line 701 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 701 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SUNDANESE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16679 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 144: {
							{
#line 702 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 702 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SYLOTI_NAGRI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16691 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 145: {
							{
#line 703 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 703 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_SYRIAC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16703 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 146: {
							{
#line 704 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 704 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAGALOG, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16715 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 147: {
							{
#line 705 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 705 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAGBANWA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16727 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 148: {
							{
#line 706 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 706 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAI_LE, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16739 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 149: {
							{
#line 707 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 707 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAI_THAM, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16751 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 150: {
							{
#line 708 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 708 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAI_VIET, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16763 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 151: {
							{
#line 709 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 709 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TAMIL, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16775 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 152: {
							{
#line 710 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 710 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TELUGU, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16787 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 153: {
							{
#line 711 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 711 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_THAANA, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16799 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 154: {
							{
#line 712 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 712 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_THAI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16811 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 155: {
							{
#line 713 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 713 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TIBETAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16823 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 156: {
							{
#line 714 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 714 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_TIFINAGH, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16835 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 157: {
							{
#line 715 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 715 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_UGARITIC, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16847 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 158: {
							{
#line 716 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 716 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_VAI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16859 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 159: {
							{
#line 717 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 717 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_YI, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16871 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 160: {
							{
#line 718 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 718 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ANY, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16883 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 161: {
							{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Unknown property"); }
								}}
							
#line 16895 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 162: {
							{
#line 581 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 581 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_C, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16907 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 163: {
							{
#line 585 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 585 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16919 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 164: {
							{
#line 587 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 587 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_L, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16931 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 165: {
							{
#line 594 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 594 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_M, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16943 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 166: {
							{
#line 596 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 596 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ME, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16955 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 167: {
							{
#line 598 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 598 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_N, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16967 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 168: {
							{
#line 602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_P, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16979 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 169: {
							{
#line 610 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 610 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_S, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 16991 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 170: {
							{
#line 615 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 615 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_Z, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17003 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 171: {
							{
#line 654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17015 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 172: {
							{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Unknown property"); }
								}}
							
#line 17027 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 173: {
							{
#line 581 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 581 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_C, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17040 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 174: {
							{
#line 585 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 585 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_CO, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17053 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 175: {
							{
#line 587 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 587 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_L, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17066 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 176: {
							{
#line 594 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 594 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_M, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17079 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 177: {
							{
#line 596 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 596 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_ME, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17092 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 178: {
							{
#line 598 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 598 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_N, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17105 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 179: {
							{
#line 602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_P, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17118 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 180: {
							{
#line 610 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 610 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_UCP_S, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17131 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 181: {
							{
#line 654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										currentCls->add(CLASS_SCRIPT_HAN, negated); {top-= 1;cs = stack[top];goto _again;} }
								}}
							
#line 17144 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 182: {
							{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = ((te))-1;
									{
#line 719 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Unknown property"); }
								}}
							
#line 17157 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 183: {
							{
#line 734 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 734 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_C, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17176 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 184: {
							{
#line 742 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 742 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_L, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17195 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 185: {
							{
#line 750 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 750 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_M, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17214 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 186: {
							{
#line 758 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 758 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_N, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;}
									}
								}}
							
#line 17233 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 187: {
							{
#line 766 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 766 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_P, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17252 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 188: {
							{
#line 774 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 774 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_S, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17271 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 189: {
							{
#line 782 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 782 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UCP_Z, negated); 
										if (!inCharClass) {
											currentCls->finalize();
											currentSeq->addComponent(std::move(currentCls));
										}
										{top-= 1;cs = stack[top];goto _again;} 
									}
								}}
							
#line 17290 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 190: {
							{
#line 791 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 791 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Unknown property"); }
								}}
							
#line 17302 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 191: {
							{
#line 797 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 797 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Unsupported POSIX collating "
										"element");
									}
								}}
							
#line 17317 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 192: {
							{
#line 804 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 804 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ALNUM, false);
									}
								}}
							
#line 17331 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 193: {
							{
#line 807 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 807 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ALNUM, true);
									}
								}}
							
#line 17345 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 194: {
							{
#line 810 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 810 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ALPHA, false);
									}
								}}
							
#line 17359 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 195: {
							{
#line 813 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 813 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ALPHA, true);
									}
								}}
							
#line 17373 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 196: {
							{
#line 816 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 816 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ASCII, false);
									}
								}}
							
#line 17387 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 197: {
							{
#line 819 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 819 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_ASCII, true);
									}
								}}
							
#line 17401 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 198: {
							{
#line 822 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 822 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_BLANK, false);
									}
								}}
							
#line 17415 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 199: {
							{
#line 825 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 825 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_BLANK, true);
									}
								}}
							
#line 17429 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 200: {
							{
#line 828 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 828 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_CNTRL, false);
									}
								}}
							
#line 17443 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 201: {
							{
#line 831 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 831 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_CNTRL, true);
									}
								}}
							
#line 17457 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 202: {
							{
#line 834 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 834 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_DIGIT, false);
									}
								}}
							
#line 17471 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 203: {
							{
#line 837 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 837 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_DIGIT, true);
									}
								}}
							
#line 17485 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 204: {
							{
#line 840 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 840 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_GRAPH, false);
									}
								}}
							
#line 17499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 205: {
							{
#line 843 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 843 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_GRAPH, true);
									}
								}}
							
#line 17513 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 206: {
							{
#line 846 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 846 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_LOWER, false);
									}
								}}
							
#line 17527 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 207: {
							{
#line 849 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 849 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_LOWER, true);
									}
								}}
							
#line 17541 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 208: {
							{
#line 852 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 852 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_PRINT, false);
									}
								}}
							
#line 17555 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 209: {
							{
#line 855 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 855 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_PRINT, true);
									}
								}}
							
#line 17569 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 210: {
							{
#line 858 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 858 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_PUNCT, false);
									}
								}}
							
#line 17583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 211: {
							{
#line 861 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 861 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_PUNCT, true);
									}
								}}
							
#line 17597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 212: {
							{
#line 865 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 865 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_SPACE, false);
									}
								}}
							
#line 17611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 213: {
							{
#line 868 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 868 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_SPACE, true);
									}
								}}
							
#line 17625 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 214: {
							{
#line 871 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 871 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UPPER, false);
									}
								}}
							
#line 17639 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 215: {
							{
#line 874 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 874 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_UPPER, true);
									}
								}}
							
#line 17653 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 216: {
							{
#line 877 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 877 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_WORD, false);
									}
								}}
							
#line 17667 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 217: {
							{
#line 880 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 880 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_WORD, true);
									}
								}}
							
#line 17681 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 218: {
							{
#line 883 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 883 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_XDIGIT, false);
									}
								}}
							
#line 17695 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 219: {
							{
#line 886 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 886 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_XDIGIT, true);
									}
								}}
							
#line 17709 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 220: {
							{
#line 891 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 891 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Invalid POSIX named class");
									}
								}}
							
#line 17723 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 221: {
							{
#line 894 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 894 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
												
												DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
												if ((int)stack.size() == top) {
													stack.resize(2 * (top + 1));
												}
											}
											stack[top] = cs; top += 1;cs = 843; goto _again;}}
								}}
							
#line 17744 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 222: {
							{
#line 897 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 897 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										/*noop*/}
								}}
							
#line 17756 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 223: {
							{
#line 899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x08');
									}
								}}
							
#line 17770 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 224: {
							{
#line 903 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 903 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x09');
									}
								}}
							
#line 17784 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 225: {
							{
#line 907 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 907 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x0a');
									}
								}}
							
#line 17798 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 226: {
							{
#line 911 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 911 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x0d');
									}
								}}
							
#line 17812 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 227: {
							{
#line 915 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 915 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x0c');
									}
								}}
							
#line 17826 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 228: {
							{
#line 919 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 919 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x07');
									}
								}}
							
#line 17840 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 229: {
							{
#line 923 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 923 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add('\x1b');
									}
								}}
							
#line 17854 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 230: {
							{
#line 927 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 927 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_HORZ, false);
									}
								}}
							
#line 17868 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 231: {
							{
#line 931 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 931 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_HORZ, true);
									}
								}}
							
#line 17882 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 232: {
							{
#line 935 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 935 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_VERT, false);
									}
								}}
							
#line 17896 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 233: {
							{
#line 939 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 939 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_VERT, true);
									}
								}}
							
#line 17910 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 234: {
							{
#line 943 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 943 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										negated = false;
										{p = p - 1; }
										{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
												
												DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
												if ((int)stack.size() == top) {
													stack.resize(2 * (top + 1));
												}
											}
											stack[top] = cs; top += 1;cs = 559; goto _again;}}
								}}
							
#line 17933 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 235: {
							{
#line 949 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 949 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										negated = false;
										{p = p - 1; }
										{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
												
												DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
												if ((int)stack.size() == top) {
													stack.resize(2 * (top + 1));
												}
											}
											stack[top] = cs; top += 1;cs = 818; goto _again;}}
								}}
							
#line 17956 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 236: {
							{
#line 955 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 955 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										negated = true;
										{p = p - 1; }
										{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
												
												DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
												if ((int)stack.size() == top) {
													stack.resize(2 * (top + 1));
												}
											}
											stack[top] = cs; top += 1;cs = 559; goto _again;}}
								}}
							
#line 17979 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 237: {
							{
#line 961 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 961 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										negated = true;
										{p = p - 1; }
										{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
												
												DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
												if ((int)stack.size() == top) {
													stack.resize(2 * (top + 1));
												}
											}
											stack[top] = cs; top += 1;cs = 818; goto _again;}}
								}}
							
#line 18002 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 238: {
							{
#line 971 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 971 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(octAccumulator);
									}
								}}
							
#line 18016 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 239: {
							{
#line 974 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 974 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(octAccumulator);
									}
								}}
							
#line 18030 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 240: {
							{
#line 978 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 978 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										string oct(ts + 3, te - ts - 4);
										unsigned long val;
										try {
											val = stoul(oct, nullptr, 8);
										} catch (const std::out_of_range &) {
											val = MAX_UNICODE + 1;
										}
										if ((!mode.utf8 && val > 255) || val > MAX_UNICODE) {
											throw LocatedParseError("Value in \\o{...} sequence is too large");
										}
										currentCls->add((unichar)val);
									}
								}}
							
#line 18054 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 241: {
							{
#line 998 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 998 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(accumulator);
									}
								}}
							
#line 18068 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 242: {
							{
#line 1002 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1002 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										// whatever we found here
										currentCls->add(*(ts + 1));
										
									}
								}}
							
#line 18084 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 243: {
							{
#line 1008 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1008 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										string hex(ts + 3, te - ts - 4);
										unsigned long val;
										try {
											val = stoul(hex, nullptr, 16);
										} catch (const std::out_of_range &) {
											val = MAX_UNICODE + 1;
										}
										if (val > MAX_UNICODE) {
											throw LocatedParseError("Value in \\x{...} sequence is too large");
										}
										currentCls->add((unichar)val);
									}
								}}
							
#line 18108 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 244: {
							{
#line 1026 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1026 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										if (te - ts < 3) {
											assert(te - ts == 2);
											throw LocatedParseError(SLASH_C_ERROR);
										} else {
											assert(te - ts == 3);
											currentCls->add(decodeCtrl(ts[2]));
										}
									}
								}}
							
#line 18128 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 245: {
							{
#line 1036 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1036 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_WORD, false);
									}
								}}
							
#line 18142 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 246: {
							{
#line 1040 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1040 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_WORD, true);
									}
								}}
							
#line 18156 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 247: {
							{
#line 1044 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1044 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_SPACE, false);
									}
								}}
							
#line 18170 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 248: {
							{
#line 1048 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1048 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_SPACE, true);
									}
								}}
							
#line 18184 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 249: {
							{
#line 1052 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1052 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_DIGIT, false);
									}
								}}
							
#line 18198 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 250: {
							{
#line 1056 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1056 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(CLASS_DIGIT, true);
									}
								}}
							
#line 18212 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 251: {
							{
#line 1059 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1059 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->addDash();
									}
								}}
							
#line 18226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 252: {
							{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										ostringstream str;
										str << "'\\" << *(ts + 1) << "' at index " << ts - ptr
										<< " not supported in a character class.";
										throw ParseError(str.str());
									}
								}}
							
#line 18243 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 253: {
							{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										ostringstream str;
										str << "'\\" << *(ts + 1) << "' at index " << ts - ptr
										<< " not supported in a character class.";
										throw ParseError(str.str());
									}
								}}
							
#line 18260 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 254: {
							{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										ostringstream str;
										str << "'\\" << *(ts + 1) << "' at index " << ts - ptr
										<< " not supported in a character class.";
										throw ParseError(str.str());
									}
								}}
							
#line 18277 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 255: {
							{
#line 1076 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1076 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										// add the literal char
										currentCls->add(*(ts + 1));
									}
								}}
							
#line 18292 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 256: {
							{
#line 1082 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1082 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										assert(mode.utf8);
										currentCls->add(readUtf8CodePoint2c(ts));
									}
								}}
							
#line 18307 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 257: {
							{
#line 1087 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1087 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										assert(mode.utf8);
										currentCls->add(readUtf8CodePoint3c(ts));
									}
								}}
							
#line 18322 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 258: {
							{
#line 1092 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1092 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										assert(mode.utf8);
										currentCls->add(readUtf8CodePoint4c(ts));
									}
								}}
							
#line 18337 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 259: {
							{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										assert(mode.utf8);
										throwInvalidUtf8();
									}
								}}
							
#line 18352 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 260: {
							{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add((u8)*ts);
									}
								}}
							
#line 18366 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 261: {
							{
#line 1107 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p+1;{
#line 1107 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->finalize();
										currentSeq->addComponent(std::move(currentCls));
										inCharClass = false;
										{cs = 746; goto _again;}}
								}}
							
#line 18382 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 262: {
							{
#line 967 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 967 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Malformed property"); }
								}}
							
#line 18394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 263: {
							{
#line 968 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 968 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										throw LocatedParseError("Malformed property"); }
								}}
							
#line 18406 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 264: {
							{
#line 971 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 971 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(octAccumulator);
									}
								}}
							
#line 18420 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 265: {
							{
#line 974 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 974 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(octAccumulator);
									}
								}}
							
#line 18434 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 266: {
							{
#line 993 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 993 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Value in \\o{...} sequence is non-octal or missing braces");
									}
								}}
							
#line 18448 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 267: {
							{
#line 998 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 998 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										currentCls->add(accumulator);
									}
								}}
							
#line 18462 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
							
							break; 
						}
						case 268: {
							{
#line 1022 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{te = p;p = p - 1;{
#line 1022 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										throw LocatedParseError("Value in \\x{...} sequence is non-hex or missing }");
								}
							}}
						
#line 18476 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
						
						break; 
					}
					case 269: {
						{
#line 1026 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							{te = p;p = p - 1;{
#line 1026 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
									
									if (te - ts < 3) {
										assert(te - ts == 2);
										throw LocatedParseError(SLASH_C_ERROR);
									} else {
										assert(te - ts == 3);
										currentCls->add(decodeCtrl(ts[2]));
									}
								}
							}}
						
#line 18496 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
						
						break; 
					}
					case 270: {
						{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							{te = p;p = p - 1;{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
									
									assert(mode.utf8);
									throwInvalidUtf8();
								}
							}}
						
#line 18511 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
						
						break; 
					}
					case 271: {
						{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							{te = p;p = p - 1;{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
									
									currentCls->add((u8)*ts);
								}
							}}
						
#line 18525 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
						
						break; 
					}
					case 272: {
						{
#line 993 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							{p = ((te))-1;
								{
#line 993 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
									
									throw LocatedParseError("Value in \\o{...} sequence is non-octal or missing braces");
								}
							}}
						
#line 18540 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
						
						break; 
					}
					case 273: {
						{
#line 1022 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							{p = ((te))-1;
								{
#line 1022 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
									
									throw LocatedParseError("Value in \\x{...} sequence is non-hex or missing }");
							}
						}}
					
#line 18555 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 274: {
					{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{p = ((te))-1;
							{
#line 1097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18571 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 275: {
					{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{p = ((te))-1;
							{
#line 1103 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentCls->add((u8)*ts);
							}
						}}
					
#line 18586 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 276: {
					{
#line 1121 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1121 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (currentCls->isNegated()) {
									// Already seen a caret; the second one is not a meta-character.
									inCharClassEarly = false;
									{p = p - 1; } {cs = 819; goto _again;}} else {
									currentCls->negate();
									// Note: we cannot switch off inCharClassEarly here, as /[^]]/
									// needs to use the right square bracket path below.
								}
							}
						}}
					
#line 18607 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 277: {
					{
#line 1134 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1134 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentCls->add(']');
								inCharClassEarly = false;
							}
						}}
					
#line 18622 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 278: {
					{
#line 1139 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1139 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 843; goto _again;}}
						}}
					
#line 18642 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 279: {
					{
#line 1140 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1140 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								/*noop*/}
						}}
					
#line 18654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 280: {
					{
#line 1143 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1143 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								inCharClassEarly = false;
								{p = p - 1; }
								{cs = 819; goto _again;}}
						}}
					
#line 18669 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 281: {
					{
#line 1143 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1143 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								inCharClassEarly = false;
								{p = p - 1; }
								{cs = 819; goto _again;}}
						}}
					
#line 18684 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 282: {
					{
#line 1155 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1155 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								{cs = 746; goto _again;}}
						}}
					
#line 18697 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 283: {
					{
#line 1160 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1160 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint2c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 18716 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 284: {
					{
#line 1169 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1169 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint3c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 18735 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 285: {
					{
#line 1178 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1178 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint4c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 18754 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 286: {
					{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18769 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 287: {
					{
#line 1193 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1193 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, *ts, mode);
							}
						}}
					
#line 18783 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 288: {
					{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18798 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 289: {
					{
#line 1193 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1193 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, *ts, mode);
							}
						}}
					
#line 18812 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 290: {
					{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{p = ((te))-1;
							{
#line 1187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18828 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 291: {
					{
#line 1203 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1203 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								{top-= 1;cs = stack[top];goto _again;}
							}
						}}
					
#line 18842 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 292: {
					{
#line 1208 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1208 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								currentCls->add(readUtf8CodePoint2c(ts));
								inCharClassEarly = false;
							}
						}}
					
#line 18858 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 293: {
					{
#line 1214 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1214 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								currentCls->add(readUtf8CodePoint3c(ts));
								inCharClassEarly = false;
							}
						}}
					
#line 18874 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 294: {
					{
#line 1220 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1220 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								currentCls->add(readUtf8CodePoint4c(ts));
								inCharClassEarly = false;
							}
						}}
					
#line 18890 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 295: {
					{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18905 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 296: {
					{
#line 1232 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1232 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentCls->add(*ts);
								inCharClassEarly = false;
							}
						}}
					
#line 18920 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 297: {
					{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18935 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 298: {
					{
#line 1232 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1232 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentCls->add(*ts);
								inCharClassEarly = false;
							}
						}}
					
#line 18950 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 299: {
					{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{p = ((te))-1;
							{
#line 1226 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 18966 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 300: {
					{
#line 1244 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1244 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								inComment = false; {cs = 746; goto _again;}}
						}}
					
#line 18978 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 301: {
					{
#line 1 "-"
						{te = p+1;}}
					
#line 18987 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 302: {
					{
#line 1256 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1256 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								inComment = false; {cs = 746; goto _again;}}
						}}
					
#line 18999 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 303: {
					{
#line 1 "-"
						{te = p+1;}}
					
#line 19008 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 304: {
					{
#line 1492 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{act = 288;}}
					
#line 19017 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 305: {
					{
#line 1509 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{act = 290;}}
					
#line 19026 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 306: {
					{
#line 1738 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{act = 330;}}
					
#line 19035 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 307: {
					{
#line 363 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 363 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (sequences.empty()) {
									throw LocatedParseError("Unmatched parentheses");
								}
								currentSeq->finalize();
								POP_SEQUENCE;
							}
						}}
					
#line 19053 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 308: {
					{
#line 1275 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1275 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addAlternation();
							}
						}}
					
#line 19067 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 309: {
					{
#line 1280 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1280 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("POSIX named classes are only "
								"supported inside a class");
							}
						}}
					
#line 19082 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 310: {
					{
#line 1287 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1287 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Unsupported POSIX collating "
								"element");
							}
						}}
					
#line 19097 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 311: {
					{
#line 1294 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1294 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								{cs = 838; goto _again;}}
						}}
					
#line 19110 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 312: {
					{
#line 1298 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1298 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								/* noop */ }
						}}
					
#line 19122 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 313: {
					{
#line 1300 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1300 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(generateComponent(CLASS_ANY, false, mode));
							}
						}}
					
#line 19136 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 314: {
					{
#line 1304 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1304 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (mode.utf8) {
									throw LocatedParseError("\\C is unsupported in UTF8");
								}
								currentSeq->addComponent(std::make_unique<ComponentByte>());
							}
						}}
					
#line 19153 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 315: {
					{
#line 1318 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1318 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(0, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_NONGREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19170 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 316: {
					{
#line 1325 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1325 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(0, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_POSSESSIVE)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19187 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 317: {
					{
#line 1339 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1339 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(1, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_NONGREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19204 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 318: {
					{
#line 1346 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1346 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(1, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_POSSESSIVE)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19221 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 319: {
					{
#line 1360 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1360 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(
								0, 1, ComponentRepeat::REPEAT_NONGREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19238 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 320: {
					{
#line 1367 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1367 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(
								0, 1, ComponentRepeat::REPEAT_POSSESSIVE)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19255 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 321: {
					{
#line 1384 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1384 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (repeatN > repeatM || repeatM == 0) {
									throwInvalidRepeat();
								} else if (!currentSeq->addRepeat(
								repeatN, repeatM,
								ComponentRepeat::REPEAT_NONGREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19275 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 322: {
					{
#line 1394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (repeatN > repeatM || repeatM == 0) {
									throwInvalidRepeat();
								} else if (!currentSeq->addRepeat(
								repeatN, repeatM,
								ComponentRepeat::REPEAT_POSSESSIVE)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 19295 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 323: {
					{
#line 323 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 323 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								inComment = true;
								{cs = 849; goto _again;}}
						}}
					
#line 19309 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 324: {
					{
#line 1411 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1411 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								{p = p - 1; } {{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 787; goto _again;}}
						}}
					
#line 19329 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 325: {
					{
#line 1415 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1415 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								assert(0); {p += 1; goto _out; } }
						}}
					
#line 19341 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 326: {
					{
#line 1422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto bound = mode.multiline ? ComponentBoundary::BEGIN_LINE
								: ComponentBoundary::BEGIN_STRING;
								currentSeq->addComponent(std::make_unique<ComponentBoundary>(bound));
							}
						}}
					
#line 19357 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 327: {
					{
#line 1429 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1429 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto bound = mode.multiline ? ComponentBoundary::END_LINE
								: ComponentBoundary::END_STRING_OPTIONAL_LF;
								currentSeq->addComponent(std::make_unique<ComponentBoundary>(bound));
							}
						}}
					
#line 19373 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 328: {
					{
#line 1435 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1435 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto bound = ComponentBoundary::BEGIN_STRING;
								currentSeq->addComponent(std::make_unique<ComponentBoundary>(bound));
							}
						}}
					
#line 19388 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 329: {
					{
#line 1440 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1440 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto bound = ComponentBoundary::END_STRING_OPTIONAL_LF;
								currentSeq->addComponent(std::make_unique<ComponentBoundary>(bound));
							}
						}}
					
#line 19403 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 330: {
					{
#line 1445 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1445 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto bound = ComponentBoundary::END_STRING;
								currentSeq->addComponent(std::make_unique<ComponentBoundary>(bound));
							}
						}}
					
#line 19418 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 331: {
					{
#line 1450 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1450 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(
								std::make_unique<ComponentWordBoundary>(ts - ptr, false, mode));
							}
						}}
					
#line 19433 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 332: {
					{
#line 1455 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1455 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(
								std::make_unique<ComponentWordBoundary>(ts - ptr, true, mode));
							}
						}}
					
#line 19448 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 333: {
					{
#line 1465 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1465 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x09', mode);
							}
						}}
					
#line 19462 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 334: {
					{
#line 1469 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1469 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x0a', mode);
							}
						}}
					
#line 19476 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 335: {
					{
#line 1473 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1473 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x0d', mode);
							}
						}}
					
#line 19490 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 336: {
					{
#line 1477 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1477 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x0c', mode);
							}
						}}
					
#line 19504 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 337: {
					{
#line 1481 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1481 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x07', mode);
							}
						}}
					
#line 19518 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 338: {
					{
#line 1485 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1485 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, '\x1b', mode);
							}
						}}
					
#line 19532 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 339: {
					{
#line 1489 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1489 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, octAccumulator, mode);
							}
						}}
					
#line 19546 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 340: {
					{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
							}
						}}
					
#line 19563 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 341: {
					{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// Accumulator is a negative offset.
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								if (accumulator >= groupIndex) {
									throw LocatedParseError("Invalid reference");
								}
								unsigned idx = groupIndex - accumulator;
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(idx));
							}
						}}
					
#line 19585 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 342: {
					{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
							}
						}}
					
#line 19602 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 343: {
					{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// Accumulator is a negative offset.
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								if (accumulator >= groupIndex) {
									throw LocatedParseError("Invalid reference");
								}
								unsigned idx = groupIndex - accumulator;
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(idx));
							}
						}}
					
#line 19624 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 344: {
					{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(label));
							}
						}}
					
#line 19638 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 345: {
					{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(label));
							}
						}}
					
#line 19652 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 346: {
					{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(label));
							}
						}}
					
#line 19666 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 347: {
					{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(label));
							}
						}}
					
#line 19680 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 348: {
					{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 499 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(label));
							}
						}}
					
#line 19694 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 349: {
					{
#line 1550 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1550 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "Onigiruma subroutine call at index " << ts - ptr <<
								" not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 19711 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 350: {
					{
#line 1561 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1561 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								string oct(ts + 3, te - ts - 4);
								unsigned long val;
								try {
									val = stoul(oct, nullptr, 8);
								} catch (const std::out_of_range &) {
									val = MAX_UNICODE + 1;
								}
								if ((!mode.utf8 && val > 255) || val > MAX_UNICODE) {
									throw LocatedParseError("Value in \\o{...} sequence is too large");
								}
								addEscapedOctal(currentSeq, (unichar)val, mode);
							}
						}}
					
#line 19735 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 351: {
					{
#line 1579 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1579 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addEscapedHex(currentSeq, accumulator, mode);
							}
						}}
					
#line 19749 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 352: {
					{
#line 1583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1583 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								string hex(ts + 3, te - ts - 4);
								unsigned long val;
								try {
									val = stoul(hex, nullptr, 16);
								} catch (const std::out_of_range &) {
									val = MAX_UNICODE + 1;
								}
								if (val > MAX_UNICODE) {
									throw LocatedParseError("Value in \\x{...} sequence is too large");
								}
								addEscapedHex(currentSeq, (unichar)val, mode);
							}
						}}
					
#line 19773 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 353: {
					{
#line 1601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (te - ts < 3) {
									assert(te - ts == 2);
									throw LocatedParseError(SLASH_C_ERROR);
								} else {
									assert(te - ts == 3);
									addLiteral(currentSeq, decodeCtrl(ts[2]), mode);
								}
							}
						}}
					
#line 19793 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 354: {
					{
#line 1611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1611 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "'\\" << *(ts + 1) << "' at index " << ts - ptr
								<< " not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 19810 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 355: {
					{
#line 1619 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1619 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_WORD, false, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19825 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 356: {
					{
#line 1624 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1624 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_WORD, true, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19840 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 357: {
					{
#line 1629 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1629 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_SPACE, false, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19855 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 358: {
					{
#line 1634 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1634 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_SPACE, true, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19870 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 359: {
					{
#line 1639 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1639 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_DIGIT, false, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 360: {
					{
#line 1644 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1644 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_DIGIT, true, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19900 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 361: {
					{
#line 1649 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1649 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_HORZ, false, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19915 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 362: {
					{
#line 1654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1654 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_HORZ, true, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19930 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 363: {
					{
#line 1659 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1659 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_VERT, false, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19945 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 364: {
					{
#line 1664 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1664 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto cc = generateComponent(CLASS_VERT, true, mode);
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 19960 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 365: {
					{
#line 1669 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1669 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!currentCls && !inCharClass);
								currentCls = getComponentClass(mode);
								negated = false;
								{p = p - 1; }
								{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 559; goto _again;}}
						}}
					
#line 19985 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 366: {
					{
#line 1677 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1677 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!currentCls && !inCharClass);
								currentCls = getComponentClass(mode);
								negated = false;
								{p = p - 1; }
								{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 818; goto _again;}}
						}}
					
#line 20010 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 367: {
					{
#line 1685 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1685 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!currentCls && !inCharClass);
								currentCls = getComponentClass(mode);
								negated = true;
								{p = p - 1; }
								{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 559; goto _again;}}
						}}
					
#line 20035 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 368: {
					{
#line 1693 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1693 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!currentCls && !inCharClass);
								currentCls = getComponentClass(mode);
								negated = true;
								{p = p - 1; }
								{{
#line 1904 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
										
										DEBUG_PRINTF("stack %zu top %d\n", stack.size(), top);
										if ((int)stack.size() == top) {
											stack.resize(2 * (top + 1));
										}
									}
									stack[top] = cs; top += 1;cs = 818; goto _again;}}
						}}
					
#line 20060 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 369: {
					{
#line 1705 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1705 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "\\R at index " << ts - ptr << " not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 20076 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 370: {
					{
#line 1712 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1712 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "\\K at index " << ts - ptr << " not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 20092 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 371: {
					{
#line 1727 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1727 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "\\G at index " << ts - ptr << " not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 20108 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 372: {
					{
#line 1733 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1733 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								currentSeq->addComponent(std::make_unique<ComponentEUS>(ts - ptr, mode));
							}
						}}
					
#line 20122 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 373: {
					{
#line 1738 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1738 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, *(ts + 1), mode);
							}
						}}
					
#line 20136 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 374: {
					{
#line 317 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 317 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								inComment = true;
								{cs = 848; goto _again;}}
						}}
					
#line 20150 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 375: {
					{
#line 434 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 434 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								mode = newMode;
								currentSeq->addComponent(std::make_unique<ComponentEmpty>());
							}
						}}
					
#line 20165 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 376: {
					{
#line 356 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 356 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								mode = newMode;
								currentSeq =
								enterSequence(currentSeq, std::make_unique<ComponentSequence>());
							}
						}}
					
#line 20182 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 377: {
					{
#line 370 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 370 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentAssertion>(ComponentAssertion::LOOKAHEAD,
								ComponentAssertion::POS));
							}
						}}
					
#line 20199 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 378: {
					{
#line 376 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 376 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentAssertion>(ComponentAssertion::LOOKAHEAD,
								ComponentAssertion::NEG));
							}
						}}
					
#line 20216 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 379: {
					{
#line 382 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 382 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentAssertion>(ComponentAssertion::LOOKBEHIND,
								ComponentAssertion::POS));
							}
						}}
					
#line 20233 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 380: {
					{
#line 388 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 388 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentAssertion>(ComponentAssertion::LOOKBEHIND,
								ComponentAssertion::NEG));
							}
						}}
					
#line 20250 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 381: {
					{
#line 394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Embedded code is not supported");
							}
						}}
					
#line 20264 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 382: {
					{
#line 394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 394 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Embedded code is not supported");
							}
						}}
					
#line 20278 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 383: {
					{
#line 417 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 417 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentAtomicGroup>());
							}
						}}
					
#line 20294 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 384: {
					{
#line 337 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 337 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!label.empty()); // should be guaranteed by machine
								char c = *label.begin();
								if (c >= '0' && c <= '9') {
									throw LocatedParseError("Group name cannot begin with a digit");
								}
								if (!groupNames.insert(label).second) {
									throw LocatedParseError("Two named subpatterns use the name '" + label + "'");
								}
								PUSH_SEQUENCE;
								auto seq = std::make_unique<ComponentSequence>();
								seq->setCaptureIndex(groupIndex++);
								seq->setCaptureName(label);
								currentSeq = enterSequence(currentSeq, std::move(seq));
							}
						}}
					
#line 20320 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 385: {
					{
#line 400 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 400 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Subpattern reference unsupported");
							}
						}}
					
#line 20334 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 386: {
					{
#line 400 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 400 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Subpattern reference unsupported");
							}
						}}
					
#line 20348 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 387: {
					{
#line 1784 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1784 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto a = std::make_unique<ComponentAssertion>(
								ComponentAssertion::LOOKAHEAD, ComponentAssertion::POS);
								ComponentAssertion *a_seq = a.get();
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(std::move(a)));
								PUSH_SEQUENCE;
								currentSeq = a_seq;
							}
						}}
					
#line 20369 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 388: {
					{
#line 1795 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1795 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto a = std::make_unique<ComponentAssertion>(
								ComponentAssertion::LOOKAHEAD, ComponentAssertion::NEG);
								ComponentAssertion *a_seq = a.get();
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(std::move(a)));
								PUSH_SEQUENCE;
								currentSeq = a_seq;
							}
						}}
					
#line 20390 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 389: {
					{
#line 1806 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1806 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto a = std::make_unique<ComponentAssertion>(
								ComponentAssertion::LOOKBEHIND, ComponentAssertion::POS);
								ComponentAssertion *a_seq = a.get();
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(std::move(a)));
								PUSH_SEQUENCE;
								currentSeq = a_seq;
							}
						}}
					
#line 20411 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 390: {
					{
#line 1817 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1817 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								auto a = std::make_unique<ComponentAssertion>(
								ComponentAssertion::LOOKBEHIND, ComponentAssertion::NEG);
								ComponentAssertion *a_seq = a.get();
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(std::move(a)));
								PUSH_SEQUENCE;
								currentSeq = a_seq;
							}
						}}
					
#line 20432 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 391: {
					{
#line 1829 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1829 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Pattern recursion not supported");
							}
						}}
					
#line 20446 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 392: {
					{
#line 403 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 403 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								PUSH_SEQUENCE;
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(accumulator));
							}
						}}
					
#line 20465 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 393: {
					{
#line 411 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 411 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								assert(!label.empty());
								currentSeq = enterSequence(currentSeq,
								std::make_unique<ComponentCondReference>(label));
							}
						}}
					
#line 20482 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 394: {
					{
#line 1845 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1845 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								ostringstream str;
								str << "Callout at index " << ts - ptr << " not supported.";
								throw ParseError(str.str());
							}
						}}
					
#line 20498 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 395: {
					{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Unrecognised character after (?");
							}
						}}
					
#line 20512 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 396: {
					{
#line 1858 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1858 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint2c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 20531 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 397: {
					{
#line 1867 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1867 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint3c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 20550 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 398: {
					{
#line 1876 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1876 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								/* leverage ComponentClass to generate the vertices */
								auto cc = getComponentClass(mode);
								cc->add(readUtf8CodePoint4c(ts));
								cc->finalize();
								currentSeq->addComponent(std::move(cc));
							}
						}}
					
#line 20569 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 399: {
					{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(mode.utf8);
								throwInvalidUtf8();
							}
						}}
					
#line 20584 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 400: {
					{
#line 1894 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1894 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (mode.ignore_space == false) {
									addLiteral(currentSeq, *ts, mode);
								}
							}
						}}
					
#line 20600 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 401: {
					{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p+1;{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, *ts, mode);
							}
						}}
					
#line 20614 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 402: {
					{
#line 329 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 329 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								PUSH_SEQUENCE;
								auto seq = std::make_unique<ComponentSequence>();
								seq->setCaptureIndex(groupIndex++);
								currentSeq = enterSequence(currentSeq, std::move(seq));
							}
						}}
					
#line 20631 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 403: {
					{
#line 422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								assert(!currentCls);
								assert(!inCharClass); // not reentrant
								currentCls = getComponentClass(mode);
								inCharClass = true;
								inCharClassEarly = true;
								currentClsBegin = ts;
								{cs = 836; goto _again;}}
						}}
					
#line 20650 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 404: {
					{
#line 1311 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1311 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(0, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_GREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 20667 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 405: {
					{
#line 1332 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1332 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(1, ComponentRepeat::NoLimit,
								ComponentRepeat::REPEAT_GREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 20684 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 406: {
					{
#line 1353 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1353 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (!currentSeq->addRepeat(
								0, 1, ComponentRepeat::REPEAT_GREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 20701 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 407: {
					{
#line 1374 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1374 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (repeatN > repeatM || repeatM == 0) {
									throwInvalidRepeat();
								} else if (!currentSeq->addRepeat(
								repeatN, repeatM,
								ComponentRepeat::REPEAT_GREEDY)) {
									throwInvalidRepeat();
								}
							}
						}}
					
#line 20721 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 408: {
					{
#line 1489 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1489 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, octAccumulator, mode);
							}
						}}
					
#line 20735 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 409: {
					{
#line 1492 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1492 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// If there are enough capturing sub expressions, this may be
								// a back reference
								accumulator = parseAsDecimal(octAccumulator);
								if (accumulator < groupIndex) {
									currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
								} else {
									addEscapedOctal(currentSeq, octAccumulator, mode);
								}
							}
						}}
					
#line 20756 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 410: {
					{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
							}
						}}
					
#line 20773 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 411: {
					{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 480 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
							}
						}}
					
#line 20790 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 412: {
					{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 487 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// Accumulator is a negative offset.
								if (accumulator == 0) {
									throw LocatedParseError("Numbered reference cannot be zero");
								}
								if (accumulator >= groupIndex) {
									throw LocatedParseError("Invalid reference");
								}
								unsigned idx = groupIndex - accumulator;
								currentSeq->addComponent(std::make_unique<ComponentBackReference>(idx));
							}
						}}
					
#line 20812 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 413: {
					{
#line 1558 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1558 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Invalid reference after \\g");
							}
						}}
					
#line 20826 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 414: {
					{
#line 1575 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1575 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Value in \\o{...} sequence is non-octal or missing braces");
							}
						}}
					
#line 20840 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 415: {
					{
#line 1579 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1579 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addEscapedHex(currentSeq, accumulator, mode);
							}
						}}
					
#line 20854 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
					
					break; 
				}
				case 416: {
					{
#line 1597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						{te = p;p = p - 1;{
#line 1597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								throw LocatedParseError("Value in \\x{...} sequence is non-hex or missing }");
						}
					}}
				
#line 20868 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 417: {
				{
#line 1601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1601 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							if (te - ts < 3) {
								assert(te - ts == 2);
								throw LocatedParseError(SLASH_C_ERROR);
							} else {
								assert(te - ts == 3);
								addLiteral(currentSeq, decodeCtrl(ts[2]), mode);
							}
						}
					}}
				
#line 20888 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 418: {
				{
#line 1701 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1701 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							throw LocatedParseError("Malformed property"); }
					}}
				
#line 20900 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 419: {
				{
#line 1702 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1702 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							throw LocatedParseError("Malformed property"); }
					}}
				
#line 20912 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 420: {
				{
#line 1720 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1720 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							ostringstream str;
							str << "\\k at index " << ts - ptr << " not supported.";
							throw ParseError(str.str());
						}
					}}
				
#line 20928 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 421: {
				{
#line 1743 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1743 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							assert(ts + 1 == pe);
							ostringstream str;
							str << "Unescaped \\ at end of input, index " << ts - ptr << ".";
							throw ParseError(str.str());
						}
					}}
				
#line 20945 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 422: {
				{
#line 397 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 397 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							throw LocatedParseError("Conditional subpattern unsupported");
						}
					}}
				
#line 20959 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 423: {
				{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							throw LocatedParseError("Unrecognised character after (?");
						}
					}}
				
#line 20973 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 424: {
				{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							assert(mode.utf8);
							throwInvalidUtf8();
						}
					}}
				
#line 20988 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 425: {
				{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{te = p;p = p - 1;{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							addLiteral(currentSeq, *ts, mode);
						}
					}}
				
#line 21002 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 426: {
				{
#line 329 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{p = ((te))-1;
						{
#line 329 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							PUSH_SEQUENCE;
							auto seq = std::make_unique<ComponentSequence>();
							seq->setCaptureIndex(groupIndex++);
							currentSeq = enterSequence(currentSeq, std::move(seq));
						}
					}}
				
#line 21020 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 427: {
				{
#line 422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{p = ((te))-1;
						{
#line 422 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							assert(!currentCls);
							assert(!inCharClass); // not reentrant
							currentCls = getComponentClass(mode);
							inCharClass = true;
							inCharClassEarly = true;
							currentClsBegin = ts;
							{cs = 836; goto _again;}}
					}}
				
#line 21040 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 428: {
				{
#line 1558 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{p = ((te))-1;
						{
#line 1558 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							throw LocatedParseError("Invalid reference after \\g");
						}
					}}
				
#line 21055 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 429: {
				{
#line 1575 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{p = ((te))-1;
						{
#line 1575 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							throw LocatedParseError("Value in \\o{...} sequence is non-octal or missing braces");
						}
					}}
				
#line 21070 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
			case 430: {
				{
#line 1597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
					{p = ((te))-1;
						{
#line 1597 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
							
							throw LocatedParseError("Value in \\x{...} sequence is non-hex or missing }");
					}
				}}
			
#line 21085 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 431: {
			{
#line 1720 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
				{p = ((te))-1;
					{
#line 1720 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						
						ostringstream str;
						str << "\\k at index " << ts - ptr << " not supported.";
						throw ParseError(str.str());
					}
				}}
			
#line 21102 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 432: {
			{
#line 397 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
				{p = ((te))-1;
					{
#line 397 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						
						throw LocatedParseError("Conditional subpattern unsupported");
					}
				}}
			
#line 21117 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 433: {
			{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
				{p = ((te))-1;
					{
#line 1853 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						
						throw LocatedParseError("Unrecognised character after (?");
					}
				}}
			
#line 21132 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 434: {
			{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
				{p = ((te))-1;
					{
#line 1885 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						
						assert(mode.utf8);
						throwInvalidUtf8();
					}
				}}
			
#line 21148 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 435: {
			{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
				{p = ((te))-1;
					{
#line 1899 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
						
						addLiteral(currentSeq, *ts, mode);
					}
				}}
			
#line 21163 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
		case 436: {
			{
#line 1 "NONE"
				{switch( act ) {
						case 288:  {
							p = ((te))-1;
							{
#line 1492 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// If there are enough capturing sub expressions, this may be
								// a back reference
								accumulator = parseAsDecimal(octAccumulator);
								if (accumulator < groupIndex) {
									currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
								} else {
									addEscapedOctal(currentSeq, octAccumulator, mode);
								}
							}
							break; 
						}
						case 290:  {
							p = ((te))-1;
							{
#line 1509 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								// if there are enough left parens to this point, back ref
								if (accumulator < groupIndex) {
									currentSeq->addComponent(std::make_unique<ComponentBackReference>(accumulator));
								} else {
									// Otherwise, we interpret the first three digits as an
									// octal escape, and the remaining characters stand for
									// themselves as literals.
									const char *s = ts;
									unsigned int accum = 0;
									unsigned int oct_digits = 0;
									assert(*s == '\\'); // token starts at backslash
									for (++s; s < te && oct_digits < 3; ++oct_digits, ++s) {
										u8 digit = *s - '0';
										if (digit < 8) {
											accum = digit + accum * 8;
										} else {
											break;
										}
									}
									
									if (oct_digits > 0) {
										addEscapedOctal(currentSeq, accum, mode);
									}
									
									// And then the rest of the digits, if any, are literal.
									for (; s < te; ++s) {
										addLiteral(currentSeq, *s, mode);
									}
								}
							}
							break; 
						}
						case 330:  {
							p = ((te))-1;
							{
#line 1738 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"
								
								addLiteral(currentSeq, *(ts + 1), mode);
							}
							break; 
						}
					}}
			}
			
#line 21236 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
			
			break; 
		}
	}
	_acts += 1;
	_nacts -= 1;
}

goto _again;

_again: {}
if ( p == eof ) {
	if ( cs >= 746 )
		goto _out;
}
else {
	_acts = ( _regex_actions + (_regex_to_state_actions[cs]));
	_nacts = (unsigned int)(*( _acts)); _acts += 1;
	while ( _nacts > 0 ) {
		switch ( (*( _acts)) ) {
			case 23: {
				{
#line 1 "NONE"
					{ts = 0;}}
				
#line 21262 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.cpp"
				
				break; 
			}
		}
		_acts += 1;
		_nacts -= 1;
	}
	
	if ( cs != 0 ) {
		p += 1;
		goto _resume;
	}
}
_out: {}
}

#line 1984 "/builddir/build/BUILD/vectorscan-d29730e1cb9daaa66bda63426cdce83505d2c809/src/parser/Parser.rl"


if (p != pe && *p != '\0') {
// didn't make it to the end of our input, but we didn't throw a ParseError?
assert(0);
ostringstream str;
str << "Parse error at index " << (p - ptr) << ".";
throw ParseError(str.str());
}

if (currentCls) {
assert(inCharClass);
assert(currentClsBegin);
ostringstream oss;
oss << "Unterminated character class starting at index "
<< currentClsBegin - ptr << ".";
throw ParseError(oss.str());
}

if (inComment) {
throw ParseError("Unterminated comment.");
}

if (!sequences.empty()) {
ostringstream str;
str << "Missing close parenthesis for group started at index "
<< sequences.back().seqOffset << ".";
throw ParseError(str.str());
}

// Unlikely, but possible
if (groupIndex > 65535) {
throw ParseError("The maximum number of capturing subexpressions is 65535.");
}

// Finalize the top-level sequence, which will take care of any
// top-level alternation.
currentSeq->finalize();
assert(currentSeq == rootSeq.get());

// Ensure that all references are valid.
checkReferences(*rootSeq, groupIndex, groupNames);

return std::move(rootSeq);
} catch (LocatedParseError &error) {
if (ts >= ptr && ts <= pe) {
error.locate(ts - ptr);
} else {
error.locate(0);
}
throw;
}
}

} // namespace ue2
